# -*- coding: utf-8 -*-
"""
This program provides motivation for seasonal adjustment of the WARN layoff
data in Appendix B.2 of "Advance Layoff Notices and Aggregate Job Loss" 
by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    WARNData_NSA_20211230.xlsx : Excel file with the state-level WARN data

The main output of this program is
    table_a2 : an array holding the values for Table A.2 in the appendix

This program also saves
    Seasonal_layoffs.pdf : Figure A.5 in the paper
"""


#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Import Pandas, NumPy, Matplotlib, and a Scipy function
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import chi2

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the program

# String with the date of the data to call
str_date = '20211230'   # Format: 20200115 corresponds to January 15, 2020

# Number of years in the balanced panel
nyears = 21

# States that create a balanced panel from 1999 through 2019
balanced_list = ['AL', 'CT', 'FL', 'IL', 'IN', 'KS', 'MI', 'NC', 'OH', 'TX',
                 'VA', 'WI']


#%% Function to Estimate with OLS

def make_ols(lhs_in, rhs_in):
    
    # Dimensions of the input matrix
    TT, NN = rhs_in.shape
    
    # Useful matrices
    Q = np.zeros((NN, NN))
    Q[:, :] = (rhs_in.T @ rhs_in)/TT
    
    Qinv = np.zeros((NN, NN))
    Qinv[:, :] = np.linalg.inv(Q)
    
    # OLS estimation
    beta_out = np.zeros((NN, 1))
    beta_out[:, :] = Qinv @ (rhs_in.T @ lhs_in)/TT
    
    # Residuals
    ee = np.zeros((TT, 1))
    ee[:, :] = lhs_in - rhs_in @ beta_out
    
    # Covariance matrix
    zz = np.zeros((TT, NN))
    zz[:, :] = np.multiply(rhs_in, (ee @ np.ones((1, NN))))
    
    omega = np.ones((NN, NN))
    omega[:, :] = (zz.T @ zz)/TT
    
    V_out = np.zeros((NN, NN))
    V_out[:, :] = Qinv @ omega @ Qinv
    
    # Standard errors
    se_out = np.zeros(NN)
    se_out[:] = np.sqrt(np.diag(V_out)/TT)
    
    # Wald statistic
    R = np.zeros((NN, 1))
    R[0, 0] = 1
    R[11, 0] = -1
    
    diff_out = np.zeros((1, 1))
    diff_out[0, 0] = R.T @ beta_out
    
    w_out = np.zeros((1, 1))
    w_out[0, 0] = TT*diff_out.T @ np.linalg.inv(R.T @ V_out @ R) @ diff_out
    
    return beta_out, V_out, se_out, diff_out, w_out


#%% Load the WARN Data

# Dataframe for number of workers affected by WARN notices by month and state
str_file_name = 'WARNData_NSA_' + str_date + '.xlsx'
path_warn_data = os.path.join(path_dir, 'data_input', str_file_name)
df_warn_all = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)


#%% Construct the Dummy Variables

# Storage array
dummy = np.zeros((12*nyears, 12))

# Loop to populate dummy variables
for jj in range(nyears):
    dummy[0 + 12*jj, 0] = 1
    dummy[1 + 12*jj, 1] = 1
    dummy[2 + 12*jj, 2] = 1
    dummy[3 + 12*jj, 3] = 1
    dummy[4 + 12*jj, 4] = 1
    dummy[5 + 12*jj, 5] = 1
    dummy[6 + 12*jj, 6] = 1
    dummy[7 + 12*jj, 7] = 1
    dummy[8 + 12*jj, 8] = 1
    dummy[9 + 12*jj, 9] = 1
    dummy[10 + 12*jj, 10] = 1
    dummy[11 + 12*jj, 11] = 1


#%% Loop to Select Balanced Panel from 1999 through 2019

# Number of states in the balanced panel
nvars = len(balanced_list)

# Dataframe to store balanced panel
df_warn_balanced = pd.DataFrame(index=df_warn_all.loc['19990101':'20191231', :].index,
                                columns=balanced_list)

# Dataframe to hold the results
table_a2 = pd.DataFrame(index=balanced_list,
                        columns=['Jan', 'Dec', 'diff', 'ratio', 'wald'])

# Storage arrays for joint inference
beta_array = np.zeros((nvars, 12))
diff_vec = np.zeros((12, 1))
zbig = np.zeros((nyears*12, nvars*12))

# Loop to compute regressions and do inference
for loop in range(nvars):
    # State name
    state_name = balanced_list[loop]
    
    # Dates with observed data
    state_index = df_warn_all.loc['19990101':'20191231', state_name].index
    
    # Populate the balanced panel dataframe
    df_warn_balanced.loc[:, state_name] = df_warn_all.loc['19990101':'20191231', state_name]
    
    # Number of observations
    nsample = len(state_index)
    
    # Convert the data to numpy
    X_nsa = np.zeros((nyears*12, 1))
    X_nsa[:, 0] = df_warn_balanced.loc[:, state_name].to_numpy()
    
    # OLS estimation
    beta_hat, V_hat, se, diff, wald = make_ols(X_nsa, dummy)
    
    beta_array[loop, :] = beta_hat[:, 0]
    diff_vec[loop, 0] = diff
    
    # Populate the array
    table_a2.loc[state_name, 'Jan'] = np.round(beta_hat[0, 0], 1)
    table_a2.loc[state_name, 'Dec'] = np.round(beta_hat[11, 0], 1)
    table_a2.loc[state_name, 'diff'] = np.round(diff[0, 0], 1)
    table_a2.loc[state_name, 'ratio'] = np.round(beta_hat[0, 0]/beta_hat[11, 0], 1)
    table_a2.loc[state_name, 'wald'] = np.round(wald[0, 0], 4)
    
    # Residuals
    resid = np.zeros((12*nyears, 1))
    resid[:, :] = X_nsa - dummy @ beta_hat
    
    # Populate matrix for joint inference
    zbig[:, loop*12:(loop+1)*12] = np.multiply(dummy, (resid @ np.ones((1, 12))))
    
    # Delete temporary variables
    del state_name, state_index, nsample

# Joint covariance matrix
Qbig = np.zeros((nvars*12, nvars*12))
Qbig[:, :] = np.kron(np.eye(nvars), np.linalg.inv((dummy.T @ dummy)/(12*nyears)))

Obig = np.zeros((nvars*12, nvars*12))
Obig[:, :] = (zbig.T @ zbig)/(nyears*12)

Vbig = np.zeros((nvars*12, nvars*12))
Vbig[:, :] = Qbig @ Obig @ Qbig

# Joint inference
R = np.zeros((nvars, nvars*12))
for jj in range(nvars):
    R[jj, jj*12] = 1
    R[jj, 11+jj*12] = -1

wald_joint = np.zeros((1, 1))
wald_joint[:, :] = 12*nyears*diff_vec.T @ np.linalg.inv(R @ Vbig @ R.T) @ diff_vec


#%% Print Aggregate Results

print('Average aggregate layoffs in Jan are '
      + str(int(np.round(np.sum(table_a2.loc[:, 'Jan']), 0))))
print('Average aggregate layoffs in Dec are '
      + str(int(np.round(np.sum(table_a2.loc[:, 'Dec']), 0))))
print('Average aggregate difference between Jan and Dec is '
      + str(int(np.round(np.sum(diff_vec[:, 0]), 0))))
print('The joint Wald statistic for differences across states is '
      + str(np.round(wald_joint[0, 0], 2)))
print('The 5% critical value from the chi-squared distribution with '
      + str(nvars) + ' degress of freedom is '
      + str(np.round(chi2.ppf(0.95, nvars), 2)))
print(' ')


#%% Figure to Summarize National Layoffs by Month

# National level average for the balanced panel states
national_avg_nsa = np.zeros(12)
national_avg_nsa[:] = np.sum(beta_array, axis=0)

# List of months for the plot
month_list = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
              'Oct', 'Nov', 'Dec']

# Figure and axes instances
fig, ax = plt.subplots()

# Plot the results
ax.plot(month_list, national_avg_nsa/1000, marker='x',
        color='black', linewidth=1)
ax.set_ylabel('Number of workers (Thousands, NSA)', fontsize=10)

# Save the figure
plt.savefig('Seasonal_layoffs.pdf', bbox_inches='tight')


#%% Disply the Run Time in Minutes

print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


