# -*- coding: utf-8 -*-
"""
This program analyzes the factor loadings from the DFM for Appendix B.4 of
"Advance Layoff Notices and Aggregate Job Loss" 
by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    seasonal_module.py : module with functions to seasonally adjust the data
    warn_module.py : module with functions to estimate the DFM
    WARNData_NSA_20211230.xlsx : Excel file with the state-level WARN data
    state_employment.xlsx : Excel file with state-level employment data

The main output of this program is
    table_a3 : an array holding the values for Table A.3 in the appendix
    table_a4 : an array holding the values for Table A.4 in the appendix
    table_a5 : an array holding the values for Table A.5 in the appendix
"""


#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Import Pandas and NumPy
import pandas as pd
import numpy as np

# Functions for seasonal adjustment
from modules.seasonal_module import sa_func

# Functions for estimating the DFM
from modules.dfm_em_module import make_guess, em_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the Program

# String with the date of the data to call
str_date = '20211230'   # Format: 20200115 corresponds to January 15, 2020

# String to choose first date for WARN data
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# Number of lags in the state equation
nlags = 1

# Parameters of the EM algorithm for seasonal adjustment
nloop = 20000       # maximum number of iterations in the EM algrorithm
thresh_sa = 1e-3    # threshold for convergence

# Parameters of the EM algorithm for the dynamic factor model
thresh = 1e-5       # threshold for convergence


#%% Load the WARN Data

# Dataframe for number of workers affected by WARN notices by month and state
str_file_name = 'WARNData_NSA_' + str_date + '.xlsx'
path_warn_data = os.path.join(path_dir, 'data_input', str_file_name)
df_warn_all = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)

# Drop small states or states with too many observations with a value of 0
df_warn = df_warn_all.drop(columns=['AK', 'ID', 'KY', 'ND', 'NM',
                                    'OK', 'RI', 'UT', 'WV'])

# Number of states in the sample
nvars = len(df_warn.columns)


#%% Load the State-Level Variables for Analysis

df_state = pd.read_excel('state_employment.xlsx', 'analysis', index_col=0)


#%% Loop to Seasonally Adjust the WARN Data

# Dataframe to store seasonally adjusted data
df_warn_sa = pd.DataFrame(index=df_warn.index, columns=df_warn.columns)

# Loop for seasonal adjustment
for loop in range(nvars):
    # State name
    state_name = df_warn.columns[loop]
    
    # Dates with observed data
    state_index = df_warn.loc[:, state_name].dropna().index
    
    # Number of observations excluding NaNs
    nsample = len(state_index)
    
    # Convert the data to NumPy while dropping NaNs
    X_nsa = np.zeros((nsample, 1))
    X_nsa[:, 0] = df_warn.loc[:, state_name].dropna().to_numpy()
    
    # Seasonally adjust the data
    data_sa = sa_func(X_nsa, nloop, thresh_sa, state_name)
    
    # Store seasonally adjusted data in dataframe
    df_warn_sa.loc[state_index, state_name] = data_sa[:, 0]
    
    # Delete temprary variables
    del state_name, state_index, nsample, data_sa


#%% Adjust the WARN Data and Construct the Selection Matrices for Missing Data

# Number of observations
nsample = len(df_warn_sa.loc[str_warn:, :].index)

# Convert the data to NumPy
X = np.zeros((nsample, nvars))
X[:, :] = df_warn_sa.loc[str_warn:, :].to_numpy()

# Set zero values to NaN and construct the selection matrices
W = np.zeros((nvars, nvars, nsample))
for jj in range(nsample):
    # Set zero values to NaN
    for ii in range(nvars):
        if X[jj, ii] < 1:
            X[jj, ii] = np.nan
    
    # Selection matrices
    W[:, :, jj] = np.diag(np.isnan(X[jj, :])==0)

# Delete temprary variables
del jj


#%% Initial Guess for the EM Algorithm

# Take logs of the data
X = np.log(X)

# Guess of model parameters
Fguess, Aguess, Qguess, Lguess, Rguess, Zstd, Avg, Var = make_guess(X, nlags)

# Guess of the initial state
F0guess = np.zeros((nlags, 1))

# Guess of covaraiance of initial state
Qvec = np.zeros((nlags*nlags, 1))
Qvec[0, 0] = Qguess[0, 0]

tmp = np.zeros((nlags*nlags, nlags*nlags))
tmp[:, :] = np.eye(nlags*nlags) - np.kron(Aguess, Aguess)

P0guess = np.zeros((nlags*nlags, 1))
P0guess[:, :] = np.linalg.solve(tmp, Qvec)
P0guess = np.reshape(P0guess, (nlags, nlags))

# Delete temprary variables
del Qvec, tmp


#%% EM Algorithm to Compute Model Parameters and WARN Factor

Fhat, Phat, Ahat, Qhat, Lhat, Rhat, Loglikes, ell = \
    em_func(F0guess, P0guess, Zstd, Lguess, Rguess, Aguess, Qguess,
            W, nloop, thresh)

# Check if the EM algoithm finished before reaching max iterations
if ell == nloop-1:
    print('No solution found with EM algorithm. '
          + 'Maximum number of iterations reached.')
else:
    print('Solution to EM algorithm found on loop ' + str(ell))


#%% Table A3

# Storage array
table_a3 = np.zeros((nvars, 4))

# Factor loadings
table_a3[:, 0] = Lhat[:, 0]

# Number of observations
for jj in range(nvars):
    table_a3[jj, 1] = np.sum(W[jj, jj, :])

# Average employment
table_a3[:, 2] = df_state.loc[:, 'emp'].to_numpy()

# Average manufacturing ratio
table_a3[:, 3] = df_state.loc[:, 'man_frac'].to_numpy()


#%% Table A4

# Storage array
table_a4 = np.zeros((2, 3))

# Compute correlations with SC
temp = np.corrcoef(table_a3[:, 0], table_a3[:, 1])
table_a4[0, 0] = temp[1, 0]

temp = np.corrcoef(table_a3[:, 0], np.log(table_a3[:, 2]))
table_a4[0, 1] = temp[1, 0]

temp = np.corrcoef(table_a3[:, 0], table_a3[:, 3])
table_a4[0, 2] = temp[1, 0]

# Compute correlations without SC
no_sc = np.zeros((nvars-1, 4))
no_sc[:, :] = np.delete(table_a3, 22, 0)

temp = np.corrcoef(no_sc[:, 0], no_sc[:, 1])
table_a4[1, 0] = temp[1, 0]

temp = np.corrcoef(no_sc[:, 0], np.log(no_sc[:, 2]))
table_a4[1, 1] = temp[1, 0]

temp = np.corrcoef(no_sc[:, 0], no_sc[:, 3])
table_a4[1, 2] = temp[1, 0]


#%% Regressions with SC

# RHS array
rhs = np.zeros((nvars, 4))
rhs[:, 0] = 1
rhs[:, 1] = table_a3[:, 1]
rhs[:, 2] = np.log(table_a3[:, 2])
rhs[:, 3] = table_a3[:, 3]

# LHS array
lhs = np.zeros((nvars, 1))
lhs[:, 0] = table_a3[:, 0]

# Ordinary least squares
Q = np.zeros((4, 4))
Q[:, :] = nvars*np.linalg.inv(rhs.T @ rhs)

phi = np.zeros((4, 1))
phi[:, :] = Q @ (rhs.T @ lhs)/nvars

# Regression errors
ee = np.zeros((nvars, 1))
ee[:, :] = lhs - rhs @ phi

# Standard errors
z = np.zeros((nvars, 4))
z[:, :] = np.multiply(rhs, ee @ np.ones((1, 4)))

omega = np.zeros((4, 4))
omega[:, :] = (z.T @ z)/nvars

V = np.zeros((4, 4))
V[:, :] = Q @ omega @ Q.T

se = np.zeros((4, 1))
se[:, 0] = np.sqrt(np.diag(V)/nvars)

# t-statistics
t_stats = np.divide(phi[:, 0], se[:, 0])


#%% First Row of Table A5

# Storage array
table_a5 = np.zeros((4, 4))

# Results with SC
table_a5[0, :] = np.round(phi[:, 0], 4)
table_a5[1, :] = np.round(se[:, 0], 4)


#%% Regressions without SC

# RHS array
rhs = np.zeros((nvars-1, 4))
rhs[:, 0] = 1
rhs[:, 1] = no_sc[:, 1]
rhs[:, 2] = np.log(no_sc[:, 2])
rhs[:, 3] = no_sc[:, 3]

# LHS array
lhs = np.zeros((nvars-1, 1))
lhs[:, 0] = no_sc[:, 0]

# Ordinary least squares
Q = np.zeros((4, 4))
Q[:, :] = (nvars - 1)*np.linalg.inv(rhs.T @ rhs)

phi = np.zeros((4, 1))
phi[:, :] = Q @ (rhs.T @ lhs)/(nvars - 1)

# Regression errors
ee = np.zeros((nvars-1, 1))
ee[:, :] = lhs - rhs @ phi

# Standard errors
z = np.zeros((nvars-1, 4))
z[:, :] = np.multiply(rhs, ee @ np.ones((1, 4)))

omega = np.zeros((4, 4))
omega[:, :] = (z.T @ z)/(nvars - 1)

V = np.zeros((4, 4))
V[:, :] = Q @ omega @ Q.T

se = np.zeros((4, 1))
se[:, 0] = np.sqrt(np.diag(V)/(nvars - 1))

# t-statistics
t_stats = np.divide(phi[:, 0], se[:, 0])


#%% Second Row of Table A5

# Results with SC
table_a5[2, :] = np.round(phi[:, 0], 4)
table_a5[3, :] = np.round(se[:, 0], 4)


#%% Disply the Run Time in Minutes

print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


