# -*- coding: utf-8 -*-
"""
This program computes the Bai and Ng (2002) criteria for number of factors for
Appendix B.5 of "Advance Layoff Notices and Aggregate Job Loss" 
by Pawel M. Krolikowski and Kurt G. Lunsford.

The sample used in this program is July 1996 through December 2019.

This program calls the following files:
    seasonal_module.py : module with functions to seasonally adjust the data
    WARNData_NSA_20211230.xlsx : Excel file with the state-level WARN data

This program prints the number of factors.
"""


#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Import Pandas and NumPy
import pandas as pd
import numpy as np

# Functions for seasonal adjustment
from modules.seasonal_module import sa_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the program

# String with the date of the data to call
str_date = '20211230'   # Format: 20200115 corresponds to January 15, 2020

# String to choose first date for WARN data
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# String to choose last date of WARN data used for Bai-Ng ICs
str_last = '20191231'   # Format: 20191231 corresponds to December 31, 2019

# Parameters of the EM algorithm for seasonal adjustment
nloop = 20000       # maximum number of iterations in the EM algrorithm
thresh_sa = 1e-3    # threshold for convergence

# Maximum number of factor for Bai & Ng computations
Kmax = 5


#%% Load the WARN Data

# Dataframe for number of workers affected by WARN notices by month and state
str_file_name = 'WARNData_NSA_' + str_date + '.xlsx'
path_warn_data = os.path.join(path_dir, 'data_input', str_file_name)
df_warn_all = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)

# Drop small states or states with too many observations with a value of 0
df_warn = df_warn_all.drop(columns=['AK', 'ID', 'KY', 'ND', 'NM',
                                    'OK', 'RI', 'UT', 'WV'])

# Adjust the end of the sample
df_warn = df_warn.loc[:str_last, :]

# Number of states in the sample
nvars = len(df_warn.columns)


#%% Loop to Seasonally Adjust the Data

# Dataframe to store seasonally adjusted data
df_warn_sa = pd.DataFrame(index=df_warn.index, columns=df_warn.columns)

# Loop for seasonal adjustment
for loop in range(nvars):
    # State name
    state_name = df_warn.columns[loop]
    
    # Dates with observed data
    state_index = df_warn.loc[:, state_name].dropna().index
    
    # Number of observations excluding NaNs
    nsample = len(state_index)
    
    # Convert the data to NumPy while dropping NaNs
    X_nsa = np.zeros((nsample, 1))
    X_nsa[:, 0] = df_warn.loc[:, state_name].dropna().to_numpy()
    
    # Seasonally adjust the data
    data_sa = sa_func(X_nsa, nloop, thresh_sa, state_name)
    
    # Store seasonally adjusted data in dataframe
    df_warn_sa.loc[state_index, state_name] = data_sa[:, 0]
    
    # Delete temprary variables
    del state_name, state_index, nsample, data_sa


#%% Adjust the Data for Zero Values

# Number of observations
nsample = len(df_warn_sa.loc[str_warn:, :].index)

# Convert the data to numpy
X = np.zeros((nsample, nvars))
X[:, :] = df_warn_sa.loc[str_warn:, :].to_numpy()

# Set zero values to NaN
for jj in range(nsample):
    for ii in range(nvars):
        if X[jj, ii] < 1:
            X[jj, ii] = np.nan

# Delete temprary variables
del jj


#%% Function for Bai & Ng Criteria

def bai_ng_func(x_in, k_in):
    # Dimensions of the data
    TT, NN = x_in.shape
    
    # Eigenvalues and eigenvectors
    S = np.zeros((NN, NN))
    S[:, :] = x_in.T @ x_in         # inner product of data
    D, P = np.linalg.eig(S)         # eigenvalues and eigenvectors
    
    # Storage arrays
    Dnew = np.zeros(NN)
    Dnew[:] = D
    
    Pnew = np.zeros((NN, NN))
    Pnew[:, :] = P
    
    Plarge = np.zeros((NN, k_in))
    
    # Collect eigenvectors associated with largest eigenvalues
    for iii in range(k_in):
        # Largest eigenvalues and corresponding eigenvector
        ind = np.argmax(Dnew)           # index of largest eigenvalue
        Plarge[:, iii] = Pnew[:, ind]   # select the corresponding eigenvector
        
        # Delete the largest eigenvalue and corresponding eigenvector
        Dnew = np.delete(Dnew, ind)
        Pnew = np.delete(Pnew, ind, 1)
    
    # Compute the loadings and the factors
    lam = np.zeros((NN, k_in))
    lam[:, :] = np.sqrt(NN)*Plarge
    
    fact = np.zeros((TT, k_in))
    fact[:, :] = (x_in @ lam)/NN
    
    # Idiosyncratic components
    ee = np.zeros((TT, NN))
    ee[:, :] = x_in - fact @ lam.T
    
    # Compute the sum of squared errors
    sse = 0
    for iii in range(NN):
        sse = sse + np.inner(ee[:, iii], ee[:, iii])
    
    # Divide by N and T
    vv = sse/(NN*TT)
    
    # Information criteria
    penalty1 = ((NN + TT)/(NN*TT))*np.log((NN*TT)/(NN + TT))
    ic1_out = np.log(vv) + k_in*penalty1
    
    penalty2 = ((NN + TT)/(NN*TT))*np.log(NN)
    ic2_out = np.log(vv) + k_in*penalty2
    
    penalty3 = np.log(NN)/NN
    ic3_out = np.log(vv) + k_in*penalty3
    
    return ic1_out, ic2_out, ic3_out


#%% Compute the Bai and Ng Information Criteria

# Take logs of the data
X = np.log(X)

# Averages of the data
avg = np.zeros((nvars, 1))
avg[:, 0] = np.nanmean(X, axis=0)

# Variances of the data
var = np.zeros((nvars, 1))
var[:, 0] = np.nanvar(X, axis=0)

# Standardize the data: de-mean, unit variance, set NaNs to zero
Xstd = np.zeros((nsample, nvars))
for jj in range(nvars):
    mask = (np.isnan(X[:, jj]) == 0)
    Xstd[mask, jj] = (X[mask, jj] - avg[jj, 0])/np.sqrt(var[jj, 0])

# Loop to compute information criteria
ic1 = np.zeros(Kmax)
ic2 = np.zeros(Kmax)
ic3 = np.zeros(Kmax)
for jj in range(Kmax):
    # Number of factors
    kk = jj + 1
    
    # Information criteria
    ic1_temp, ic2_temp, ic3_temp = bai_ng_func(Xstd, kk)
    ic1[jj] = ic1_temp
    ic2[jj] = ic2_temp
    ic3[jj] = ic3_temp

# Compute the number of factors implied by the information criteria
k1 = np.argmin(ic1) + 1
k2 = np.argmin(ic2) + 1
k3 = np.argmin(ic3) + 1

# Print the results
print(' ')
print("Number of factors according to Bai & Ng's first information criteria "
      + "is " + str(k1) + ".")
print(' ')
print("Number of factors according to Bai & Ng's second information criteria "
      + "is " + str(k2) + ".")
print(' ')
print("Number of factors according to Bai & Ng's third information criteria "
      + "is " + str(k3) + ".")
print(' ')


#%% Disply the Run Time in Minutes

print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


