# -*- coding: utf-8 -*-
"""
This program produces Figure A.6 in the Appendix of "Advance Layoff Notices
and Aggregate Job Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    seasonal_module.py : module with functions to seasonally adjust the data
    WARNData_NSA_20211230.xlsx : Excel file with the state-level WARN data
    nber_recessions.xlsx : Excel file with NBER recession indicators
    WARNFactors_20211230.xlsx : the output of warn_dfm_program.py in
                                WARN_RepFiles\section_4

This program prints correlations reported in Appendix B.6.

This program also saves
    WARNhat_and_panel.pdf : Figure A.6 in the appendix
"""

#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Import Pandas, NumPy and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# Functions for seasonal adjustment
from modules.seasonal_module import sa_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the program

# String with the date of the data to call
str_date = '20211230'   # Format: 20200115 corresponds to January 15, 2020

# String to choose first date for balanced panel
str_first = '20060101'   # Format: 20060101 corresponds to January 1, 2006

# Parameters of the EM algorithm for seasonal adjustment
nloop = 20000       # maximum number of iterations in the EM algrorithm
thresh_sa = 1e-3    # threshold for convergence


#%% Load the WARN Data

# Dataframe for number of workers affected by WARN notices by month and state
str_file_name = 'WARNData_NSA_' + str_date + '.xlsx'
path_warn_data = os.path.join(path_dir, 'data_input', str_file_name)
df_warn_all = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)

# Drop small states or states with too many observations with a value of 0
df_warn = df_warn_all.drop(columns=['AK', 'ID', 'KY', 'ND', 'NM',
                                    'OK', 'RI', 'UT', 'WV'])

# Number of states in the sample
nvars = len(df_warn.columns)

# Load the WARN factor and WARN_hat
path_warn_data = os.path.join(path_dir, 'section_4', 'WARNFactors_20211230.xlsx')
df_factors = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)


#%% Collect States in the Balanced Panel

# Initial dataframe
df_balanced = df_warn.loc[str_first:, :]

# Loop to drop states not in the balanced panel
for loop in range(nvars):
    # Column name
    str_col = df_warn.columns[loop]
    
    # Index of first month
    str_date = df_balanced.index[0]
    
    # Drop state if needed
    if np.isnan(df_balanced.loc[str_date, str_col]) == True:
        df_balanced = df_balanced.drop(columns=[str_col])

# Number of states in the sample
nvars = len(df_balanced.columns)


#%% Loop to Seasonally Adjust the Data

# Dataframe to store seasonally adjusted data
df_balanced_sa = pd.DataFrame(index=df_warn.index, columns=df_balanced.columns)

# Loop for seasonal adjustment
for loop in range(nvars):
    # State name
    state_name = df_balanced.columns[loop]
    
    # Dates with observed data
    state_index = df_warn.loc[:, state_name].dropna().index
    
    # Number of observations excluding NaNs
    nsample = len(state_index)
    
    # Convert the data to NumPy while dropping NaNs
    X_nsa = np.zeros((nsample, 1))
    X_nsa[:, 0] = df_warn.loc[:, state_name].dropna().to_numpy()
    
    # Seasonally adjust the data
    data_sa = sa_func(X_nsa, nloop, thresh_sa, state_name)
    
    # Store seasonally adjusted data in dataframe
    df_balanced_sa.loc[state_index, state_name] = data_sa[:, 0]
    
    # Delete temprary variables
    del state_name, state_index, nsample, data_sa

# Truncate the seasonally adjusted sample
df_balanced_sa = df_balanced_sa.loc[str_first:, :]

# The time sample
sample = df_balanced_sa.index

# The time sample size
nsample = len(sample)


#%% Compute the Balanced Panel

# The sum over the balanced panel
warn_balanced = np.zeros(nsample)
for jj in range(nvars):
    state_name = df_balanced_sa.columns[jj]
    
    temp = np.zeros(nsample)
    temp[:] = df_balanced_sa.loc[:, state_name].to_numpy()
    
    warn_balanced[np.isnan(temp)==False] = warn_balanced[np.isnan(temp)==False] + temp[np.isnan(temp)==False]


#%% Load the NBER Recession Indicators for the Figures

path_rec_data = os.path.join(path_dir, 'data_input', 'nber_recessions.xlsx')
df_recession = pd.read_excel(path_rec_data, 'Sheet1', index_col=0)
df_recession = df_recession.loc[str_first:, :]


#%% Plot the Results

# Recession indicator
rec = np.zeros(nsample)
rec[:] = 10000*df_recession.loc[sample, 'USREC']

# WARN_hat
warn_hat = np.zeros(nsample)
warn_hat[:] = df_factors.loc[sample, 'WARN_hat']

# Figure and axes instances
fig, ax = plt.subplots(2, 1, figsize=(7, 7))

# Plot the WARN factor
ax[0].fill_between(sample, rec, -rec, color='black', alpha=0.15, linewidth=0)
ax[0].plot(sample, np.zeros(nsample), color='black', linewidth=1)
line1, = ax[0].plot(sample, warn_hat/1000, color='black', linewidth=3)
line2, = ax[0].plot(sample, warn_balanced/1000, color='dimgray', linewidth=2)
ax[0].set_xlim(sample[0], sample[-1])
ax[0].set_ylim(0, 500)
ax[0].set_ylabel('Number of workers (Thousands, SA)', fontsize=10)
ax[0].set_title('January 2006 through November 2021', fontsize=12)
ax[0].legend([line1, line2], ['$\widehat{WARN}_t$', 'WARN layoffs from balanced panel'])

# Plot the WARN_hat
num = 14*12
ax[1].fill_between(sample[0:num], rec[0:num], -rec[0:num],
                   color='black', alpha=0.15, linewidth=0)
ax[1].plot(sample[0:num], np.zeros(num), color='black', linewidth=1)
line3, = ax[1].plot(sample[0:num], warn_hat[0:num]/1000, color='black',linewidth=3)
line4, = ax[1].plot(sample[0:num], warn_balanced[0:num]/1000, color='dimgray', linewidth=1.5)
ax[1].set_xlim(sample[0], sample[num-1])
ax[1].set_ylim(0, 90)
ax[1].set_ylabel('Number of workers (Thousands, SA)', fontsize=10)
ax[1].set_title('January 2006 through December 2019', fontsize=12)
ax[1].legend([line3, line4], ['$\widehat{WARN}_t$', 'WARN layoffs from balanced panel'])

fig.tight_layout()

# Save the figure
plt.savefig('WARNhat_and_panel.pdf', bbox_inches='tight')


#%% Compute and Print Correlations

temp = np.corrcoef(warn_balanced, warn_hat)
print('The correlation from Jan 2006 through Nov 2021 is '
      + str(np.round(temp[1, 0], 2)))
print(' ')

temp = np.corrcoef(warn_balanced[0:num], warn_hat[0:num])
print('The correlation from Jan 2006 through Dec 2019 is '
      + str(np.round(temp[1, 0], 2)))
print(' ')


#%% Disply the Run Time in Minutes

print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')

