# -*- coding: utf-8 -*-
"""
This program produces Figures A.8, A.9, and A.10 in "Advance Layoff Notices
and Aggregate Job Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    fcast_eval_module.py : module with functions to compute and evaluate forecasts
    warn_factors_rt.xlsx : Excel file with pseudo real-time WARN factors and
                           values of WARN hat from Equation (4)
    labor_rates_rt.xlsx : Excel file with real-time uneployment rates and 
                          labor market flow rates
    uspriv_alfred.xlsx : Excel file with real-time private employment levels
    icsa_fred.xlsx : Excel file holding monthly averages of initial
                     unemployment insurance claims

This program saves
    RMSPE_ssur_roll.pdf : Figure A.8 in the appendix
    RMSPE_varu_roll.pdf : Figure A.9 in the appendix
    RMSPE_vare_roll.pdf : Figure A.10 in the appendix
"""

#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Pandas, NumPy and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# Functions for making and evaluating forecasts
from modules.fcast_eval_module import fcast_var_func, fcast_bn_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the program

# First date that WARN factor is available
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# First date for forecast evaluation
str_first = '20080101'  # Format: 20080101 corresponds to January 1, 2008
                        # 20030101 is the earliest permissable choice

# Last date for forecast evaluation
str_last = '20191231'   # Format: 20191231 corresponds to December 31, 2019

# Number of variables in the VAR including WARN factor or WARN hat
nvars = 6

# Number of lags in the forecasting models
nlags = 2

# Longest available sample size for estimating the forecasting models
nsample = len(pd.date_range(start=str_warn, end=str_first, freq='MS')) - 2

# Maximum horizon to forecast
hmax = 24

# Rolling sample size
m = 30


#%% Load the Data

# Number of months to truncate
KK = len(pd.date_range(start='20030101', end=str_first, freq='MS')) - 1

# Change in warn factor
path_factors = os.path.join(path_dir, 'section_6_2', 'warn_factors_rt.xlsx')
df_factor = pd.read_excel(path_factors, 'factors', index_col=0)
df_factor = df_factor.diff()
df_factor = df_factor.loc[str_warn:, :]
df_factor = df_factor.iloc[:, KK:]

# Change in warn_hat
df_warn = pd.read_excel(path_factors, 'warn_hat', index_col=0)
df_warn = df_warn.diff()
df_warn = df_warn.loc[str_warn:, :]
df_warn = df_warn.iloc[:, KK:]

# Unemployment rate
path_labor = os.path.join(path_dir, 'section_6_2', 'labor_rates_rt.xlsx')
df_urate = pd.read_excel(path_labor, 'urate', index_col=0)
df_urate = df_urate.loc[str_warn:, :]
df_urate = df_urate.iloc[:, KK:]

# Change in log unemployment rate
df_uchange = np.log(pd.read_excel(path_labor, 'urate', index_col=0))
df_uchange = df_uchange.diff()
df_uchange = df_uchange.loc[str_warn:, :]
df_uchange = df_uchange.iloc[:, KK:]

# Log finding rate
df_grate = np.log(pd.read_excel(path_labor, 'grate', index_col=0))
df_grate = df_grate.loc[str_warn:, :]
df_grate = df_grate.iloc[:, KK:]

# Log separation rate
df_srate = np.log(pd.read_excel(path_labor, 'srate', index_col=0))
df_srate = df_srate.loc[str_warn:, :]
df_srate = df_srate.iloc[:, KK:]

# Level and change of private employment
path_emp_data = os.path.join(path_dir, 'data_input', 'uspriv_alfred.xlsx')
df_emp = pd.read_excel(path_emp_data, 'Sheet1', index_col=0)
df_demp = df_emp.diff()
df_emp = df_emp.loc[str_warn:, :]
df_demp = df_demp.loc[str_warn:, :]
df_demp = df_demp.iloc[:, KK:]

# Initial claims
path_claims_data = os.path.join(path_dir, 'data_input', 'icsa_alfred.xlsx')
df_claim = np.log(pd.read_excel(path_claims_data, 'Sheet2', index_col=0))
df_claim = df_claim.loc[str_warn:, :]
df_claim = df_claim.iloc[:, KK:]


#%% Compute the Forecasts

# Pseudo real-time dates that forecasts are produced for
fcast_index = pd.date_range(start=str_first, end=str_last, freq='MS')

# Number of forecasts to compute
nfcast = len(fcast_index)

# Storage arrays for finding rate forecasts
g_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
g_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
g_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for separation rate forecasts
s_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
s_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
s_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for unemployment rate forecasts
u_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
u_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
u_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat
u_fcast_bn = np.zeros((nfcast, hmax+1))     # forecasts from BN model without WARN
u_fcast_bnf = np.zeros((nfcast, hmax+1))    # forecasts from BN model with WARN factor
u_fcast_bnw = np.zeros((nfcast, hmax+1))    # forecasts from BN model with WARN_hat

# Storage arrays for employment change forecasts
e_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
e_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
e_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Loop to compute the forecasts
for loop in range(nfcast):
    # Number of observations in this loop
    nobs = len(df_uchange.iloc[:, loop].dropna())
    
    # Most recent value of unemployment rate available to forecaster
    u_init = np.zeros(1)
    u_init[0] = df_urate.iloc[nobs-1, loop]
    
    # Variables for VAR forecasts without WARN
    var_data = np.zeros((nsample, nvars-1))
    var_data[:, 0] = df_grate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 1] = df_srate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 2] = df_uchange.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 3] = df_claim.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 4] = df_demp.iloc[nobs-nsample:nobs, loop].to_numpy()
    
    # Variables for VAR forecasts with WARN factor
    varf_data = np.zeros((nsample, nvars))
    varf_data[:, 0:nvars-1] = var_data[:, :]
    varf_data[:, nvars-1] = df_factor.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Variables for VAR forecasts with WARN_hat
    varw_data = np.zeros((nsample, nvars))
    varw_data[:, 0:nvars-1] = var_data[:, :]
    varw_data[:, nvars-1] = df_warn.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Forecasts of finding and separation rates from the VAR without WARN
    var_fcast = np.zeros((hmax+1, nvars-1))
    var_fcast[:, :] = fcast_var_func(var_data, hmax, nlags)
    g_fcast_var[loop, :] = np.exp(var_fcast[:, 0])
    s_fcast_var[loop, :] = np.exp(var_fcast[:, 1])
    e_fcast_var[loop, :] = var_fcast[:, 4]
    
    # Forecasts of unemployment rate from the VAR without WARN
    temp = np.zeros(hmax+1)
    temp[:] = np.cumsum(var_fcast[:, 2])
    u_fcast_var[loop, :] = np.exp(temp + np.log(u_init[0]))
    
    # Forecasts of finding and separation rates from the VAR with WARN factor
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varf_data, hmax, nlags)
    g_fcast_varf[loop, :] = np.exp(var_fcast[:, 0])
    s_fcast_varf[loop, :] = np.exp(var_fcast[:, 1])
    e_fcast_varf[loop, :] = var_fcast[:, 4]
    
    # Forecasts of unemployment rate from the VAR with WARN factor
    temp = np.zeros(hmax+1)
    temp[:] = np.cumsum(var_fcast[:, 2])
    u_fcast_varf[loop, :] = np.exp(temp + np.log(u_init[0]))
    
    # Forecasts of finding and separation rates from the VAR with WARN_hat
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varw_data, hmax, nlags)
    g_fcast_varw[loop, :] = np.exp(var_fcast[:, 0])
    s_fcast_varw[loop, :] = np.exp(var_fcast[:, 1])
    e_fcast_varw[loop, :] = var_fcast[:, 4]
    
    # Forecasts of unemployment rate from the VAR with WARN_hat
    temp = np.zeros(hmax+1)
    temp[:] = np.cumsum(var_fcast[:, 2])
    u_fcast_varw[loop, :] = np.exp(temp + np.log(u_init[0]))
    
    # Forecasts unemployment rate from BN model without WARN
    u_fcast_bn[loop, :] = fcast_bn_func(g_fcast_var[loop, :],
                                        s_fcast_var[loop, :], u_init[0], hmax)
    
    # Forecasts unemployment rate from BN model with WARN factor
    u_fcast_bnf[loop, :] = fcast_bn_func(g_fcast_varf[loop, :],
                                         s_fcast_varf[loop, :], u_init[0], hmax)
    
    # Forecasts unemployment rate from BN model with WARN_hat
    u_fcast_bnw[loop, :] = fcast_bn_func(g_fcast_varw[loop, :],
                                         s_fcast_varw[loop, :], u_init[0], hmax)
    
    print(loop)


#%% Compute Forecast Errors

# The unemployment rate outcomes
u_actual = df_urate.loc[str_first:str_last, :]
u_actual = u_actual.iloc[:, -1]

# The change in employment outcomes
e_actual = df_demp.loc[str_first:str_last, :]
e_actual = e_actual.iloc[:, -1]

# Storage arrays for unemployment rate forecast errors
du_var = np.zeros((nfcast, hmax+1))     # forecasts from VAR without WARN
du_varf = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN factor
du_varw = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN_hat
du_bn = np.zeros((nfcast, hmax+1))      # forecasts from BN model without WARN
du_bnf = np.zeros((nfcast, hmax+1))     # forecasts from BN model with WARN factor
du_bnw = np.zeros((nfcast, hmax+1))     # forecasts from BN model with WARN_hat

# Storage arrays for employment change forecast errors
de_var = np.zeros((nfcast, hmax+1))     # forecasts from VAR without WARN
de_varf = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN factor
de_varw = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN_hat

# Loop to compute forecast errors and RMSPEs
for jj in range(hmax+1):
    # Forecast errors for unemployment rates
    du_var[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                 - u_fcast_var[0:nfcast-jj, jj])
    du_varf[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                  - u_fcast_varf[0:nfcast-jj, jj])
    du_varw[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                  - u_fcast_varw[0:nfcast-jj, jj])
    du_bn[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                - u_fcast_bn[0:nfcast-jj, jj])
    du_bnf[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                 - u_fcast_bnf[0:nfcast-jj, jj])
    du_bnw[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                 - u_fcast_bnw[0:nfcast-jj, jj])
    
    # Forecast errors for employment changes
    de_var[jj:nfcast, jj] = e_actual[jj:nfcast] - e_fcast_var[0:nfcast-jj, jj]
    de_varf[jj:nfcast, jj] = e_actual[jj:nfcast] - e_fcast_varf[0:nfcast-jj, jj]
    de_varw[jj:nfcast, jj] = e_actual[jj:nfcast] - e_fcast_varw[0:nfcast-jj, jj]


#%% Compute Rolling RMSPEs

# Storage arrays for RMSPEs
rmspe_u_var = np.zeros((nfcast-m, hmax+1))
rmspe_u_varf = np.zeros((nfcast-m, hmax+1))
rmspe_u_varw = np.zeros((nfcast-m, hmax+1))
rmspe_u_bn = np.zeros((nfcast-m, hmax+1))
rmspe_u_bnf = np.zeros((nfcast-m, hmax+1))
rmspe_u_bnw = np.zeros((nfcast-m, hmax+1))

rmspe_e_var = np.zeros((nfcast-m, hmax+1))
rmspe_e_varf = np.zeros((nfcast-m, hmax+1))
rmspe_e_varw = np.zeros((nfcast-m, hmax+1))

# Compute RMSPEs
for ii in range(hmax+1):
    for jj in range(nfcast-m-ii):
        rmspe_u_var[jj+ii, ii] = np.sqrt(np.mean(np.power(du_var[jj+ii:jj+ii+m, ii], 2)))
        rmspe_u_varf[jj+ii, ii] = np.sqrt(np.mean(np.power(du_varf[jj+ii:jj+ii+m, ii], 2)))
        rmspe_u_varw[jj+ii, ii] = np.sqrt(np.mean(np.power(du_varw[jj+ii:jj+ii+m, ii], 2)))
        
        rmspe_u_bn[jj+ii, ii] = np.sqrt(np.mean(np.power(du_bn[jj+ii:jj+ii+m, ii], 2)))
        rmspe_u_bnf[jj+ii, ii] = np.sqrt(np.mean(np.power(du_bnf[jj+ii:jj+ii+m, ii], 2)))
        rmspe_u_bnw[jj+ii, ii] = np.sqrt(np.mean(np.power(du_bnw[jj+ii:jj+ii+m, ii], 2)))
        
        rmspe_e_var[jj+ii, ii] = np.sqrt(np.mean(np.power(de_var[jj+ii:jj+ii+m, ii], 2)))
        rmspe_e_varf[jj+ii, ii] = np.sqrt(np.mean(np.power(de_varf[jj+ii:jj+ii+m, ii], 2)))
        rmspe_e_varw[jj+ii, ii] = np.sqrt(np.mean(np.power(de_varw[jj+ii:jj+ii+m, ii], 2)))


#%% Plot Unemployment RMSPEs with BN Model

# Sample period to plot
sample = pd.date_range(start=str_first, end=str_last, freq='MS')

# Truncate the sample
indexm = int(np.round(m/2))
sample = sample[indexm-1:nfcast-indexm-1]

# Figure and axes instances
fig, ax = plt.subplots(3, 2, figsize=(10, 10))

h = 6

series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = rmspe_u_bn[h:nfcast-m, h]

seriesw = np.zeros(nfcast - m)
seriesw[:] = np.nan
seriesw[h:nfcast-m] = rmspe_u_bnw[h:nfcast-m, h]

ax[0, 0].plot(sample, series, linewidth=3, color='black')
ax[0, 0].plot(sample, seriesw, linewidth=3, color='grey')
ax[0, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[0, 0].set_xlim(sample[0], sample[-1])
ax[0, 0].set_ylim(0, 1.5)
ax[0, 0].set_yticks([0, 0.5, 1.0, 1.5])
ax[0, 0].set_ylabel('RMSPE (percentage points)')
ax[0, 0].set_title('Unemployment forecasts with SSUR models, h = ' + str(h), fontsize=10)
ax[0, 0].legend(['SSUR-baseline', 'SSUR-w'], loc='upper right')

diff = np.zeros(nfcast - m)
diff[:] = seriesw - series

ax[0, 1].plot(sample, diff, linewidth=3, color='black')
ax[0, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[0, 1].set_xlim(sample[0], sample[-1])
ax[0, 1].set_ylim(-0.1, 0)
ax[0, 1].set_yticks([-0.1, -0.05, 0])
ax[0, 1].set_ylabel('RMSPE (percentage points)')
ax[0, 1].set_title('Difference between RMSPEs with SSUR models, h = ' + str(h), fontsize=10)

h = 12

series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = rmspe_u_bn[h:nfcast-m, h]

seriesw = np.zeros(nfcast - m)
seriesw[:] = np.nan
seriesw[h:nfcast-m] = rmspe_u_bnw[h:nfcast-m, h]

ax[1, 0].plot(sample, series, linewidth=3, color='black')
ax[1, 0].plot(sample, seriesw, linewidth=3, color='grey')
ax[1, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[1, 0].set_xlim(sample[0], sample[-1])
ax[1, 0].set_ylim(0, 3)
ax[1, 0].set_yticks([0, 1.0, 2.0, 3.0])
ax[1, 0].set_ylabel('RMSPE (percentage points)')
ax[1, 0].set_title('Unemployment forecasts with SSUR models, h = ' + str(h), fontsize=10)
ax[1, 0].legend(['SSUR-baseline', 'SSUR-w'], loc='upper right')

diff = np.zeros(nfcast - m)
diff[:] = seriesw - series

ax[1, 1].plot(sample, diff, linewidth=3, color='black')
ax[1, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[1, 1].set_xlim(sample[0], sample[-1])
ax[1, 1].set_ylim(-0.3, 0)
ax[1, 1].set_yticks([-0.3, -0.2, -0.1, 0])
ax[1, 1].set_ylabel('RMSPE (percentage points)')
ax[1, 1].set_title('Difference between RMSPEs with SSUR models, h = ' + str(h), fontsize=10)

h = 18

series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = rmspe_u_bn[h:nfcast-m, h]

seriesw = np.zeros(nfcast - m)
seriesw[:] = np.nan
seriesw[h:nfcast-m] = rmspe_u_bnw[h:nfcast-m, h]

ax[2, 0].plot(sample, series, linewidth=3, color='black')
ax[2, 0].plot(sample, seriesw, linewidth=3, color='grey')
ax[2, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[2, 0].set_xlim(sample[0], sample[-1])
ax[2, 0].set_ylim(0, 4)
ax[2, 0].set_yticks([0, 1, 2, 3, 4])
ax[2, 0].set_ylabel('RMSPE (percentage points)')
ax[2, 0].set_title('Unemployment forecasts with SSUR models, h = ' + str(h), fontsize=10)
ax[2, 0].legend(['SSUR-baseline', 'SSUR-w'], loc='upper right')

diff = np.zeros(nfcast - m)
diff[:] = seriesw - series

ax[2, 1].plot(sample, diff, linewidth=3, color='black')
ax[2, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[2, 1].set_xlim(sample[0], sample[-1])
ax[2, 1].set_ylim(-0.5, 0)
ax[2, 1].set_yticks([-0.4, -0.2, 0])
ax[2, 1].set_ylabel('RMSPE (percentage points)')
ax[2, 1].set_title('Difference between RMSPEs with SSUR models, h = ' + str(h), fontsize=10)

fig.tight_layout()
plt.savefig('RMSPE_ssur_roll.pdf', bbox_inches='tight')


#%% Plot Unemployment RMSPEs with VAR Model

# Figure and axes instances
fig, ax = plt.subplots(3, 2, figsize=(10, 10))

h = 6

series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = rmspe_u_var[h:nfcast-m, h]

seriesw = np.zeros(nfcast - m)
seriesw[:] = np.nan
seriesw[h:nfcast-m] = rmspe_u_varw[h:nfcast-m, h]

ax[0, 0].plot(sample, series, linewidth=3, color='black')
ax[0, 0].plot(sample, seriesw, linewidth=3, color='grey')
ax[0, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[0, 0].set_xlim(sample[0], sample[-1])
ax[0, 0].set_ylim(0, 1.5)
ax[0, 0].set_yticks([0, 0.5, 1.0, 1.5])
ax[0, 0].set_ylabel('RMSPE (percentage points)')
ax[0, 0].set_title('Unemployment forecasts with VAR models, h = ' + str(h), fontsize=10)
ax[0, 0].legend(['VAR-5', 'VAR-w'], loc='upper right')

diff = np.zeros(nfcast - m)
diff[:] = seriesw - series

ax[0, 1].plot(sample, diff, linewidth=3, color='black')
ax[0, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[0, 1].set_xlim(sample[0], sample[-1])
ax[0, 1].set_ylim(-0.15, 0)
ax[0, 1].set_yticks([-0.15, -0.1, -0.05, 0])
ax[0, 1].set_ylabel('RMSPE (percentage points)')
ax[0, 1].set_title('Difference between RMSPEs with VAR models, h = ' + str(h), fontsize=10)

h = 12

series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = rmspe_u_var[h:nfcast-m, h]

seriesw = np.zeros(nfcast - m)
seriesw[:] = np.nan
seriesw[h:nfcast-m] = rmspe_u_varw[h:nfcast-m, h]

ax[1, 0].plot(sample, series, linewidth=3, color='black')
ax[1, 0].plot(sample, seriesw, linewidth=3, color='grey')
ax[1, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[1, 0].set_xlim(sample[0], sample[-1])
ax[1, 0].set_ylim(0, 3)
ax[1, 0].set_yticks([0, 1.0, 2.0, 3.0])
ax[1, 0].set_ylabel('RMSPE (percentage points)')
ax[1, 0].set_title('Unemployment forecasts with VAR models, h = ' + str(h), fontsize=10)
ax[1, 0].legend(['VAR-5', 'VAR-w'], loc='upper right')

diff = np.zeros(nfcast - m)
diff[:] = seriesw - series

ax[1, 1].plot(sample, diff, linewidth=3, color='black')
ax[1, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[1, 1].set_xlim(sample[0], sample[-1])
ax[1, 1].set_ylim(-0.36, 0)
ax[1, 1].set_yticks([-0.3, -0.2, -0.1, 0])
ax[1, 1].set_ylabel('RMSPE (percentage points)')
ax[1, 1].set_title('Difference between RMSPEs with VAR models, h = ' + str(h), fontsize=10)

h = 18

series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = rmspe_u_var[h:nfcast-m, h]

seriesw = np.zeros(nfcast - m)
seriesw[:] = np.nan
seriesw[h:nfcast-m] = rmspe_u_varw[h:nfcast-m, h]

ax[2, 0].plot(sample, series, linewidth=3, color='black')
ax[2, 0].plot(sample, seriesw, linewidth=3, color='grey')
ax[2, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[2, 0].set_xlim(sample[0], sample[-1])
ax[2, 0].set_ylim(0, 4)
ax[2, 0].set_yticks([0, 1, 2, 3, 4])
ax[2, 0].set_ylabel('RMSPE (percentage points)')
ax[2, 0].set_title('Unemployment forecasts with VAR models, h = ' + str(h), fontsize=10)
ax[2, 0].legend(['VAR-5', 'VAR-w'], loc='upper right')

diff = np.zeros(nfcast - m)
diff[:] = seriesw - series

ax[2, 1].plot(sample, diff, linewidth=3, color='black')
ax[2, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[2, 1].set_xlim(sample[0], sample[-1])
ax[2, 1].set_ylim(-0.55, 0)
ax[2, 1].set_yticks([-0.4, -0.2, 0])
ax[2, 1].set_ylabel('RMSPE (percentage points)')
ax[2, 1].set_title('Difference between RMSPEs with VAR models, h = ' + str(h), fontsize=10)

fig.tight_layout()
plt.savefig('RMSPE_varu_roll.pdf', bbox_inches='tight')


#%% Plot Unemployment RMSPEs with VAR Model

# Figure and axes instances
fig, ax = plt.subplots(3, 2, figsize=(10, 10))

h = 6

series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = rmspe_e_var[h:nfcast-m, h]

seriesw = np.zeros(nfcast - m)
seriesw[:] = np.nan
seriesw[h:nfcast-m] = rmspe_e_varw[h:nfcast-m, h]

ax[0, 0].plot(sample, series, linewidth=3, color='black')
ax[0, 0].plot(sample, seriesw, linewidth=3, color='grey')
ax[0, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[0, 0].set_xlim(sample[0], sample[-1])
ax[0, 0].set_ylim(0, 450)
ax[0, 0].set_yticks([0, 100, 200, 300, 400])
ax[0, 0].set_ylabel('RMSPE (workers in thousands)')
ax[0, 0].set_title('Employment forecasts with VAR models, h = ' + str(h), fontsize=10)
ax[0, 0].legend(['VAR-5', 'VAR-w'], loc='upper right')

diff = np.zeros(nfcast - m)
diff[:] = seriesw - series

ax[0, 1].plot(sample, diff, linewidth=3, color='black')
ax[0, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[0, 1].set_xlim(sample[0], sample[-1])
ax[0, 1].set_ylim(-33, 15)
ax[0, 1].set_yticks([-30, -15, 0, 15])
ax[0, 1].set_ylabel('RMSPE (workers in thousands)')
ax[0, 1].set_title('Difference between RMSPEs with VAR models, h = ' + str(h), fontsize=10)

h = 12

series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = rmspe_e_var[h:nfcast-m, h]

seriesw = np.zeros(nfcast - m)
seriesw[:] = np.nan
seriesw[h:nfcast-m] = rmspe_e_varw[h:nfcast-m, h]

ax[1, 0].plot(sample, series, linewidth=3, color='black')
ax[1, 0].plot(sample, seriesw, linewidth=3, color='grey')
ax[1, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[1, 0].set_xlim(sample[0], sample[-1])
ax[1, 0].set_ylim(0, 470)
ax[1, 0].set_yticks([0, 100, 200, 300, 400])
ax[1, 0].set_ylabel('RMSPE (workers in thousands)')
ax[1, 0].set_title('Employment forecasts with VAR models, h = ' + str(h), fontsize=10)
ax[1, 0].legend(['VAR-5', 'VAR-w'], loc='upper right')

diff = np.zeros(nfcast - m)
diff[:] = seriesw - series

ax[1, 1].plot(sample, diff, linewidth=3, color='black')
ax[1, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[1, 1].set_xlim(sample[0], sample[-1])
ax[1, 1].set_ylim(-60, 0)
ax[1, 1].set_yticks([-60, -40, -20, 0])
ax[1, 1].set_ylabel('RMSPE (workers in thousands)')
ax[1, 1].set_title('Difference between RMSPEs with VAR models, h = ' + str(h), fontsize=10)

h = 18

series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = rmspe_e_var[h:nfcast-m, h]

seriesw = np.zeros(nfcast - m)
seriesw[:] = np.nan
seriesw[h:nfcast-m] = rmspe_e_varw[h:nfcast-m, h]

ax[2, 0].plot(sample, series, linewidth=3, color='black')
ax[2, 0].plot(sample, seriesw, linewidth=3, color='grey')
ax[2, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[2, 0].set_xlim(sample[0], sample[-1])
ax[2, 0].set_ylim(0, 300)
ax[2, 0].set_yticks([0, 100, 200, 300])
ax[2, 0].set_ylabel('RMSPE (workers in thousands)')
ax[2, 0].set_title('Employment forecasts with VAR models, h = ' + str(h), fontsize=10)
ax[2, 0].legend(['VAR-5', 'VAR-w'], loc='upper right')

diff = np.zeros(nfcast - m)
diff[:] = seriesw - series

ax[2, 1].plot(sample, diff, linewidth=3, color='black')
ax[2, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[2, 1].set_xlim(sample[0], sample[-1])
ax[2, 1].set_ylim(-70, 0)
ax[2, 1].set_yticks([-60, -40, -20, 0])
ax[2, 1].set_ylabel('RMSPE (workers in thousands)')
ax[2, 1].set_title('Difference between RMSPEs with VAR models, h = ' + str(h), fontsize=10)

fig.tight_layout()
plt.savefig('RMSPE_vare_roll.pdf', bbox_inches='tight')


#%% Disply the Run Time

# This line display the run time in minutes
print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


