# -*- coding: utf-8 -*-
"""
This program produces Figure A.11 in "Advance Layoff Notices and Aggregate Job
Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    fcast_eval_module.py : module with functions to compute and evaluate forecasts
    warn_factors_rt.xlsx : Excel file with pseudo real-time WARN factors and
                           values of WARN hat from Equation (4)
    labor_rates_rt.xlsx : Excel file with real-time uneployment rates and 
                          labor market flow rates
    uspriv_alfred.xlsx : Excel file with real-time private employment levels
    icsa_fred.xlsx : Excel file holding monthly averages of initial
                     unemployment insurance claims

This program saves
    GiacominiRossiStats.pdf : Figure A.11 in the appendix
"""

#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Pandas, NumPy and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# Functions for making and evaluating forecasts
from modules.fcast_eval_module import fcast_var_func, fcast_bn_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the program

# First date that WARN factor is available
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# First date for forecast evaluation
str_first = '20080101'  # Format: 20080101 corresponds to January 1, 2008
                        # 20030101 is the earliest permissable choice

# Last date for forecast evaluation
str_last = '20191231'   # Format: 20191231 corresponds to December 31, 2019

# Number of variables in the VAR including WARN factor or WARN hat
nvars = 6

# Number of lags in the forecasting models
nlags = 2

# Longest available sample size for estimating the forecasting models
nsample = len(pd.date_range(start=str_warn, end=str_first, freq='MS')) - 2

# Maximum horizon to forecast
hmax = 24

# Rolling sample size
m = 30


#%% Load the Data

# Number of months to truncate
KK = len(pd.date_range(start='20030101', end=str_first, freq='MS')) - 1

# Change in warn factor
path_factors = os.path.join(path_dir, 'section_6_2', 'warn_factors_rt.xlsx')
df_factor = pd.read_excel(path_factors, 'factors', index_col=0)
df_factor = df_factor.diff()
df_factor = df_factor.loc[str_warn:, :]
df_factor = df_factor.iloc[:, KK:]

# Change in warn_hat
df_warn = pd.read_excel(path_factors, 'warn_hat', index_col=0)
df_warn = df_warn.diff()
df_warn = df_warn.loc[str_warn:, :]
df_warn = df_warn.iloc[:, KK:]

# Unemployment rate
path_labor = os.path.join(path_dir, 'section_6_2', 'labor_rates_rt.xlsx')
df_urate = pd.read_excel(path_labor, 'urate', index_col=0)
df_urate = df_urate.loc[str_warn:, :]
df_urate = df_urate.iloc[:, KK:]

# Change in log unemployment rate
df_uchange = np.log(pd.read_excel(path_labor, 'urate', index_col=0))
df_uchange = df_uchange.diff()
df_uchange = df_uchange.loc[str_warn:, :]
df_uchange = df_uchange.iloc[:, KK:]

# Log finding rate
df_grate = np.log(pd.read_excel(path_labor, 'grate', index_col=0))
df_grate = df_grate.loc[str_warn:, :]
df_grate = df_grate.iloc[:, KK:]

# Log separation rate
df_srate = np.log(pd.read_excel(path_labor, 'srate', index_col=0))
df_srate = df_srate.loc[str_warn:, :]
df_srate = df_srate.iloc[:, KK:]

# Level and change of private employment
path_emp_data = os.path.join(path_dir, 'data_input', 'uspriv_alfred.xlsx')
df_emp = pd.read_excel(path_emp_data, 'Sheet1', index_col=0)
df_demp = df_emp.diff()
df_emp = df_emp.loc[str_warn:, :]
df_demp = df_demp.loc[str_warn:, :]
df_demp = df_demp.iloc[:, KK:]

# initial claims
path_claims_data = os.path.join(path_dir, 'data_input', 'icsa_alfred.xlsx')
df_claim = np.log(pd.read_excel(path_claims_data, 'Sheet2', index_col=0))
df_claim = df_claim.loc[str_warn:, :]
df_claim = df_claim.iloc[:, KK:]


#%% Compute the Forecasts

# Pseudo real-time dates that forecasts are produced for
fcast_index = pd.date_range(start=str_first, end=str_last, freq='MS')

# Number of forecasts to compute
nfcast = len(fcast_index)

# Storage arrays for finding rate forecasts
g_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
g_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
g_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for separation rate forecasts
s_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
s_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
s_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for unemployment rate forecasts
u_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
u_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
u_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat
u_fcast_bn = np.zeros((nfcast, hmax+1))     # forecasts from BN model without WARN
u_fcast_bnf = np.zeros((nfcast, hmax+1))    # forecasts from BN model with WARN factor
u_fcast_bnw = np.zeros((nfcast, hmax+1))    # forecasts from BN model with WARN_hat

# Storage arrays for employment change forecasts
e_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
e_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
e_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Loop to compute the forecasts
for loop in range(nfcast):
    # Number of observations in this loop
    nobs = len(df_uchange.iloc[:, loop].dropna())
    
    # Most recent value of unemployment rate available to forecaster
    u_init = np.zeros(1)
    u_init[0] = df_urate.iloc[nobs-1, loop]
    
    # Variables for VAR forecasts without WARN
    var_data = np.zeros((nsample, nvars-1))
    var_data[:, 0] = df_grate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 1] = df_srate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 2] = df_uchange.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 3] = df_claim.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 4] = df_demp.iloc[nobs-nsample:nobs, loop].to_numpy()
    
    # Variables for VAR forecasts with WARN factor
    varf_data = np.zeros((nsample, nvars))
    varf_data[:, 0:nvars-1] = var_data[:, :]
    varf_data[:, nvars-1] = df_factor.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Variables for VAR forecasts with WARN_hat
    varw_data = np.zeros((nsample, nvars))
    varw_data[:, 0:nvars-1] = var_data[:, :]
    varw_data[:, nvars-1] = df_warn.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Forecasts of finding and separation rates from the VAR without WARN
    var_fcast = np.zeros((hmax+1, nvars-1))
    var_fcast[:, :] = fcast_var_func(var_data, hmax, nlags)
    g_fcast_var[loop, :] = np.exp(var_fcast[:, 0])
    s_fcast_var[loop, :] = np.exp(var_fcast[:, 1])
    e_fcast_var[loop, :] = var_fcast[:, 4]
    
    # Forecasts of unemployment rate from the VAR without WARN
    temp = np.zeros(hmax+1)
    temp[:] = np.cumsum(var_fcast[:, 2])
    u_fcast_var[loop, :] = np.exp(temp + np.log(u_init[0]))
    
    # Forecasts of finding and separation rates from the VAR with WARN factor
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varf_data, hmax, nlags)
    g_fcast_varf[loop, :] = np.exp(var_fcast[:, 0])
    s_fcast_varf[loop, :] = np.exp(var_fcast[:, 1])
    e_fcast_varf[loop, :] = var_fcast[:, 4]
    
    # Forecasts of unemployment rate from the VAR with WARN factor
    temp = np.zeros(hmax+1)
    temp[:] = np.cumsum(var_fcast[:, 2])
    u_fcast_varf[loop, :] = np.exp(temp + np.log(u_init[0]))
    
    # Forecasts of finding and separation rates from the VAR with WARN_hat
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varw_data, hmax, nlags)
    g_fcast_varw[loop, :] = np.exp(var_fcast[:, 0])
    s_fcast_varw[loop, :] = np.exp(var_fcast[:, 1])
    e_fcast_varw[loop, :] = var_fcast[:, 4]
    
    # Forecasts of unemployment rate from the VAR with WARN_hat
    temp = np.zeros(hmax+1)
    temp[:] = np.cumsum(var_fcast[:, 2])
    u_fcast_varw[loop, :] = np.exp(temp + np.log(u_init[0]))
    
    # Forecasts unemployment rate from BN model without WARN
    u_fcast_bn[loop, :] = fcast_bn_func(g_fcast_var[loop, :],
                                        s_fcast_var[loop, :], u_init[0], hmax)
    
    # Forecasts unemployment rate from BN model with WARN factor
    u_fcast_bnf[loop, :] = fcast_bn_func(g_fcast_varf[loop, :],
                                         s_fcast_varf[loop, :], u_init[0], hmax)
    
    # Forecasts unemployment rate from BN model with WARN_hat
    u_fcast_bnw[loop, :] = fcast_bn_func(g_fcast_varw[loop, :],
                                         s_fcast_varw[loop, :], u_init[0], hmax)
    
    print(loop)


#%% Compute Forecast Errors

# The unemployment rate outcomes
u_actual = df_urate.loc[str_first:str_last, :]
u_actual = u_actual.iloc[:, -1]

# The change in employment outcomes
e_actual = df_demp.loc[str_first:str_last, :]
e_actual = e_actual.iloc[:, -1]

# Storage arrays for unemployment rate forecast errors
du_var = np.zeros((nfcast, hmax+1))     # forecasts from VAR without WARN
du_varf = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN factor
du_varw = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN_hat
du_bn = np.zeros((nfcast, hmax+1))      # forecasts from BN model without WARN
du_bnf = np.zeros((nfcast, hmax+1))     # forecasts from BN model with WARN factor
du_bnw = np.zeros((nfcast, hmax+1))     # forecasts from BN model with WARN_hat

# Storage arrays for employment change forecast errors
de_var = np.zeros((nfcast, hmax+1))     # forecasts from VAR without WARN
de_varf = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN factor
de_varw = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN_hat

# Loop to compute forecast errors and RMSPEs
for jj in range(hmax+1):
    # Forecast errors for unemployment rates
    du_var[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                 - u_fcast_var[0:nfcast-jj, jj])
    du_varf[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                  - u_fcast_varf[0:nfcast-jj, jj])
    du_varw[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                  - u_fcast_varw[0:nfcast-jj, jj])
    du_bn[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                - u_fcast_bn[0:nfcast-jj, jj])
    du_bnf[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                 - u_fcast_bnf[0:nfcast-jj, jj])
    du_bnw[jj:nfcast, jj] = 100*(u_actual[jj:nfcast]
                                 - u_fcast_bnw[0:nfcast-jj, jj])
    
    # Forecast errors for employment changes
    de_var[jj:nfcast, jj] = e_actual[jj:nfcast] - e_fcast_var[0:nfcast-jj, jj]
    de_varf[jj:nfcast, jj] = e_actual[jj:nfcast] - e_fcast_varf[0:nfcast-jj, jj]
    de_varw[jj:nfcast, jj] = e_actual[jj:nfcast] - e_fcast_varw[0:nfcast-jj, jj]


#%% Function to Compute Long-Run Variance

def lrv_func(d_in, h_in):
    # Number of observations
    PP = len(d_in)
    
    # The average difference of the losses
    d_bar = np.zeros(1)
    d_bar[0] = np.mean(d_in)
    
    # Remove the average from the difference in the losses
    d_tilde = np.zeros(PP)
    d_tilde[:] = d_in - d_bar[0]
    
    # The unadjusted long-run variance
    lrv_temp = np.zeros(1)
    lrv_temp[:] = np.inner(d_tilde, d_tilde)/PP
    if h_in > 1:
        for iii in range(h_in-1):
            cov = np.zeros(1)
            cov[:] = np.inner(d_tilde[0:PP-iii-1], d_tilde[iii+1:PP])/PP
            lrv_temp[:] = lrv_temp + 2*cov
    
    # Constant to adjust the long-run variance from Harvey, Leybourne and Newbold
    const = PP + 1 - 2*h_in + h_in*(h_in - 1)/PP
    
    # The adjusted long-run variance
    lrv_out = PP*lrv_temp[0]/const
    
    return lrv_out


#%% Compute Delta Losses and Giacomini-Rossi Fluctuation Statistics

# Compute the delta loss values
diffu_bnw = np.zeros((nfcast, hmax+1))
diffu_bnw[:, :] = np.power(du_bn, 2) - np.power(du_bnw, 2)
diffu_varw = np.zeros((nfcast, hmax+1))
diffu_varw[:, :] = np.power(du_var, 2) - np.power(du_varw, 2)
diffe_varw = np.zeros((nfcast, hmax+1))
diffe_varw[:, :] = np.power(de_var, 2) - np.power(de_varw, 2)

# Storage arrays for the long-run variances
lrvu_bn = np.zeros(hmax+1)
lrvu_var = np.zeros(hmax+1)
lrve_var = np.zeros(hmax+1)

# Compute the long-run variances
for jj in range(hmax+1):
    hh = jj + 1
    lrvu_bn[jj] = lrv_func(diffu_bnw[jj:nfcast, jj], hh)
    lrvu_var[jj] = lrv_func(diffu_varw[jj:nfcast, jj], hh)
    lrve_var[jj] = lrv_func(diffe_varw[jj:nfcast, jj], hh)

# Storage arrays for fluctuation test statistics
Fu_bnw = np.zeros((nfcast-m, hmax+1))
Fu_varw = np.zeros((nfcast-m, hmax+1))
Fe_varw = np.zeros((nfcast-m, hmax+1))

# Compute the fluctuation test statistics
for ii in range(hmax+1):
    for jj in range(nfcast-m-ii):
        Fu_bnw[jj+ii, ii] = np.sum(diffu_bnw[jj+ii:jj+ii+m, ii])/np.sqrt(m*lrvu_bn[ii])
        Fu_varw[jj+ii, ii] = np.sum(diffu_varw[jj+ii:jj+ii+m, ii])/np.sqrt(m*lrvu_var[ii])
        Fe_varw[jj+ii, ii] = np.sum(diffe_varw[jj+ii:jj+ii+m, ii])/np.sqrt(m*lrve_var[ii])


#%% Plot Giacomini-Rossi Fluctuation Statistics

# Sample period to plot
sample = pd.date_range(start=str_first, end=str_last, freq='MS')

# Truncate the sample
indexm = int(np.round(m/2))
sample = sample[indexm-1:nfcast-indexm-1]

# Figure and axes instances
fig, ax = plt.subplots(3, 3, figsize=(14, 10))

h = 6
series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = Fu_bnw[h:nfcast-m, h]

ax[0, 0].plot(sample, series, linewidth=3, color='grey')
ax[0, 0].plot(sample, np.ones(nfcast-m)*3.179, linewidth=2, color='black', linestyle='--')
ax[0, 0].plot(sample, np.ones(nfcast-m)*2.948, linewidth=2, color='black', linestyle=':')
ax[0, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[0, 0].set_xlim(sample[0], sample[-1])
ax[0, 0].set_ylim(-1, 4.4)
ax[0, 0].set_yticks([-1,0, 1, 2, 3, 4])
ax[0, 0].set_ylabel('Relative performance')
ax[0, 0].set_title('Unemployment forecasts with SSUR models, h = ' + str(h), fontsize=10)

h = 12
series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = Fu_bnw[h:nfcast-m, h]

ax[1, 0].plot(sample, series, linewidth=3, color='grey')
ax[1, 0].plot(sample, np.ones(nfcast-m)*3.179, linewidth=2, color='black', linestyle='--')
ax[1, 0].plot(sample, np.ones(nfcast-m)*2.948, linewidth=2, color='black', linestyle=':')
ax[1, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[1, 0].set_xlim(sample[0], sample[-1])
ax[1, 0].set_ylim(-1, 4.4)
ax[1, 0].set_yticks([-1,0, 1, 2, 3, 4])
ax[1, 0].set_ylabel('Relative performance')
ax[1, 0].set_title('Unemployment forecasts with SSUR models, h = ' + str(h), fontsize=10)

h = 18
series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = Fu_bnw[h:nfcast-m, h]

ax[2, 0].plot(sample, series, linewidth=3, color='grey')
ax[2, 0].plot(sample, np.ones(nfcast-m)*3.179, linewidth=2, color='black', linestyle='--')
ax[2, 0].plot(sample, np.ones(nfcast-m)*2.948, linewidth=2, color='black', linestyle=':')
ax[2, 0].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[2, 0].set_xlim(sample[0], sample[-1])
ax[2, 0].set_ylim(-1, 4.4)
ax[2, 0].set_yticks([-1,0, 1, 2, 3, 4])
ax[2, 0].set_ylabel('Relative performance')
ax[2, 0].set_title('Unemployment forecasts with SSUR models, h = ' + str(h), fontsize=10)

h = 6
series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = Fu_varw[h:nfcast-m, h]

ax[0, 1].plot(sample, series, linewidth=3, color='grey')
ax[0, 1].plot(sample, np.ones(nfcast-m)*3.179, linewidth=2, color='black', linestyle='--')
ax[0, 1].plot(sample, np.ones(nfcast-m)*2.948, linewidth=2, color='black', linestyle=':')
ax[0, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[0, 1].set_xlim(sample[0], sample[-1])
ax[0, 1].set_ylim(-1, 4.4)
ax[0, 1].set_yticks([-1,0, 1, 2, 3, 4])
ax[0, 1].set_ylabel('Relative performance')
ax[0, 1].set_title('Unemployment forecasts with VAR models, h = ' + str(h), fontsize=10)

h = 12
series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = Fu_varw[h:nfcast-m, h]

ax[1, 1].plot(sample, series, linewidth=3, color='grey')
ax[1, 1].plot(sample, np.ones(nfcast-m)*3.179, linewidth=2, color='black', linestyle='--')
ax[1, 1].plot(sample, np.ones(nfcast-m)*2.948, linewidth=2, color='black', linestyle=':')
ax[1, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[1, 1].set_xlim(sample[0], sample[-1])
ax[1, 1].set_ylim(-1, 4.4)
ax[1, 1].set_yticks([-1,0, 1, 2, 3, 4])
ax[1, 1].set_ylabel('Relative performance')
ax[1, 1].set_title('Unemployment forecasts with VAR models, h = ' + str(h), fontsize=10)

h = 18
series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = Fu_varw[h:nfcast-m, h]

ax[2, 1].plot(sample, series, linewidth=3, color='grey')
ax[2, 1].plot(sample, np.ones(nfcast-m)*3.179, linewidth=2, color='black', linestyle='--')
ax[2, 1].plot(sample, np.ones(nfcast-m)*2.948, linewidth=2, color='black', linestyle=':')
ax[2, 1].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[2, 1].set_xlim(sample[0], sample[-1])
ax[2, 1].set_ylim(-1, 4.4)
ax[2, 1].set_yticks([-1,0, 1, 2, 3, 4])
ax[2, 1].set_ylabel('Relative performance')
ax[2, 1].set_title('Unemployment forecasts with VAR models, h = ' + str(h), fontsize=10)

h = 6
series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = Fe_varw[h:nfcast-m, h]

ax[0, 2].plot(sample, series, linewidth=3, color='grey')
ax[0, 2].plot(sample, np.ones(nfcast-m)*3.179, linewidth=2, color='black', linestyle='--')
ax[0, 2].plot(sample, np.ones(nfcast-m)*2.948, linewidth=2, color='black', linestyle=':')
ax[0, 2].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[0, 2].set_xlim(sample[0], sample[-1])
ax[0, 2].set_ylim(-1, 4.4)
ax[0, 2].set_yticks([-1,0, 1, 2, 3, 4])
ax[0, 2].set_ylabel('Relative performance')
ax[0, 2].set_title('Employment forecasts with VAR models, h = ' + str(h), fontsize=10)

h = 12
series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = Fe_varw[h:nfcast-m, h]

ax[1, 2].plot(sample, series, linewidth=3, color='grey')
ax[1, 2].plot(sample, np.ones(nfcast-m)*3.179, linewidth=2, color='black', linestyle='--')
ax[1, 2].plot(sample, np.ones(nfcast-m)*2.948, linewidth=2, color='black', linestyle=':')
ax[1, 2].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[1, 2].set_xlim(sample[0], sample[-1])
ax[1, 2].set_ylim(-1, 4.4)
ax[1, 2].set_yticks([-1,0, 1, 2, 3, 4])
ax[1, 2].set_ylabel('Relative performance')
ax[1, 2].set_title('Employment forecasts with VAR models, h = ' + str(h), fontsize=10)

h = 18
series = np.zeros(nfcast - m)
series[:] = np.nan
series[h:nfcast-m] = Fe_varw[h:nfcast-m, h]

ax[2, 2].plot(sample, series, linewidth=3, color='grey')
ax[2, 2].plot(sample, np.ones(nfcast-m)*3.179, linewidth=2, color='black', linestyle='--')
ax[2, 2].plot(sample, np.ones(nfcast-m)*2.948, linewidth=2, color='black', linestyle=':')
ax[2, 2].plot(sample, np.zeros(nfcast-m), linewidth=1, color='black', linestyle='-')
ax[2, 2].set_xlim(sample[0], sample[-1])
ax[2, 2].set_ylim(-1, 3.99)
ax[2, 2].set_yticks([-1,0, 1, 2, 3])
ax[2, 2].set_ylabel('Relative performance')
ax[2, 2].set_title('Employment forecasts with the VAR models, h = ' + str(h), fontsize=10)

fig.tight_layout()
plt.savefig('GiacominiRossiStats.pdf', bbox_inches='tight')


#%% Disply the Run Time

# This line display the run time in minutes
print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


