# -*- coding: utf-8 -*-
"""
This program produces Table A.6 in "Advance Layoff Notices and Aggregate Job
Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    fcast_eval_module.py : module with functions to compute and evaluate forecasts
    warn_factors_rt.xlsx : Excel file with pseudo real-time WARN factors and
                           values of WARN hat from Equation (4)
    labor_rates_rt.xlsx : Excel file with real-time uneployment rates and 
                          labor market flow rates
    uspriv_alfred.xlsx : Excel file with real-time private employment levels
    icsa_fred.xlsx : Excel file holding monthly averages of initial
                     unemployment insurance claims

The main output of this program is
    table : an array holding the values for Table A.6 in the paper
    tstats : an array with the t-statistics used to assess statistical
             significance in Table A.6
"""

#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Pandas and NumPy
import pandas as pd
import numpy as np

# Functions for making and evaluating forecasts
from modules.fcast_eval_module import fcast_var_func, tstat_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the program

# First date that WARN factor is available
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# First date for forecast evaluation
str_first = '20080101'  # Format: 20080101 corresponds to January 1, 2008
                        # 20030101 is the earliest permissable choice

# Last date for forecast evaluation
str_last = '20191231'   # Format: 20191231 corresponds to December 31, 2019

# Number of variables in the VAR including WARN factor or WARN hat
nvars = 6

# Number of lags in the forecasting models
nlags = 2

# Longest available sample size for estimating the forecasting models
nsample = len(pd.date_range(start=str_warn, end=str_first, freq='MS')) - 2

# Maximum horizon to forecast
hmax = 24


#%% Load the Data

# Number of months to truncate
KK = len(pd.date_range(start='20030101', end=str_first, freq='MS')) - 1

# Change in warn factor
path_factors = os.path.join(path_dir, 'section_6_2', 'warn_factors_rt.xlsx')
df_factor = pd.read_excel(path_factors, 'factors', index_col=0)
df_factor = df_factor.diff()
df_factor = df_factor.loc[str_warn:, :]
df_factor = df_factor.iloc[:, KK:]

# Change in warn_hat
df_warn = pd.read_excel(path_factors, 'warn_hat', index_col=0)
df_warn = df_warn.diff()
df_warn = df_warn.loc[str_warn:, :]
df_warn = df_warn.iloc[:, KK:]

# Unemployment rate
path_labor = os.path.join(path_dir, 'section_6_2', 'labor_rates_rt.xlsx')
df_urate = pd.read_excel(path_labor, 'urate', index_col=0)
df_urate = df_urate.loc[str_warn:, :]
df_urate = df_urate.iloc[:, KK:]

# Change in log unemployment rate
df_uchange = np.log(pd.read_excel(path_labor, 'urate', index_col=0))
df_uchange = df_uchange.diff()
df_uchange = df_uchange.loc[str_warn:, :]
df_uchange = df_uchange.iloc[:, KK:]

# Log finding rate
df_grate = np.log(pd.read_excel(path_labor, 'grate', index_col=0))
df_grate = df_grate.loc[str_warn:, :]
df_grate = df_grate.iloc[:, KK:]

# Log separation rate
df_srate = np.log(pd.read_excel(path_labor, 'srate', index_col=0))
df_srate = df_srate.loc[str_warn:, :]
df_srate = df_srate.iloc[:, KK:]

# Level and change of private employment
path_emp_data = os.path.join(path_dir, 'data_input', 'uspriv_alfred.xlsx')
df_emp = pd.read_excel(path_emp_data, 'Sheet1', index_col=0)
df_demp = df_emp.diff()
df_emp = df_emp.loc[str_warn:, :]
df_demp = df_demp.loc[str_warn:, :]
df_demp = df_demp.iloc[:, KK:]

# initial claims
path_claims_data = os.path.join(path_dir, 'data_input', 'icsa_alfred.xlsx')
df_claim = np.log(pd.read_excel(path_claims_data, 'Sheet2', index_col=0))
df_claim = df_claim.loc[str_warn:, :]
df_claim = df_claim.iloc[:, KK:]


#%% Compute the Forecasts

# Pseudo real-time dates that forecasts are produced for
fcast_index = pd.date_range(start=str_first, end=str_last, freq='MS')

# Number of forecasts to compute
nfcast = len(fcast_index)

# Storage arrays for finding rate forecasts
g_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
g_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
g_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for separation rate forecasts
s_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
s_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
s_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Loop to compute the forecasts
for loop in range(nfcast):
    # Number of observations in this loop
    nobs = len(df_uchange.iloc[:, loop].dropna())
    
    # Variables for VAR forecasts without WARN
    var_data = np.zeros((nsample, nvars-1))
    var_data[:, 0] = df_grate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 1] = df_srate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 2] = df_uchange.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 3] = df_claim.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 4] = df_demp.iloc[nobs-nsample:nobs, loop].to_numpy()
    
    # Variables for VAR forecasts with WARN factor
    varf_data = np.zeros((nsample, nvars))
    varf_data[:, 0:nvars-1] = var_data[:, :]
    varf_data[:, nvars-1] = df_factor.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Variables for VAR forecasts with WARN_hat
    varw_data = np.zeros((nsample, nvars))
    varw_data[:, 0:nvars-1] = var_data[:, :]
    varw_data[:, nvars-1] = df_warn.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Forecasts of finding and separation rates from the VAR without WARN
    var_fcast = np.zeros((hmax+1, nvars-1))
    var_fcast[:, :] = fcast_var_func(var_data, hmax, nlags)
    g_fcast_var[loop, :] = np.exp(var_fcast[:, 0])
    s_fcast_var[loop, :] = np.exp(var_fcast[:, 1])
    
    # Forecasts of finding and separation rates from the VAR with WARN factor
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varf_data, hmax, nlags)
    g_fcast_varf[loop, :] = np.exp(var_fcast[:, 0])
    s_fcast_varf[loop, :] = np.exp(var_fcast[:, 1])
    
    # Forecasts of finding and separation rates from the VAR with WARN_hat
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varw_data, hmax, nlags)
    g_fcast_varw[loop, :] = np.exp(var_fcast[:, 0])
    s_fcast_varw[loop, :] = np.exp(var_fcast[:, 1])
    
    print(loop)


#%% Forecast Evaluation: Compute Forecast Errors and RMSPEs

# The separation rate outcomes
s_actual = np.exp(df_srate.loc[str_first:str_last, :])
s_actual = s_actual.iloc[:, -1]

# The finding rate outcomes
g_actual = np.exp(df_grate.loc[str_first:str_last, :])
g_actual = g_actual.iloc[:, -1]

# Storage arrays for flows forecast errors
ds_var = np.zeros((nfcast, hmax+1))     # forecasts from VAR without WARN
ds_varf = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN factor
ds_varw = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN_hat

dg_var = np.zeros((nfcast, hmax+1))     # forecasts from VAR without WARN
dg_varf = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN factor
dg_varw = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN_hat

# Storage arrays for flows RMSPEs
rmspe_s_var = np.zeros(hmax+1)
rmspe_s_varf = np.zeros(hmax+1)
rmspe_s_varw = np.zeros(hmax+1)

rmspe_g_var = np.zeros(hmax+1)
rmspe_g_varf = np.zeros(hmax+1)
rmspe_g_varw = np.zeros(hmax+1)

# Loop to compute forecast errors and RMSPEs
for jj in range(hmax+1):
    # Forecast errors for flows
    ds_var[jj:nfcast, jj] = s_actual[jj:nfcast] - s_fcast_var[0:nfcast-jj, jj]
    ds_varf[jj:nfcast, jj] = s_actual[jj:nfcast] - s_fcast_varf[0:nfcast-jj, jj]
    ds_varw[jj:nfcast, jj] = s_actual[jj:nfcast] - s_fcast_varw[0:nfcast-jj, jj]
    
    dg_var[jj:nfcast, jj] = g_actual[jj:nfcast] - g_fcast_var[0:nfcast-jj, jj]
    dg_varf[jj:nfcast, jj] = g_actual[jj:nfcast] - g_fcast_varf[0:nfcast-jj, jj]
    dg_varw[jj:nfcast, jj] = g_actual[jj:nfcast] - g_fcast_varw[0:nfcast-jj, jj]
    
    # RMSPEs for flows
    rmspe_s_var[jj] = np.sqrt(np.mean(np.power(ds_var[jj:nfcast, jj], 2)))
    rmspe_s_varf[jj] = np.sqrt(np.mean(np.power(ds_varf[jj:nfcast, jj], 2)))
    rmspe_s_varw[jj] = np.sqrt(np.mean(np.power(ds_varw[jj:nfcast, jj], 2)))
    
    rmspe_g_var[jj] = np.sqrt(np.mean(np.power(dg_var[jj:nfcast, jj], 2)))
    rmspe_g_varf[jj] = np.sqrt(np.mean(np.power(dg_varf[jj:nfcast, jj], 2)))
    rmspe_g_varw[jj] = np.sqrt(np.mean(np.power(dg_varw[jj:nfcast, jj], 2)))


#%% Collect RMSPEs and Relative RMSPEs for a Table

# Storage array
table = np.zeros((7, hmax+1))

# Panel A
table[0, :] = rmspe_s_var
table[1, :] = np.divide(rmspe_s_varf, rmspe_s_var)
table[2, :] = np.divide(rmspe_s_varw, rmspe_s_var)

# Panel B
table[3, :] = rmspe_g_var
table[4, :] = np.divide(rmspe_g_varf, rmspe_g_var)
table[5, :] = np.divide(rmspe_g_varw, rmspe_g_var)

# Round to 4 digits for presentation
table = np.round(table, 4)

# Sample sizes
for jj in range(hmax+1):
    table[6, jj] = nfcast - jj


#%% Forecast Evaluation: Tests of Statistical Significance

# Storage arrays for unemployment rate t statistics
tstats = np.zeros((4, hmax+1))

# Loop to evaluate forecasts
for jj in range(hmax+1):
    # Forecast horizon
    h = jj + 1
    
    # The t-statistics for the table
    tstats[0, jj] = tstat_func(ds_var[jj:nfcast, jj], ds_varf[jj:nfcast, jj], h)
    tstats[1, jj] = tstat_func(ds_var[jj:nfcast, jj], ds_varw[jj:nfcast, jj], h)
    tstats[2, jj] = tstat_func(dg_var[jj:nfcast, jj], dg_varf[jj:nfcast, jj], h)
    tstats[3, jj] = tstat_func(dg_var[jj:nfcast, jj], dg_varw[jj:nfcast, jj], h)


#%% Disply the Run Time

# This line display the run time in minutes
print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


