# -*- coding: utf-8 -*-
"""
This program produces Table A.7 in "Advance Layoff Notices and Aggregate Job
Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    fcast_eval_module.py : module with functions to compute and evaluate forecasts
    warn_factors_rt.xlsx : Excel file with pseudo real-time WARN factors and
                           values of WARN hat from Equation (4)
    labor_rates_rt.xlsx : Excel file with real-time uneployment rates and 
                          labor market flow rates
    uspriv_alfred.xlsx : Excel file with real-time private employment levels
    manemp_alfred.xlsx : Excel file with real-time manufacturing employment levels
    icsa_fred.xlsx : Excel file holding monthly averages of initial
                     unemployment insurance claims

The main output of this program is
    table : an array holding the values for Table A.6 in the paper
    tstats : an array with the t-statistics used to assess statistical
             significance in Table A.7
"""

#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Pandas and NumPy
import pandas as pd
import numpy as np

# Functions for making and evaluating forecasts
from modules.fcast_eval_module import fcast_var_func, tstat_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the program

# First date that WARN factor is available
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# First date for forecast evaluation
str_first = '20080101'  # Format: 20080101 corresponds to January 1, 2008
                        # 20030101 is the earliest permissable choice

# Last date for forecast evaluation
str_last = '20191231'   # Format: 20191231 corresponds to December 31, 2019

# Number of variables in the VAR including WARN factor or WARN hat
nvars = 6

# Number of lags in the forecasting models
nlags = 2

# Longest available sample size for estimating the forecasting models
nsample = len(pd.date_range(start=str_warn, end=str_first, freq='MS')) - 2

# Maximum horizon to forecast
hmax = 24


#%% Load the Data

# Number of months to truncate
KK = len(pd.date_range(start='20030101', end=str_first, freq='MS')) - 1

# Change in warn factor
path_factors = os.path.join(path_dir, 'section_6_2', 'warn_factors_rt.xlsx')
df_factor = pd.read_excel(path_factors, 'factors', index_col=0)
df_factor = df_factor.diff()
df_factor = df_factor.loc[str_warn:, :]
df_factor = df_factor.iloc[:, KK:]

# Change in warn_hat
df_warn = pd.read_excel(path_factors, 'warn_hat', index_col=0)
df_warn = df_warn.diff()
df_warn = df_warn.loc[str_warn:, :]
df_warn = df_warn.iloc[:, KK:]

# Unemployment rate
path_labor = os.path.join(path_dir, 'section_6_2', 'labor_rates_rt.xlsx')
df_urate = pd.read_excel(path_labor, 'urate', index_col=0)
df_urate = df_urate.loc[str_warn:, :]
df_urate = df_urate.iloc[:, KK:]

# Change in log unemployment rate
df_uchange = np.log(pd.read_excel(path_labor, 'urate', index_col=0))
df_uchange = df_uchange.diff()
df_uchange = df_uchange.loc[str_warn:, :]
df_uchange = df_uchange.iloc[:, KK:]

# Log finding rate
df_grate = np.log(pd.read_excel(path_labor, 'grate', index_col=0))
df_grate = df_grate.loc[str_warn:, :]
df_grate = df_grate.iloc[:, KK:]

# Log separation rate
df_srate = np.log(pd.read_excel(path_labor, 'srate', index_col=0))
df_srate = df_srate.loc[str_warn:, :]
df_srate = df_srate.iloc[:, KK:]

# Level and change of private employment
path_emp_data = os.path.join(path_dir, 'data_input', 'uspriv_alfred.xlsx')
df_emp = pd.read_excel(path_emp_data, 'Sheet1', index_col=0)
df_demp = df_emp.diff()
df_emp = df_emp.loc[str_warn:, :]
df_demp = df_demp.loc[str_warn:, :]
df_demp = df_demp.iloc[:, KK:]

# Level and change of manufacturing employment
path_man_data = os.path.join(path_dir, 'data_input', 'manemp_alfred.xlsx')
df_man = pd.read_excel(path_man_data, 'Sheet1', index_col=0)
df_dman = df_man.diff()
df_man = df_man.loc[str_warn:, :]
df_dman = df_dman.loc[str_warn:, :]
df_dman = df_dman.iloc[:, KK:]

# Level and change of private employment ex manufacturing
df_dempx = pd.DataFrame(index=df_demp.index, columns=df_demp.columns)
df_dempx.loc[:, :] = df_demp.to_numpy() - df_dman.to_numpy()

# initial claims
path_claims_data = os.path.join(path_dir, 'data_input', 'icsa_alfred.xlsx')
df_claim = np.log(pd.read_excel(path_claims_data, 'Sheet2', index_col=0))
df_claim = df_claim.loc[str_warn:, :]
df_claim = df_claim.iloc[:, KK:]


#%% Compute the Forecasts

# First date that forecasts are produced for
fcast_index = pd.date_range(start=str_first, end=str_last, freq='MS')

# Number of forecasts to compute
nfcast = len(fcast_index)

# Storage arrays for private employment forecasts
e_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
e_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
e_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for manufacturing employment forecasts
man_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
man_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
man_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for private employment ex manufacturing forecasts
ex_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
ex_fcast_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
ex_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Loop to compute the forecasts
for loop in range(nfcast):
    # Number of observations in this loop
    nobs = len(df_uchange.iloc[:, loop].dropna())
    
    # Variables for VAR forecasts without WARN
    var_data = np.zeros((nsample, nvars-1))
    var_data[:, 0] = df_grate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 1] = df_srate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 2] = df_uchange.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 3] = df_claim.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 4] = df_demp.iloc[nobs-nsample:nobs, loop].to_numpy()
    
    # Variables for VAR forecasts with WARN factor
    varf_data = np.zeros((nsample, nvars))
    varf_data[:, 0:nvars-1] = var_data[:, :]
    varf_data[:, nvars-1] = df_factor.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Variables for VAR forecasts with WARN_hat
    varw_data = np.zeros((nsample, nvars))
    varw_data[:, 0:nvars-1] = var_data[:, :]
    varw_data[:, nvars-1] = df_warn.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Forecasts of private employment from the VAR without WARN
    var_fcast = np.zeros((hmax+1, nvars-1))
    var_fcast[:, :] = fcast_var_func(var_data, hmax, nlags)
    e_fcast_var[loop, :] = var_fcast[:, 4]
    
    # Forecasts of private employment from the VAR with WARN factor
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varf_data, hmax, nlags)
    e_fcast_varf[loop, :] = var_fcast[:, 4]
    
    # Forecasts of private employment from the VAR with WARN_hat
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varw_data, hmax, nlags)
    e_fcast_varw[loop, :] = var_fcast[:, 4]
        
    # Variables for VAR forecasts without WARN
    var_data = np.zeros((nsample, nvars-1))
    var_data[:, 0] = df_grate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 1] = df_srate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 2] = df_uchange.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 3] = df_claim.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 4] = df_dman.iloc[nobs-nsample:nobs, loop].to_numpy()
    
    # Variables for VAR forecasts with WARN factor
    varf_data = np.zeros((nsample, nvars))
    varf_data[:, 0:nvars-1] = var_data[:, :]
    varf_data[:, nvars-1] = df_factor.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Variables for VAR forecasts with WARN_hat
    varw_data = np.zeros((nsample, nvars))
    varw_data[:, 0:nvars-1] = var_data[:, :]
    varw_data[:, nvars-1] = df_warn.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Forecasts of manufacturing employment from the VAR without WARN
    var_fcast = np.zeros((hmax+1, nvars-1))
    var_fcast[:, :] = fcast_var_func(var_data, hmax, nlags)
    man_fcast_var[loop, :] = var_fcast[:, 4]
    
    # Forecasts of manufacturing employment from the VAR with WARN factor
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varf_data, hmax, nlags)
    man_fcast_varf[loop, :] = var_fcast[:, 4]
    
    # Forecasts of manufacturing employment from the VAR with WARN_hat
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varw_data, hmax, nlags)
    man_fcast_varw[loop, :] = var_fcast[:, 4]
    
    # Variables for VAR forecasts without WARN
    var_data = np.zeros((nsample, nvars-1))
    var_data[:, 0] = df_grate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 1] = df_srate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 2] = df_uchange.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 3] = df_claim.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 4] = df_dempx.iloc[nobs-nsample:nobs, loop].to_numpy()
    
    # Variables for VAR forecasts with WARN factor
    varf_data = np.zeros((nsample, nvars))
    varf_data[:, 0:nvars-1] = var_data[:, :]
    varf_data[:, nvars-1] = df_factor.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Variables for VAR forecasts with WARN_hat
    varw_data = np.zeros((nsample, nvars))
    varw_data[:, 0:nvars-1] = var_data[:, :]
    varw_data[:, nvars-1] = df_warn.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Forecasts of private employment ex manufacturing from the VAR without WARN
    var_fcast = np.zeros((hmax+1, nvars-1))
    var_fcast[:, :] = fcast_var_func(var_data, hmax, nlags)
    ex_fcast_var[loop, :] = var_fcast[:, 4]
    
    # Forecasts of private employment ex manufacturing from the VAR with WARN factor
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varf_data, hmax, nlags)
    ex_fcast_varf[loop, :] = var_fcast[:, 4]
    
    # Forecasts of private employment ex manufacturing from the VAR with WARN_hat
    var_fcast = np.zeros((hmax+1, nvars))
    var_fcast[:, :] = fcast_var_func(varw_data, hmax, nlags)
    ex_fcast_varw[loop, :] = var_fcast[:, 4]
    
    print(loop)


#%% Forecast Evaluation: Compute Forecast Errors and RMSPEs

# The private employment outcomes
e_actual = df_demp.loc[str_first:str_last, :]
e_actual = e_actual.iloc[:, -1]

# The manufacturing employment outcomes
man_actual = df_dman.loc[str_first:str_last, :]
man_actual = man_actual.iloc[:, -1]

# The private employment ex manufacturing outcomes
ex_actual = df_dempx.loc[str_first:str_last, :]
ex_actual = ex_actual.iloc[:, -1]

# Storage arrays for private employment forecast errors
de_var = np.zeros((nfcast, hmax+1))     # forecasts from VAR without WARN
de_varf = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN factor
de_varw = np.zeros((nfcast, hmax+1))    # forecasts from VAR with WARN_hat

# Storage arrays for manufacturing employment forecast errors
dman_var = np.zeros((nfcast, hmax+1))   # forecasts from VAR without WARN
dman_varf = np.zeros((nfcast, hmax+1))  # forecasts from VAR with WARN factor
dman_varw = np.zeros((nfcast, hmax+1))  # forecasts from VAR with WARN_hat

# Storage arrays for private employment ex manufacturing forecast errors
dex_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
dex_varf = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN factor
dex_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for private employment RMSPEs
rmspe_e_var = np.zeros(hmax+1)
rmspe_e_varf = np.zeros(hmax+1)
rmspe_e_varw = np.zeros(hmax+1)

# Storage arrays for manufacturing employment RMSPEs
rmspe_man_var = np.zeros(hmax+1)
rmspe_man_varf = np.zeros(hmax+1)
rmspe_man_varw = np.zeros(hmax+1)

# Storage arrays for private employment ex manufacturing RMSPEs
rmspe_ex_var = np.zeros(hmax+1)
rmspe_ex_varf = np.zeros(hmax+1)
rmspe_ex_varw = np.zeros(hmax+1)

# Loop to compute forecast errors and RMSPEs
for jj in range(hmax+1):
    # Forecast errors for employment changes
    de_var[jj:nfcast, jj] = e_actual[jj:nfcast] - e_fcast_var[0:nfcast-jj, jj]
    de_varf[jj:nfcast, jj] = e_actual[jj:nfcast] - e_fcast_varf[0:nfcast-jj, jj]
    de_varw[jj:nfcast, jj] = e_actual[jj:nfcast] - e_fcast_varw[0:nfcast-jj, jj]
    
    dman_var[jj:nfcast, jj] = man_actual[jj:nfcast] - man_fcast_var[0:nfcast-jj, jj]
    dman_varf[jj:nfcast, jj] = man_actual[jj:nfcast] - man_fcast_varf[0:nfcast-jj, jj]
    dman_varw[jj:nfcast, jj] = man_actual[jj:nfcast] - man_fcast_varw[0:nfcast-jj, jj]
    
    dex_var[jj:nfcast, jj] = ex_actual[jj:nfcast] - ex_fcast_var[0:nfcast-jj, jj]
    dex_varf[jj:nfcast, jj] = ex_actual[jj:nfcast] - ex_fcast_varf[0:nfcast-jj, jj]
    dex_varw[jj:nfcast, jj] = ex_actual[jj:nfcast] - ex_fcast_varw[0:nfcast-jj, jj]
    
    # RMSPEs for employment changes
    rmspe_e_var[jj] = np.sqrt(np.mean(np.power(de_var[jj:nfcast, jj], 2)))
    rmspe_e_varf[jj] = np.sqrt(np.mean(np.power(de_varf[jj:nfcast, jj], 2)))
    rmspe_e_varw[jj] = np.sqrt(np.mean(np.power(de_varw[jj:nfcast, jj], 2)))
    
    rmspe_man_var[jj] = np.sqrt(np.mean(np.power(dman_var[jj:nfcast, jj], 2)))
    rmspe_man_varf[jj] = np.sqrt(np.mean(np.power(dman_varf[jj:nfcast, jj], 2)))
    rmspe_man_varw[jj] = np.sqrt(np.mean(np.power(dman_varw[jj:nfcast, jj], 2)))
    
    rmspe_ex_var[jj] = np.sqrt(np.mean(np.power(dex_var[jj:nfcast, jj], 2)))
    rmspe_ex_varf[jj] = np.sqrt(np.mean(np.power(dex_varf[jj:nfcast, jj], 2)))
    rmspe_ex_varw[jj] = np.sqrt(np.mean(np.power(dex_varw[jj:nfcast, jj], 2)))


#%% Collect RMSPEs and Relative RMSPEs for a Table

# Storage array
table = np.zeros((10, hmax+1))

# Panel A
table[0, :] = rmspe_e_var
table[1, :] = np.divide(rmspe_e_varf, rmspe_e_var)
table[2, :] = np.divide(rmspe_e_varw, rmspe_e_var)

# Panel B
table[3, :] = rmspe_man_var
table[4, :] = np.divide(rmspe_man_varf, rmspe_man_var)
table[5, :] = np.divide(rmspe_man_varw, rmspe_man_var)

# Panel C
table[6, :] = rmspe_ex_var
table[7, :] = np.divide(rmspe_ex_varf, rmspe_ex_var)
table[8, :] = np.divide(rmspe_ex_varw, rmspe_ex_var)

# Round to 2 digits for presentation
table = np.round(table, 2)

# Sample sizes
for jj in range(hmax+1):
    table[9, jj] = nfcast - jj


#%% Forecast Evaluation: Tests of Statistical Significance

# Storage arrays for unemployment rate t statistics
tstats = np.zeros((6, hmax+1))

# Loop to evaluate forecasts
for jj in range(hmax+1):
    # Forecast horizon
    h = jj + 1
    
    # The t-statistics for the table
    tstats[0, jj] = tstat_func(de_var[jj:nfcast, jj], de_varf[jj:nfcast, jj], h)
    tstats[1, jj] = tstat_func(de_var[jj:nfcast, jj], de_varw[jj:nfcast, jj], h)
    tstats[2, jj] = tstat_func(dman_var[jj:nfcast, jj], dman_varf[jj:nfcast, jj], h)
    tstats[3, jj] = tstat_func(dman_var[jj:nfcast, jj], dman_varw[jj:nfcast, jj], h)
    tstats[4, jj] = tstat_func(dex_var[jj:nfcast, jj], dex_varf[jj:nfcast, jj], h)
    tstats[5, jj] = tstat_func(dex_var[jj:nfcast, jj], dex_varw[jj:nfcast, jj], h)


#%% Disply the Run Time

# This line display the run time in minutes
print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


