# -*- coding: utf-8 -*-
"""
This program computes pseudo real-time WARN factors used in "Advance Layoff
Notices and Aggregate Job Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

The WARN factors are computed from the DFM described in the paper.
We also compute pseudo real-time values of WARN_hat from equation (4) in the paper.

This program calls the following modules:
    seasonal_module.py : module with functions to seasonally adjust the data
    warn_module.py : module with functions to estimate the DFM

This program also calls the following Excel files that hold real-time WARN data
    WARNData_NSA_20200115.xlsx
    WARNData_NSA_20200214.xlsx
    WARNData_NSA_20200316.xlsx
    WARNData_NSA_20200415.xlsx
    WARNData_NSA_20200515.xlsx
    WARNData_NSA_20200615.xlsx
    WARNData_NSA_20200715.xlsx
    WARNData_NSA_20200814.xlsx
    WARNData_NSA_20200915.xlsx
    WARNData_NSA_20201015.xlsx
    WARNData_NSA_20201113.xlsx
    WARNData_NSA_20201215.xlsx
    WARNData_NSA_20210115.xlsx
    WARNData_NSA_20210212.xlsx
    WARNData_NSA_20210315.xlsx
    WARNData_NSA_20210415.xlsx
    WARNData_NSA_20210514.xlsx
    WARNData_NSA_20210615.xlsx
    WARNData_NSA_20210715.xlsx
    WARNData_NSA_20210813.xlsx
    WARNData_NSA_20210915.xlsx
    WARNData_NSA_20211015.xlsx
    WARNData_NSA_20211115.xlsx
    WARNData_NSA_20211215.xlsx
Each of these files has a YYYMMDD date stamp that indicates when the data in
the file was collected by.

The main output of this program is
    df_factor : a Pandas DataFrame holding pseudo real-time WARN factors
    df_warn_hat : a Pandas DataFrame holding pseudo real-time values of WARN_hat
These DataFrames are saved to warn_factors_rt.xlsx.
"""


#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Pandas and NumPy
import pandas as pd
import numpy as np

# Functions from seasonal module
from modules.seasonal_module import sa_func

# Functions from the WARN module
from modules.dfm_em_module import make_guess, em_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the program

# String to choose first date for WARN data
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# String to choose the initial period for forecasting
str_init = '20191231'   # Format: 20021231 corresponds to December 31, 2002

# Number of lags in the state equation
nlags = 1

# Parameters of EM algorithm for seasonal adjustment
nloop = 20000       # maximum number of iterations in EM algrorithm
thresh_sa = 1e-3    # threshold for convergence

# Parameters of the EM algorithm
thresh = 1e-5       # threshold for convergence

# Date strings for calling the real-time WARN data
str_dates = ['20200115', '20200214', '20200316', '20200415', '20200515',
             '20200615', '20200715', '20200814', '20200915', '20201015',
             '20201113', '20201215', '20210115', '20210212', '20210315',
             '20210415', '20210514', '20210615', '20210715', '20210813',
             '20210915', '20211015', '20211115', '20211215']


#%% Produce the WARN factors

# Number of factor vintages to compute
nfactors = len(str_dates)

# Column dates
col_dates = pd.to_datetime(str_dates)

# Dataframe to save the estimated factor vintages
fcast_index = pd.date_range(start=str_warn, end='20211130', freq='MS')
df_factor = pd.DataFrame(index=fcast_index, columns=col_dates)
df_warn_hat = pd.DataFrame(index=fcast_index, columns=col_dates)

# Loop to produce the factor vintages
for loop in range(nfactors):
    # Dataframe for number of workers affected by WARN notices by month and state
    str_file_name = 'WARNData_NSA_' + str_dates[loop] + '.xlsx'
    path_warn_data = os.path.join(path_dir, 'data_input', str_file_name)
    df_warn_all = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)
    
    # Drop small states or states with too many observations with a value of 0
    df_warn = df_warn_all.drop(columns=['AK', 'ID', 'KY', 'ND', 'NM',
                                        'OK', 'RI', 'UT', 'WV'])
    
    # Number of states in the sample
    nvars = len(df_warn.columns)
    
    # Dataframe for seasonally adjusted data
    df_warn_sa = pd.DataFrame(index=df_warn.index, columns=df_warn.columns)
    
    # Loop for seasonal adjustment
    for jj in range(nvars):
        # State name
        state_name = df_warn.columns[jj]
        
        # Dates with observed data
        state_index = df_warn.loc[:, state_name].dropna().index
        
        # Number of observations excluding NaNs
        nsample = len(state_index)
        
        # Convert the data to numpy while drpping NaNs
        X_nsa = np.zeros((nsample, 1))
        X_nsa[:, 0] = df_warn.loc[:, state_name].dropna().to_numpy()
        
        # Seasonally adjust the data
        data_sa = sa_func(X_nsa, nloop, thresh_sa, state_name)
        
        # Save the data
        df_warn_sa.loc[state_index, state_name] = data_sa[:, 0]
        
        # Delete temprary variables
        del state_name, state_index, nsample, X_nsa, data_sa, jj
    
    # Adjust the sample of the seasonally adjusted data
    df_warn_sa = df_warn_sa.loc[str_warn:]
    
    # Number of observations in the sample
    nsample = len(df_warn_sa.index)
    
    # Convert the data to numpy
    X = np.zeros((nsample, nvars))
    X[:, :] = df_warn_sa.to_numpy()
    
    # Set zero values to NaN and construct the selection matrices
    W = np.zeros((nvars, nvars, nsample))
    for jj in range(nsample):
        # Set zero values to NaN
        for ii in range(nvars):
            if X[jj, ii] < 1:
                X[jj, ii] = np.nan
        
        # Selection matrices
        W[:, :, jj] = np.diag(np.isnan(X[jj, :])==0)
    
    # Delete temprary variables
    del ii, jj
    
    # Take logs of the data
    X = np.log(X)
    
    # Guess of model parameters
    Fguess, Aguess, Qguess, Lguess, Rguess, Zstd, Avg, Var = \
        make_guess(X, nlags)
    
    # Guess of the initial state
    F0guess = np.zeros((nlags, 1))
    
    # Guess of covaraiance of initial state
    Qvec = np.zeros((nlags*nlags, 1))
    Qvec[0, 0] = Qguess[0, 0]
    
    tmp = np.zeros((nlags*nlags, nlags*nlags))
    tmp[:, :] = np.eye(nlags*nlags) - np.kron(Aguess, Aguess)
    
    P0guess = np.zeros((nlags*nlags, 1))
    P0guess[:, :] = np.linalg.solve(tmp, Qvec)
    P0guess = np.reshape(P0guess, (nlags, nlags))
    
    # Delete temprary variables
    del Qvec, tmp
    
    # EM algorithm to compute factor and model parameters
    Fhat, Phat, Ahat, Qhat, Lhat, Rhat, Loglikes, ell = \
        em_func(F0guess, P0guess, Zstd, Lguess, Rguess, Aguess, Qguess,
                W, nloop, thresh)
    
    # Check if the EM algoithm finished before reaching max iterations
    if ell == nloop-1:
        print('No solution found with EM algorithm. '
              + 'Maximum number of iterations reached.')
    else:
        print('Solution to EM algorithm found on loop ' + str(ell))
    
    # Compute WARN factor in term of units of employees affected
    tmp = np.zeros((nvars, 1))
    warn_hat = np.zeros((nsample, 1))
    for jj in range(nsample):
        tmp[:, 0] = np.multiply(np.sqrt(Var[:, 0]), (Lhat[:, 0]*Fhat[0, jj+1]))
        warn_hat[jj, 0] = np.sum(np.exp(Avg + tmp))
    
    # Store the factors
    df_factor.iloc[0:nsample, loop] = Fhat[0, 1:nsample+1]
    df_warn_hat.iloc[0:nsample, loop] = warn_hat[:, 0]
    
    # Delete temprary variables
    del str_file_name, path_warn_data, df_warn_all, df_warn, nvars, df_warn_sa
    del Fguess, Aguess, Qguess, Lguess, Rguess, Zstd, Avg, Var
    del Fhat, Phat, Ahat, Qhat, Lhat, Rhat, Loglikes, ell, tmp, jj
    
    print(' ')
    print('Loop number ' + str(loop+1) + ' of ' + str(nfactors) + ' is done.')
    print(' ')


#%% Save the Output

str_file = 'warn_factors_rt.xlsx'
with pd.ExcelWriter(str_file, datetime_format='YYYY-MM') as writer:
    df_factor.to_excel(writer, sheet_name='factors')
    df_warn_hat.to_excel(writer, sheet_name='warn_hat')


#%% Disply the Run Time

# This line display the run time in minutes
print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


