# -*- coding: utf-8 -*-
"""
This program produces Figures A.12 and A.13 "Advance Layoff Notices and
Aggregate Job Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    fcast_eval_module.py : module with functions to compute and evaluate forecasts
    warn_factors_rt.xlsx : Excel file with pseudo real-time WARN factors and
                           values of WARN hat from Equation (4)
    labor_rates_rt.xlsx : Excel file with real-time uneployment rates and 
                          labor market flow rates
    uspriv_alfred.xlsx : Excel file with real-time private employment levels
    icsa_fred.xlsx : Excel file holding monthly averages of initial
                     unemployment insurance claims

This program saves
    forecasts_2020.pdf : Figure A.12 in the appendix
    forecasts_2020_pe.pdf : Figure A.13 in the appendix
"""

#%% Import Packages and Functions

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Pandas, NumPy and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# Functions for making and evaluating forecasts
from modules.fcast_eval_module import fcast_var_func, fcast2020_var_func, fcast_bn_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the program

# First date that WARN factor is available
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# First date for forecast evaluation
str_first = '20200101'  # Format: 20080101 corresponds to January 1, 2008

# Last date for forecast evaluation
str_last = '20211130'   # Format: 20191231 corresponds to December 31, 2019

# Number of variables in the VAR including WARN factor or WARN hat
nvars = 6

# Number of lags in the forecasting models
nlags = 2
 
# Sample size for estimating the forecasting models
# Chosen to match the rolling sample used for the 2008-2019 forecasts
nsample = 12*11 + 5

# Maximum horizon to forecast
hmax = 0


#%% Load the Data

# Change in warn factors
df_factor = pd.read_excel('warn_factors_rt.xlsx', 'factors', index_col=0)
df_factor = df_factor.diff()
df_factor = df_factor.loc[str_warn:, :]

# Change in warn_hat factors
df_warn = pd.read_excel('warn_factors_rt.xlsx', 'warn_hat', index_col=0)
df_warn = df_warn.diff()
df_warn = df_warn.loc[str_warn:, :]

# Unemployment rate
df_urate = pd.read_excel('labor_rates_rt.xlsx', 'urate', index_col=0)
df_urate = df_urate.loc[str_warn:, :]

# Change in log unemployment rate
df_uchange = np.log(pd.read_excel('labor_rates_rt.xlsx', 'urate', index_col=0))
df_uchange = df_uchange.diff()
df_uchange = df_uchange.loc[str_warn:, :]

# Log finding rate
df_grate = np.log(pd.read_excel('labor_rates_rt.xlsx', 'grate', index_col=0))
df_grate = df_grate.loc[str_warn:, :]

# Log separation rate
df_srate = np.log(pd.read_excel('labor_rates_rt.xlsx', 'srate', index_col=0))
df_srate = df_srate.loc[str_warn:, :]

# Number of months to truncate
KK = len(pd.date_range(start='20030101', end=str_first, freq='MS')) - 1

# Level and change of private employment
path_emp_data = os.path.join(path_dir, 'data_input', 'uspriv_alfred.xlsx')
df_emp = pd.read_excel(path_emp_data, 'Sheet1', index_col=0)
df_demp = df_emp.diff()
df_emp = df_emp.loc[str_warn:, :]
df_demp = df_demp.loc[str_warn:, :]
df_demp = df_demp.iloc[:, KK:]

# Initial claims
path_claims_data = os.path.join(path_dir, 'data_input', 'icsa_alfred.xlsx')
df_claim = np.log(pd.read_excel(path_claims_data, 'Sheet2', index_col=0))
df_claim = df_claim.loc[str_warn:, :]
df_claim = df_claim.iloc[:, KK:]


#%% Compute the Forecasts

# Pseudo real-time dates that forecasts are produced for
fcast_index = pd.date_range(start=str_first, end=str_last, freq='MS')

# Number of forecasts to compute
nfcast = len(fcast_index)

# Storage arrays for finding rate forecasts
g_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
g_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for separation rate forecasts
s_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
s_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Storage arrays for unemployment rate forecasts
u_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
u_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat
u_fcast_bn = np.zeros((nfcast, hmax+1))     # forecasts from BN model without WARN
u_fcast_bnw = np.zeros((nfcast, hmax+1))    # forecasts from BN model with WARN_hat

# Storage arrays for employment change forecasts
e_fcast_var = np.zeros((nfcast, hmax+1))    # forecasts from VAR without WARN
e_fcast_varw = np.zeros((nfcast, hmax+1))   # forecasts from VAR with WARN_hat

# Loop to compute the forecasts
for loop in range(nfcast):
    # Number of observations in this loop
    nobs = len(df_uchange.iloc[:, loop].dropna())
    
    # Most recent value of unemployment rate available to forecaster
    u_init = np.zeros(1)
    u_init[0] = df_urate.iloc[nobs-1, loop]
    
    # Variables for VAR forecasts
    var_data = np.zeros((nsample, nvars))
    var_data[:, 0] = df_grate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 1] = df_srate.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 2] = df_uchange.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 3] = df_claim.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 4] = df_demp.iloc[nobs-nsample:nobs, loop].to_numpy()
    var_data[:, 5] = df_warn.iloc[nobs-nsample:nobs, loop].dropna()
    
    # Construct the arrays of dummy variables
    if loop == 3:
        dummy = np.zeros((nsample, 1))
        dummy[nsample-loop+2, 0] = 1
    elif loop == 4:
        dummy = np.zeros((nsample, 2))
        dummy[nsample-loop+2, 1] = 1
        dummy[nsample-loop+3, 0] = 1
    elif loop >= 5:
        dummy = np.zeros((nsample, 3))
        dummy[nsample-loop+2, 2] = 1
        dummy[nsample-loop+3, 1] = 1
        dummy[nsample-loop+4, 0] = 1
    
    # Forecasts of finding and separation rates from the VAR without WARN
    if loop < 3:
        fcast_temp = \
            fcast_var_func(var_data[:, 0:nvars-1], hmax, nlags)
        
        var_fcast = np.zeros((hmax+1, nvars-1))
        var_fcast[:, :] = fcast_temp
        
        g_fcast_var[loop, :] = np.exp(var_fcast[:, 0])
        s_fcast_var[loop, :] = np.exp(var_fcast[:, 1])
        e_fcast_var[loop, :] = var_fcast[:, 4]
    elif loop >= 3:
        fcast_temp = \
            fcast2020_var_func(var_data[:, 0:nvars-1], dummy, hmax, nlags)
        
        var_fcast = np.zeros((hmax+1, nvars-1))
        var_fcast[:, :] = fcast_temp
        
        g_fcast_var[loop, :] = np.exp(var_fcast[:, 0])
        s_fcast_var[loop, :] = np.exp(var_fcast[:, 1])
        e_fcast_var[loop, :] = var_fcast[:, 4]
    
    # Forecasts of unemployment rate from the VAR without WARN
    temp = np.zeros(hmax+1)
    temp[:] = np.cumsum(var_fcast[:, 2])
    u_fcast_var[loop, :] = np.exp(temp + np.log(u_init[0]))
    
    # Forecasts of finding and separation rates from the VAR with WARN_hat
    if loop < 3:
        fcast_temp = fcast_var_func(var_data, hmax, nlags)
        
        var_fcast = np.zeros((hmax+1, nvars))
        var_fcast[:, :] = fcast_temp       
        
        g_fcast_varw[loop, :] = np.exp(var_fcast[:, 0])
        s_fcast_varw[loop, :] = np.exp(var_fcast[:, 1])
        e_fcast_varw[loop, :] = var_fcast[:, 4]
    elif loop >= 3:
        fcast_temp = fcast2020_var_func(var_data, dummy, hmax, nlags)
        
        var_fcast = np.zeros((hmax+1, nvars))
        var_fcast[:, :] = fcast_temp
        
        g_fcast_varw[loop, :] = np.exp(var_fcast[:, 0])
        s_fcast_varw[loop, :] = np.exp(var_fcast[:, 1])
        e_fcast_varw[loop, :] = var_fcast[:, 4]
    
    # Forecasts of unemployment rate from the VAR with WARN_hat
    temp = np.zeros(hmax+1)
    temp[:] = np.cumsum(var_fcast[:, 2])
    u_fcast_varw[loop, :] = np.exp(temp + np.log(u_init[0]))
    
    # Forecasts unemployment rate from BN model without WARN
    u_fcast_bn[loop, :] = fcast_bn_func(g_fcast_var[loop, :],
                                        s_fcast_var[loop, :], u_init[0], hmax)
    
    # Forecasts unemployment rate from BN model with WARN_hat
    u_fcast_bnw[loop, :] = fcast_bn_func(g_fcast_varw[loop, :],
                                         s_fcast_varw[loop, :], u_init[0], hmax)
    
    print(loop)


#%% Compute Statistics for the Appendix

# The unemployment rate outcomes
u_actual = df_urate.loc[str_first:str_last, :]
u_actual = u_actual.iloc[:, -1]

# The change in employment outcomes
e_actual = df_demp.loc[str_first:str_last, :]
e_actual = e_actual.iloc[:, -1]

# Unemployment rate forecast errors
du_bn = np.zeros(nfcast)
du_bn[:] = 100*(u_actual.to_numpy() - u_fcast_bn[:, 0])
du_bnw = np.zeros(nfcast)
du_bnw[:] = 100*(u_actual.to_numpy() - u_fcast_bnw[:, 0])
du_var = np.zeros(nfcast)
du_var[:] = 100*(u_actual.to_numpy() - u_fcast_var[:, 0])
du_varw = np.zeros(nfcast)
du_varw[:] = 100*(u_actual.to_numpy() - u_fcast_varw[:, 0])

# Employment change forecast errors
de_var = np.zeros(nfcast)
de_var[:] = e_actual.to_numpy() - e_fcast_var[:, 0]
de_varw = np.zeros(nfcast)
de_varw[:] = e_actual.to_numpy() - e_fcast_varw[:, 0]

# Unemployment rate RMSPEs
rmspe_bn = np.sqrt(np.inner(du_bn[0:12], du_bn[0:12]))
rmspe_bnw = np.sqrt(np.inner(du_bnw[0:12], du_bnw[0:12]))
rmspe_var = np.sqrt(np.inner(du_var[0:12], du_var[0:12]))
rmspe_varw = np.sqrt(np.inner(du_varw[0:12], du_varw[0:12]))

# Employment change RMSPEs
rmspe_e_var = np.sqrt(np.inner(de_var[0:12], de_var[0:12]))
rmspe_e_varw = np.sqrt(np.inner(de_varw[0:12], de_varw[0:12]))


#%% Plot Unemployment Rate Figure

# Figure and axes instances
fig, ax = plt.subplots(2, 1, figsize=(9, 7))

# The plot
ax[0].plot(fcast_index, 100*u_actual, color='black', linewidth=3)
ax[0].plot(fcast_index, 100*u_fcast_bn[:, 0], color=(0.25, 0.25, 0.25),
           marker='x', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='-', linewidth=1.5)
ax[0].plot(fcast_index, 100*u_fcast_bnw[:, 0], color=(0.25, 0.25, 0.25),
           marker='o', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='-', linewidth=1.5)
ax[0].plot(fcast_index, 100*u_fcast_var[:, 0], color=(0.55, 0.55, 0.55),
           marker='^', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='--', linewidth=1.5)
ax[0].plot(fcast_index, 100*u_fcast_varw[:, 0], color=(0.55, 0.55, 0.55),
           marker='s', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='--', linewidth=1.5)
ax[0].set_xlim(fcast_index[0], fcast_index[-1])
ax[0].set_ylim(0, 26)
ax[0].set_ylabel('percent', fontsize=12)
ax[0].set_title('Realized Unemployment and Forecasts', fontsize=12)
ax[0].legend(['Unemp', 'SSUR-baseline', 'SSUR-w', 'VAR-5', 'VAR-w'])

ax[1].plot(fcast_index, np.abs(du_bn), color=(0.25, 0.25, 0.25),
           marker='x', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='-', linewidth=1.5)
ax[1].plot(fcast_index, np.abs(du_bnw), color=(0.25, 0.25, 0.25),
           marker='o', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='-', linewidth=1.5)
ax[1].plot(fcast_index, np.abs(du_var), color=(0.55, 0.55, 0.55),
           marker='^', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='--', linewidth=1.5)
ax[1].plot(fcast_index, np.abs(du_varw), color=(0.55, 0.55, 0.55),
           marker='s', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='--', linewidth=1.5)
ax[1].set_xlim(fcast_index[0], fcast_index[-1])
ax[1].set_ylim(0, 15)
ax[1].set_ylabel('percent', fontsize=12)
ax[1].set_title('Absolute Forecast Errors', fontsize=12)
ax[1].legend(['SSUR-baseline', 'SSUR-w', 'VAR-5', 'VAR-w'])

fig.tight_layout()

# Save the figure
plt.savefig('forecasts_2020.pdf', bbox_inches='tight')


#%% Plot Employment Rate Figure

# Figure and axes instances
fig, ax = plt.subplots(2, 1, figsize=(9, 7))

# The plot
ax[0].plot(fcast_index, e_actual, color='black', linewidth=3)
ax[0].plot(fcast_index, e_fcast_var[:, 0], color=(0.55, 0.55, 0.55),
           marker='^', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='--', linewidth=1.5)
ax[0].plot(fcast_index, e_fcast_varw[:, 0], color=(0.55, 0.55, 0.55),
           marker='s', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='--', linewidth=1.5)
ax[0].set_xlim(fcast_index[0], fcast_index[-1])
ax[0].set_ylabel('percent', fontsize=12)
ax[0].set_title('Realized Private Employment Changes and Forecasts', fontsize=12)
ax[0].legend(['Priv Emp', 'VAR-5', 'VAR-w'])

ax[1].plot(fcast_index, np.abs(de_var), color=(0.55, 0.55, 0.55),
           marker='^', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='--', linewidth=1.5)
ax[1].plot(fcast_index, np.abs(de_varw), color=(0.55, 0.55, 0.55),
           marker='s', markerfacecolor='None', markeredgewidth=1.5,
           linestyle='--', linewidth=1.5)
ax[1].set_xlim(fcast_index[0], fcast_index[-1])
ax[1].set_ylim(0, 20000)
ax[1].set_ylabel('percent', fontsize=12)
ax[1].set_title('Absolute Forecast Errors', fontsize=12)
ax[1].legend(['VAR-5', 'VAR-w'])

fig.tight_layout()

# Save the figure
plt.savefig('forecasts_2020_pe.pdf', bbox_inches='tight')


#%% Disply the Run Time

# This line display the run time in minutes
print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


