# -*- coding: utf-8 -*-
"""
This program computes real-time labor market flow rates and unemploymet rates,
used in "Advance Layoff Notices and Aggregate Job Loss" by Pawel M. Krolikowski
and Kurt G. Lunsford.

The flow rates are computed as in Barnichon and Nekarda (2012, BPEA).
The job separation rate is computed with the method of bisection.

This program calls the following files:
    ce16ov_alfred.xlsx : Excel file with real-time employment levels
    unemploy_alfred.xlsx : Excel file with real-time unemployment levels
    uemplt5_alfred.xlsx : Excel file with real-time levels of number of people
                          unemployed for less than 5 weeks

The main output of this program is
    df_urate : a Pandas DataFrame holding real-time unemployment rates
    df_g : a Pandas DataFrame holding real-time job finding rates
    df_s : a Pandas DataFrame holding real-time job separating rates
These DataFrames are saved to labor_rates_rt.xlsx.
"""


#%% Import Packages and Functions

# Pandas and NumPy
import pandas as pd
import numpy as np

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the Program

# Maximum number of loops when using the method of bisection
maxloops = 2000

# Error tolerance for method of bisection
tol = 1e-10


#%% Read in the Data

# Get replication files directory
import os
path_dir_appdx = os.path.dirname(os.getcwd())
path_dir = os.path.dirname(path_dir_appdx)

# Pandas dataframe for number employed
path_emp_data = os.path.join(path_dir, 'data_input', 'ce16ov_alfred.xlsx')
df_emp_all = pd.read_excel(path_emp_data, 'Sheet1', index_col=0)
df_emp = df_emp_all.loc[:, 'CE16OV_20200110':]

# Pandas dataframe for number unemployed
path_unemp_data = os.path.join(path_dir, 'data_input', 'unemploy_alfred.xlsx')
df_unemp_all = pd.read_excel(path_unemp_data, 'Sheet1', index_col=0)
df_unemp = df_unemp_all.loc[:, 'UNEMPLOY_20200110':]

# Pandas dataframe for number unemployed less than 5 weeks
path_unemp5_data = os.path.join(path_dir, 'data_input', 'uemplt5_alfred.xlsx')
df_unemp5_all = pd.read_excel(path_unemp5_data, 'Sheet1', index_col=0)
df_unemp5 = df_unemp5_all.loc[:, 'UEMPLT5_20200110':]

# Multiply number unemployed less than 5 weeks by 1.1549
df_unemp5 = df_unemp5.mul(1.1549)


#%% Convert to NumPy Arrays

# Number of observations and vintages in the sample
nobs, nvint = df_emp.shape

# Convert number employed to numpy array
e = np.zeros((nobs, nvint))
e[:, :] = df_emp.to_numpy()

# Convert number unemployed to numpy array
u = np.zeros((nobs, nvint))
u[:, :] = df_unemp.to_numpy()

# Convert number unemployed less than 5 weeks to numpy array
u5 = np.zeros((nobs, nvint))
u5[:, :] = df_unemp5.to_numpy()


#%% Compute Some Labor Market Data and Store to DataFrames

# Labor force
lf = np.zeros((nobs, nvint))
lf[:, :] = e[:, :] + u[:, :]

# Unemployment rate
urate = np.zeros((nobs, nvint))
urate[:, :] = np.divide(u, lf)

# Save unemployment rate to dataframe
temp_str = df_emp.columns[0]
temp_str = temp_str[6:]
col_str = ['urate' + temp_str]
for jj in range(nvint-1):
    temp_str = df_emp.columns[jj+1]
    temp_str = temp_str[6:]
    col_str.append('urate' + temp_str)
df_urate = pd.DataFrame(urate, index=df_emp.index, columns=col_str)

# Job finding probability
G = np.zeros((nobs-1, nvint))
G[:, :] = (np.ones((nobs-1, nvint)) 
           - np.divide((u[1:nobs, :] - u5[1:nobs, :]), u[0:nobs-1, :]))

# Job finding rate
g = np.zeros((nobs-1, nvint))
g[:, :] = -np.log(np.ones((nobs-1, nvint))  - G[:, :])

# Save job finding rate to dataframe
temp_str = df_emp.columns[0]
temp_str = temp_str[6:]
col_str = ['grate' + temp_str]
for jj in range(nvint-1):
    temp_str = df_emp.columns[jj+1]
    temp_str = temp_str[6:]
    col_str.append('grate' + temp_str)
df_g = pd.DataFrame(index=df_emp.index, columns=col_str)
df_g.iloc[1:nobs, :] = g[:, :]


#%% Compute Job Separation Rate with Method of Bisection

# Storage dataframe
temp_str = df_emp.columns[0]
temp_str = temp_str[6:]
col_str = ['srate' + temp_str]
for jj in range(nvint-1):
    temp_str = df_emp.columns[jj+1]
    temp_str = temp_str[6:]
    col_str.append('srate' + temp_str)
df_s = pd.DataFrame(index=df_g.index, columns=col_str)

for loop_outer in range(nvint):
    # Number of obervations for this loop
    ntemp = len(df_urate.iloc[:, loop_outer].dropna())
    
    # Temporary variables for this loop
    u_temp = np.zeros(ntemp)
    u_temp[:] = u[0:ntemp, loop_outer]
    
    u5_temp = np.zeros(ntemp)
    u5_temp[:] = u5[0:ntemp, loop_outer]
    
    lf_temp = np.zeros(ntemp)
    lf_temp[:] = lf[0:ntemp, loop_outer]
    
    g_temp = np.zeros(ntemp-1)
    g_temp[:] = g[0:ntemp-1, loop_outer]
    
    # Guesses for method of bisection
    min_guess = np.zeros(ntemp-1)
    max_guess = np.ones(ntemp-1)
    
    # Method of bisection
    for loop_inner in range(maxloops):
        # Job separation probability based on guess values
        S_temp = np.zeros(ntemp-1)
        S_temp[:] = (min_guess + max_guess)/2
        
        # Job separation rate based on guess values
        s_temp = np.zeros(ntemp-1)
        s_temp[:] = -np.log(np.ones(ntemp-1) - S_temp)
        
        # Compute the equation errors
        temp1 = np.zeros(ntemp-1)
        temp1[:] = np.exp(-g_temp - s_temp)
        
        temp2 = np.zeros(ntemp-1)
        temp2[:] = np.multiply((np.ones(ntemp-1) - temp1), s_temp)
        
        temp3 = np.zeros(ntemp-1)
        temp3[:] = np.multiply(temp2, lf_temp[0:ntemp-1])
        
        errors = np.zeros(ntemp-1)
        errors[:] = (u_temp[1:ntemp] - np.divide(temp3, (g_temp + s_temp))
                     - np.multiply(temp1, u_temp[0:ntemp-1]))
        
        # End loop if the magnitude of the errors is less than the tolerance
        if np.sum(np.abs(errors) > tol) == 0:
            print('Solution found on loop ' + str(loop_inner) + '.')
            break
        
        # Update the guesses
        mask_pos = (errors > 0)
        min_guess[mask_pos] = S_temp[mask_pos]
        
        mask_neg = (errors < 0)
        max_guess[mask_neg] = S_temp[mask_neg]
    
    # Store the results
    df_s.iloc[1:ntemp, loop_outer] = s_temp


#%% Save the Output

str_file = 'labor_rates_rt.xlsx'
with pd.ExcelWriter(str_file, datetime_format='YYYY-MM') as writer:
    df_urate.to_excel(writer, sheet_name='urate')
    df_g.to_excel(writer, sheet_name='grate')
    df_s.to_excel(writer, sheet_name='srate')


#%% Disply the Run Time

# This line display the run time in minutes
print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')

