# -*- coding: utf-8 -*-
"""
Module with functions to estimate a dynamic factor model (DFM).

Estimation uses the expectation maximization (EM) algorithm in Banbura and
Modugno (2014) and Kalman filter and smoother equations in the appendix of
Shumway and Stoffer (1982).  

This module was written for "Advance Layoff Notices and Aggregate Job Loss" by
Pawel M. Krolikowski and Kurt G. Lunsford.

Functions
---------
make_autocov : computes the autocovariances of an autoregressive process
make_guess : computes the initial guess of the DFM parameters
kalman_func : kalman filter and smoother
maximize_func : computes the maximization step of the EM algorithm
em_func : executes the EM algorithm

Notes
-----
The DFM assumes a vector of data z_t follows

    z_t = Lambda f_t + e_t

in which f_t is the unobserved factor that is a scalar process and Lambda is a
vector of factor loadings.  e_t is an iid normal vector process with mean zero
and diagonal covariance matrix.  The factor is an AR(p) process that follows

    f_t = A_1 f_{t-1} + .. A_p f_{t-p} + u_t

in which u_t is an iid normal scalar process.

When p > 1 lags are used in the state equation, the state becomes S_t = 
[y_t, y_{t-1},...,y_{t-p+1}]'.  Then, the obseration equation is

    z_t = Lambda S_t + e_t

in which the first column of Lambda has non-zero values and all other columns
contain zeros.  The state equation becomes

    S_t = A S_{t-1} + U_t

in which A is the companion form matrix of A_1,...,A_p and U_t is a vector
process with u_t as the first element and zeros for all other elements.

The EM algorithm estimates Lambda, A_1,...,A_p, the covariance matrix of e_t,
and the variance of u_t.  Given these parameters, the Kalman filter and
smoother can the provide smoothed estimates of f_t.
"""

import numpy as np
from numba import njit


@njit
def make_autocov(Ain, Qin):
    """
    Function to compute the autocovariances of an AR(r) process.
    """
    
    # Number of lags in the AR process
    rr = Ain[:, 0].size
    
    # Output array
    mat_out = np.zeros((rr, rr))
    
    # Compute the covariance matrix    
    if rr == 1:
        # Compute the variance
        var = Qin[0, 0]/(1 - Ain[0, 0]*Ain[0, 0])
        
        # Store the variance in a matrix
        mat_out[0, 0] = var
    else:
        # Construct some useful matrices
        mat_1 = np.zeros((rr, rr))
        for nnn in range(rr):
            mat_1[nnn:, nnn] = Ain[0, 0:rr-nnn]
        
        mat_2 = np.zeros((rr, rr))
        for nnn in range(rr-1):
            mat_2[nnn, 1:rr-nnn] = Ain[0, 1+nnn:rr]
        
        mat_3 = np.zeros((rr, rr))
        mat_3[:, :] = np.eye(rr) - Ain @ (mat_1 + mat_2)
        
        # Solve for the autocovariance structure
        mat_inv = np.empty((rr, rr))
        mat_inv[:, :] = np.linalg.inv(mat_3)
        
        covar = np.zeros(rr)
        covar[:] = mat_inv[:, 0]*Qin[0, 0]
        
        # Convert autocovariances to a matrix
        mat_out[:, 0] = covar[:]
        for nnn in range(rr-1):
            mat_out[nnn+1:rr, nnn+1] = covar[0:rr-nnn-1]
            mat_out[nnn, nnn+1:rr] = covar[1:rr-nnn]
    
    return mat_out


def make_guess(zz, pp):
    """
    Function to compute the initial guess of DFM parameters.
    """
    # Dimensions of the data
    TT, NN = zz.shape
    
    # Averages of the data
    avg = np.zeros((NN, 1))
    avg[:, 0] = np.nanmean(zz, axis=0)
    
    # Variances of the data
    var = np.zeros((NN, 1))
    var[:, 0] = np.nanvar(zz, axis=0)
    
    # Standardize the data: de-mean, unit variance, set NaNs to zero
    xx = np.zeros((TT, NN))
    for jj in range(NN):
        mask = (np.isnan(zz[:, jj]) == 0)
        xx[mask, jj] = (zz[mask, jj] - avg[jj, 0])/np.sqrt(var[jj, 0])
    
    # Principal components to guess the factor and Lambda
    S = (xx.T @ xx)/TT          # covariance matrix of the data
    D, P = np.linalg.eig(S)     # eigenvalues and eigenvectors
    
    ind = np.argmax(D)          # index of largest eigenvalue
    D = D[ind]                  # select the largest eigenvalue
    P = P[:, ind:ind+1]         # select the corresponding eigenvector
    
    FF = np.zeros((TT, 1))
    FF = (xx @ P)/np.sqrt(D)    # guess of the factor
    
    Lam = np.zeros((NN, pp))
    Lam[:, 0:1] = P*np.sqrt(D)  # guess Lambda
    
    # Make sure the sign of the factor is positively correlated with the data
    check = np.zeros(TT)
    check[:] = np.mean(xx, axis=1)
    check = check - np.mean(check)
    if np.inner(check, FF[:, 0]) < 0:
        FF[:, 0] = -FF[:, 0]
        Lam[:, 0] = -Lam[:, 0]
    
    # Compute the AR parameters of the state equations
    lhs = np.zeros((TT-pp, 1))
    lhs = FF[pp:TT, :]
    
    rhs = np.zeros((TT-pp, pp))
    for jj in range(pp):
        rhs[:, jj:jj+1] = FF[pp-jj-1:TT-jj-1, :]
    
    beta = np.zeros((pp, 1))
    beta[:, :] = np.linalg.inv(rhs.T @ rhs) @ (rhs.T @ lhs)
    
    # Construct guess of A matrix in companion form
    AA = np.zeros((pp, pp))
    if pp == 1:
        AA[0, 0] = beta[0, 0]
    elif pp > 1:
        AA[0, :] = beta[0:pp, 0]
        AA[1:pp, 0:pp-1] = np.eye(pp-1)
    
    # Compute guess of variance matrix of the innovations of the state equation
    u = np.zeros((TT-pp, 1))
    u[:, :] = lhs - rhs @ beta
    
    QQ = np.zeros((pp, pp))
    QQ[0, 0] = (u.T @ u)/(TT - pp)
    
    # Compute the variance of the factor
    covar = make_autocov(AA, QQ)
    f_var = covar[0, 0]
    
    # Normalize the parameters for that factor has variance of 1
    QQ[0, 0] = QQ[0, 0]/f_var
    Lam[:, 0] = Lam[:, 0]*np.sqrt(f_var)
    FF[:, 0] = FF[:, 0]/np.sqrt(f_var)
    
    # Variance of innovations of measurement equation
    ee = np.zeros((TT, NN))
    ee[:, :] = xx - FF @ Lam[:, 0:1].T
    
    Rtemp = np.zeros((NN, NN))
    Rtemp[:, :] = (ee.T @ ee)/TT
    
    RR = np.zeros((NN, NN))
    for nnn in range(NN):
        RR[nnn, nnn] = Rtemp[nnn, nnn]
    
    return FF, AA, QQ, Lam, RR, xx, avg, var


@njit
def kalman_func(s0, P0, ZZ, LL, RR, AA, QQ, WW):
    """
    The Kalman filter and smoother.
    
    This function is used to compute the expectation step of the EM algorithm.
    """
    
    # Dimensions of the data
    TT, NN = ZZ.shape
    
    # Number of states
    rr = AA[:, 0].size
    
    # Storage arrays for the Kalman filter
    s_pred = np.zeros((rr, TT))         # predictions of the state
    s_filt = np.zeros((rr, TT+1))       # filtered state
    P_pred = np.zeros((rr, rr, TT))     # prediction of variance of state
    P_filt = np.zeros((rr, rr, TT+1))   # filtered variance of state
    Jmats = np.zeros((rr, rr, TT))      # matrices for smoothing
    
    # Initial values
    s_filt[:, 0] = s0[:, 0]
    P_filt[:, :, 0] = P0
    
    #Kalman filter
    llike = 0
    for jjj in range(TT):
        
        # Adjust the Lambda matrix to handle missing observations
        tmp = np.zeros(NN)
        for nnn in range(NN):
            tmp[nnn] = WW[nnn, nnn, jjj]
        mask = (tmp == 1)
        count = int(np.sum(tmp))
        
        Ladj = np.zeros((count, 1))
        Ladj[:, 0] = LL[mask, 0]
        
        # Adjust the R matrix to handle missing observations
        Rtemp = np.zeros((count, NN))
        Rtemp[:, :] = RR[mask, :]
        
        Radj = np.zeros((count, count))
        Radj[:, :] = Rtemp[:, mask]
        
        # Call filtered state from previous loop
        s_t1t1 = np.zeros((rr, 1))
        s_t1t1[:, 0] = s_filt[:, jjj]
        
        # Prediction of the state: s_{t|t-1} = E(s_t|z_1,...,z_{t-1})
        s_tt1 = np.zeros((rr, 1))
        s_tt1[:, :] = AA @ s_t1t1
        s_pred[:, jjj] = s_tt1[:, 0]
        
        # Call filtered state variance from previous loop
        P_t1t1 = np.zeros((rr, rr))
        P_t1t1[:, :] = P_filt[:, :, jjj]
        
        # Prediction of state variance: P_{t|t-1} = Var(s_t|z_1,...,z_{t-1})
        PAmat = np.zeros((rr, rr))
        PAmat[:, :] = P_t1t1 @ AA.T
        
        P_tt1 = np.zeros((rr, rr))
        P_tt1[:, :] = AA @ PAmat + QQ
        P_pred[:, :, jjj] = P_tt1[:, :]
        
        # Matrix for Kalman smoother
        Jtemp = np.zeros((rr, rr))
        Jtemp[:, :] = np.linalg.solve(P_tt1, PAmat.T)
        Jmats[:, :, jjj] = Jtemp.T
        
        # Variance of one-step ahead forecast
        H = np.zeros((count, count))
        H[:, :] = P_tt1[0, 0]*(Ladj @ Ladj.T) + Radj
                
        # Determinant of H
        Hlower = np.zeros((count, count))
        Hlower[:, :] = np.linalg.cholesky(H)
        
        Hdet = 1
        for nnn in range(count):
            Hdet = Hdet*Hlower[nnn, nnn]
        Hdet = Hdet*Hdet
        
        # Inverse of H
        Rinv = np.zeros((count, count))
        RLmat = np.zeros((count, 1))
        for nnn in range(count):
            Rinv[nnn, nnn] = 1/Radj[nnn, nnn]
            RLmat[nnn, 0] = Rinv[nnn, nnn]*Ladj[nnn, 0]
        
        Hinv = np.zeros((count, count))
        Hinv_temp = (1 + P_tt1[0, 0]*(Ladj.T @ RLmat))
        Hinv[:, :] = Rinv - P_tt1[0, 0]*(RLmat @ RLmat.T)/Hinv_temp[0, 0]
        
        # A useful matrix
        PLmat = np.zeros((rr, count))
        for nnn in range(rr):
            PLmat[nnn, :] = P_tt1[nnn, 0]*Ladj.T
        
        # Kalman gain
        K = np.zeros((rr, count))
        K[:, :] = PLmat @ Hinv
        
        # One step ahead forecast error: v_t = y_t - E(y_t|z_1,...,z_{t-1})
        vv = np.zeros((count, 1))
        vv[:, 0:1] = ZZ[jjj:jjj+1, mask].T - Ladj*s_tt1[0, 0]
        
        # Updating step:
        # s_{t|t} = E(s_t|z_1,...,z_t) and P_{t|t} = Var(s_t|z_1,...,z_t)
        s_filt[:, jjj+1:jjj+2] = s_tt1 + K @ vv
        P_filt[:, :, jjj+1] = P_tt1 - K @ PLmat.T
        
        # Log-likelihood
        temp = np.zeros((1, 1))
        temp[:, :] = (vv.T @ Hinv) @ vv
        llike = llike - (np.log(Hdet) + temp[0, 0])/2
    
    # Storage arrays for the Kalman smoother
    s_smooth = np.zeros((rr, TT+1))         # smoothed state
    P_smooth = np.zeros((rr, rr, TT+1))     # smothed variance of state
    Pcov_smooth = np.zeros((rr, rr, TT))    # smoothed autocovariance of state
    
    # Last values
    s_smooth[:, TT:TT+1] = s_filt[:, TT:TT+1]
    P_smooth[:, :, TT] = P_filt[:, :, TT]
    
    P_tt = np.zeros((rr, rr))
    P_tt[:, :] = P_filt[:, :, TT-1]
    
    A1 = np.zeros((1, rr))
    A1[0, :] = AA[0, :]
    Pcov_smooth[:, :, TT-1] = (AA - K @ (Ladj @ A1)) @ P_tt
    
    # Some matrices to be used with numba @jit decorator
    J1_temp = np.zeros((rr, rr))
    J2_temp = np.zeros((rr, rr))
    
    # Kalman smoother
    for jjj in range(TT-1):
        # Define the index
        kk = TT - jjj - 1
        
        # Collect some matrices
        J1_temp[:, :] = Jmats[:, :, kk]
        J2_temp[:, :] = Jmats[:, :, kk-1]
        P_tt[:, :] = P_filt[:, :, kk]
        
        # State smoother: E(s_t|z_1,...,z_T)
        temp = np.zeros((rr, 1))
        temp[:, :] = J1_temp @ (s_smooth[:, kk+1:kk+2] - s_pred[:, kk:kk+1])
        s_smooth[:, kk:kk+1] = s_filt[:, kk:kk+1] + temp
        
        # Variance smoother: Var(s_t|z_1,...,z_T)
        temp = np.zeros((rr, rr))
        temp[:, :] = J1_temp @ (P_smooth[:, :, kk+1] - P_pred[:, :, kk])
        P_smooth[:, :, kk] = P_tt + temp @ J1_temp.T
        
        # Autocovariance smoother: Cov(s_t, s_{t-1}|z_1,...,z_T)
        temp = np.zeros((rr, rr))
        temp[:, :] = P_tt + J1_temp @ (Pcov_smooth[:, :, kk] - AA @ P_tt)
        Pcov_smooth[:, :, kk-1] = temp @ J2_temp.T
    
    # Last loop of the Kalman smoother
    J1_temp[:, :] = Jmats[:, :, 0]
    P_tt[:, :] = P_filt[:, :, 0]
    
    temp = np.zeros((rr, 1))
    temp[:, :] = J1_temp @ (s_smooth[:, 1:2] - s_pred[:, 0:1])
    s_smooth[:, 0:1] = s_filt[:, 0:1] + temp
    
    temp = np.zeros((rr, rr))
    temp[:, :] = J1_temp @ (P_smooth[:, :, 1] - P_pred[:, :, 0]) @ J1_temp.T
    P_smooth[:, :, 0] = P_tt + temp
    
    return s_smooth, P_smooth, Pcov_smooth, llike


@njit
def maximize_func(XX, Wmats, RR, FF, PP, PPcov):
    """
    Function for maximization step in the EM algorithm.
    """

    # Dimensions of the data
    TT, NN = XX.shape
    
    # Number of states
    rr = FF[:, 0].size
    
    # Output arrays
    Aupdate = np.zeros((rr, rr))
    Qupdate = np.zeros((rr, rr))
    Lupdate = np.zeros((NN, rr))
    Rupdate = np.zeros((NN, NN))
    
    # Sums to compute updates of A and Q matrices
    ff0 = np.zeros((rr, TT))
    ff0[:, :] = FF[:, 0:TT]
    
    ff1 = np.zeros((1, TT))
    ff1[0, :] = FF[0, 1:TT+1]
    
    sum1 = np.zeros((rr, rr))
    sum2 = np.zeros((1, rr))
    sum3 = np.zeros((1, 1))
    
    sum1[:, :] = ff0 @ ff0.T
    sum2[:, :] = ff1 @ ff0.T        
    sum3[:, :] = ff1 @ ff1.T
    
    for jjj in range(TT):
        sum1 = sum1 + PP[:, :, jjj]
        sum2 = sum2 + PPcov[0:1, :, jjj]
        sum3 = sum3 + PP[0:1, 0:1, jjj+1]
    
    # New value of A matrix
    Atemp = np.zeros((rr, 1))
    Atemp[:, :] = np.linalg.solve(sum1, sum2.T)
    
    if rr == 1:
        Aupdate[0, 0] = Atemp[0, 0]
    elif rr > 1:
        Aupdate[1:rr, 0:rr-1] = np.eye(rr-1)
        Aupdate[0, :] = Atemp[:, 0]
    
    # New value of Q matrix
    Qtemp = (sum3 - Atemp.T @ sum2.T)/TT
    Qupdate[0, 0] = Qtemp[0, 0]
    
    # New value of Lambda
    Lsum1 = np.zeros((NN, NN))
    Lsum2 = np.zeros((NN, 1))
    for jjj in range(TT):
        # Arrays used for computation
        Ltmp = np.zeros((NN, 1))
        
        # Loop for computation
        for nnn in range(NN):
            if Wmats[nnn, nnn, jjj] == 1:
                Ltmp[nnn, 0] = XX[jjj, nnn]
            else:
                Ltmp[nnn, 0] = 0
        
        # Compute the relevant sums
        Lsum1 = Lsum1 + (FF[0, jjj+1]*FF[0, jjj+1] + PP[0, 0, jjj+1])*Wmats[:, :, jjj]
        Lsum2 = Lsum2 + Ltmp*FF[0, jjj+1]
    
    for jjj in range(NN):
        Lsum1[jjj, jjj] = 1/Lsum1[jjj, jjj]
    Lupdate[:, 0:1] = Lsum1 @ Lsum2
    
    # New value of Lambda and the R matrix
    Rsum = np.zeros((NN, NN))
    for jjj in range(TT):
        # Arrays used for computation
        tmp1 = np.zeros((NN, 1))
        tmp2 = np.zeros((NN, 1))
        tmp3 = np.zeros((NN, 1))
        Radj = np.zeros((NN, 1))
        
        # Loop for computation
        for nnn in range(NN):
            if Wmats[nnn, nnn, jjj] == 1:
                Ltmp[nnn, 0] = XX[jjj, nnn]
                tmp1[nnn, 0] = XX[jjj, nnn]*XX[jjj, nnn]
                tmp2[nnn, 0] = XX[jjj, nnn]*Lupdate[nnn, 0]
                tmp3[nnn, 0] = Lupdate[nnn, 0]*Lupdate[nnn, 0]
                Radj[nnn, 0] = 0
            else:
                Ltmp[nnn, 0] = 0
                tmp1[nnn, 0] = 0
                tmp2[nnn, 0] = 0
                tmp3[nnn, 0] = 0
                Radj[nnn, 0] = RR[nnn, nnn]
        
        # Compute the relevant sums
        Rsum = (Rsum + tmp1 - 2*tmp2*FF[0, jjj+1]
                + tmp3*(FF[0, jjj+1]*FF[0, jjj+1] + PP[0, 0, jjj+1]) + Radj)
    
    for jjj in range(NN):
        Rupdate[jjj, jjj] = Rsum[jjj, jjj]/TT
    
    return Aupdate, Qupdate, Lupdate, Rupdate


def em_func(s0, P0, zz, Lin, Rin, Ain, Qin, WW, maxloop, threshold):
    """
    Function to execute the expectation maximization (EM) algorithm.
    """
    
    # Dimensions of the data
    TT, NN = zz.shape
    
    # Number of states
    rr = Ain[:, 0].size
    
    # Initial conditions for Kalman filter
    initF = np.zeros((rr, 1))
    initF[:, 0] = s0[:, 0]
    
    initP = np.zeros((rr, rr))
    initP[:, :] = P0[:, :]
    
    # Initial expectation step via the Kalman filter and smoother
    F, P, Pcov, Loglike = \
        kalman_func(initF, initP, zz, Lin, Rin, Ain, Qin, WW)
    
    # Store the initial log-likelihood
    LL_save = np.zeros(maxloop+1)
    LL_save[0] = Loglike
    
    # Define the arrays for holding updated parameters
    Lnew = np.zeros((NN, rr))
    Rnew = np.zeros((NN, NN))
    Anew = np.zeros((rr, rr))
    Qnew = np.zeros((rr, rr))
    
    # Initial updates
    Lnew[:, :] = Lin[:, :]
    Rnew[:, :] = Rin[:, :]
    Anew[:, :] = Ain[:, :]
    Qnew[:, :] = Qin[:, :]
    
    # EM algorithm loop
    for loop in range(maxloop):
        # Maximization step
        Amax, Qmax, Lmax, Rmax = maximize_func(zz, WW, Rnew, F, P, Pcov)
        
        # Update parameters for the expectation step
        ev = np.linalg.eigvals(Anew)
        if np.max(np.absolute(ev)) < 1:
            # New value of A matrix
            Anew[:, :] = Amax[:, :]
            
            # New value of Q matrix
            Qnew[0, 0] = Qmax[0, 0]
            
            # New value of Lambda
            Lnew[:, :] = Lmax[:, :]
            
            # New value of R matrix
            Rnew[:, :] = Rmax[:, :]
            
            # The variance of the factor
            sig = make_autocov(Anew, Qnew)
            f_var = sig[0, 0]
            
            # Normalize the parameters so that factor has variance of 1
            Qnew[0, 0] = Qnew[0, 0]/f_var
            Lnew[:, :] = Lnew[:, :]*np.sqrt(f_var)
            
            # Initial values for expectation step
            initF[:, 0] = F[:, 0]/np.sqrt(f_var)
            initP[:, :] = sig[:, :]/f_var
        else:
            # New value of Lambda
            Lnew[:, :] = Lmax[:, :]
            
            # New value of R matrix
            Rnew[:, :] = Rmax[:, :]
            
            # Initial values for expectation step
            initF[:, 0] = F[:, 0]
        
        # Expectation step via the Kalman filter and smoother
        F, P, Pcov, Loglike = \
            kalman_func(initF, initP, zz, Lnew, Rnew, Anew, Qnew, WW)
        
        # Store the log-likelihood
        LL_save[loop+1] = Loglike
        
        # Check for convergence
        dist = np.absolute(Loglike - LL_save[loop])
        if dist < threshold:
            break
            
    # Remove log-likelihood values not used
    LL_save = LL_save[0:loop+2]
    
    # Define output arrays
    Lout = np.zeros((NN, rr))
    Rout = np.zeros((NN, NN))
    Aout = np.zeros((rr, rr))
    Qout = np.zeros((rr, rr))
    
    # Populate output arrays
    Lout[:, :] = Lnew[:, :]
    Rout[:, :] = Rnew[:, :]
    Aout[:, :] = Anew[:, :]
    Qout[:, :] = Qnew[:, :]
    
    return F, P, Aout, Qout, Lout, Rout, LL_save, loop

