# -*- coding: utf-8 -*-
"""
Module with functions to make and evaluate forecasts for the paper "Advance
Layoff Notices and Aggregate Job Loss" by Pawel M. Krolikowski and Kurt G.
Lunsford.

Functions
---------
fcast_var_func : computes forecasts with a VAR model
fcast2020_var_func : computes forecasts with a VAR model, allowing the use of
                     dummy variables to account for the COVID-19 pandemic
fcast_bn_func : computes forecasts with Barnichon and Nekarda's flows model
tstat_func : computes the t-stastics for forecast acuracy comparisons
"""

import numpy as np


def fcast_var_func(xx, numh, pp):
    """
    Function to compute VAR forecasts
    """
    
    # Dimensions of the data
    TT, NN = xx.shape
    
    # Collect the right-hand side variables
    rhs = np.ones((TT-pp, NN*pp+1))
    for jjj in range(pp):
        rhs[:, NN*jjj:NN*(jjj+1)] = xx[pp-jjj-1:TT-jjj-1, :]
    
    # Least squares regression
    mat1 = rhs.T @ rhs
    mat2 = rhs.T @ xx[pp:TT, :]
    gamma = np.linalg.solve(mat1, mat2)
    
    # Collect the conditioning variables
    zz = np.ones((1, NN*pp+1))
    for jjj in range(pp):
        zz[0, NN*jjj:NN*(jjj+1)] = xx[TT-jjj-1, :]
    
    # Loop to compute forecasts
    fcast_out = np.zeros((numh+1, NN))
    for jjj in range(numh+1):
        # Compute the forecast
        temp = np.zeros((1, NN))
        temp[:, :] = zz @ gamma
        fcast_out[jjj, :] = temp[0, :]
        
        # Update the conditioning variables
        zz_new = np.ones((1, NN*pp+1))
        zz_new[0, 0:NN] = temp[0, :]
        zz_new[0, NN:NN*pp] = zz[0, 0:NN*(pp-1)]
        
        zz[0, :] = zz_new[0, :]
    
    return fcast_out


def fcast2020_var_func(xx, dd, numh, pp):
    """
    Function to compute VAR forecasts and accound for COVID-19 with dummy variables
    """
    
    # Dimensions of the data
    TT, NN = xx.shape
    
    # Number of dummy variables
    ndum = dd[0, :].size
    
    # Collect the right-hand side variables
    rhs = np.ones((TT-pp, NN*pp+1+ndum))
    for jjj in range(pp):
        rhs[:, NN*jjj:NN*(jjj+1)] = xx[pp-jjj-1:TT-jjj-1, :]
    rhs[:, NN*pp+1:NN*pp+1+ndum] = dd[pp:TT, :]
    
    # Least squares regression
    mat1 = rhs.T @ rhs
    mat2 = rhs.T @ xx[pp:TT, :]
    gamma = np.linalg.solve(mat1, mat2)
    
    # Collect the conditioning variables
    zz = np.ones((1, NN*pp+1))
    for jjj in range(pp):
        zz[0, NN*jjj:NN*(jjj+1)] = xx[TT-jjj-1, :]
    
    # Loop to compute forecasts
    fcast_out = np.zeros((numh+1, NN))
    for jjj in range(numh+1):
        # Compute the forecast without the dummy coefficients
        temp = np.zeros((1, NN))
        temp[:, :] = zz @ gamma[0:NN*pp+1, :]
        fcast_out[jjj, :] = temp[0, :]
        
        # Update the conditioning variables
        zz_new = np.ones((1, NN*pp+1))
        zz_new[0, 0:NN] = temp[0, :]
        zz_new[0, NN:NN*pp] = zz[0, 0:NN*(pp-1)]
        
        zz[0, :] = zz_new[0, :]
    
    return fcast_out


def fcast_bn_func(gfcast, sfcast, u0, numh):
    """
    Function to compute forecasts from Barnichon and Nekarda model
    """
    
    # Storage array for the output
    ufcast_out = np.zeros(numh+1)
    
    # Most recent value of unemployment rate available to forecaster
    u_hat = np.zeros(1)
    u_hat[0] = u0
    
    # Loop to compute forecasts
    for jjj in range(numh+1):
        # Compute u*
        u_star = np.zeros(1)
        u_star[0] = sfcast[jjj]/(gfcast[jjj] + sfcast[jjj])
        
        # Compute beta
        beta = np.zeros(1)
        beta[0] = 1 - np.exp(-gfcast[jjj] - sfcast[jjj])
        
        # Compute the unemployment rate forecast
        ufcast_out[jjj] = beta[0]*u_star[0] + (1 - beta[0])*u_hat[0]
        
        # Store the unemployment rate forecast for the next loop
        u_hat[0] = ufcast_out[jjj]
    
    return ufcast_out


def tstat_func(d1, d2, hh):
    
    # Dimensions of the data
    TT = len(d1)
    
    # The difference in the losses
    diff = np.zeros((TT, 1))
    diff[:, 0] = np.power(d1, 2) - np.power(d2, 2)
    
    # The average difference of the losses
    d_bar = np.zeros(1)
    d_bar[0] = np.mean(diff[:, 0])
    
    # Remove the average from the difference in the losses
    d_tilde = diff - d_bar[0]
    
    # The long-run variance
    lrv = np.zeros((1, 1))
    lrv[:, :] = (d_tilde.T @ d_tilde)/TT
    if hh > 1:
        for iii in range(hh-1):
            cov = np.zeros((1, 1))
            cov[:, :] = (d_tilde[0:TT-iii-1, :].T @ d_tilde[iii+1:TT, :])/TT
            lrv[:, :] = lrv + 2*cov
    
    # Computations depending if lrv is positive
    if lrv[0, 0] > 0:
        # The conventional standard error
        se = np.zeros(1)
        se[0] = np.sqrt(lrv[0, 0]/TT)
        
        # The Diebold and Mariano (DM) statistic
        dm_stat = d_bar[0]/se[0]
        
        # Constant to adjust the DM statistic from Harvey, Leybourne and Newbold
        const = TT + 1 - 2*hh + hh*(hh - 1)/TT
        
        # The t statistic
        t_stat = np.sqrt(const/TT)*dm_stat
    else:
        # The t statistic
        t_stat = np.nan
    
    return t_stat

