# -*- coding: utf-8 -*-
"""
Module with functions to seasonally adjust a monthly data series.

The seasonal adjustment in this module uses the state space model on page 80 of
Wright (2013).  Parameters of the state space model are estimated with an
expectation maximization (EM) algorithm.

This module was written for "Advance Layoff Notices and Aggregate Job Loss" by
Pawel M. Krolikowski and Kurt G. Lunsford.

Functions
---------
phi_mult_func : premultiplies a matrix with Phi matrix from the state equation
kalman_func : kalman filter and smoother
maximize_func : computes the maximization step of the EM algorithm
em_func : executes the EM algorithm
sa_func : executes the seasonal adjustment

Notes
-----
The state space model is as follows.  The data series of interest, x_t, is 
modeled with

    x_t = tau_t + gamma_t + nu_t

where tau_t is the trend component, gamma_t is the seasonal component, and nu_t
is the noise component.  The trend is modeled with

    tau_t = tau_{t-1} + beta_{t-1} + epsilon_{1,t}

where beta_t is a slope component modeled with

    beta_t = beta_{t-1} + epsilon_{2,t}.

The seasonal component is modeled with

    gamma_t = -gamma_{t-1} - gamma_{t-2} - ... - gamma_{t-11} + epsilon_{3,t}.

The innovations nu_t, epsilon_{1,t}, epsilon_{2,t} and epsilon_{3,t} are 
mutually and serially independent normal random variables with mean zero.

The whole state space model is written as an observation equation

    x_t = M [tau_t, beta_t, gamma_t, ..., gamma_{t-10}]' + nu_t

and a state eqation

    [tau_t, beta_t, gamma_t, ..., gamma_{t-10}]'
        = Phi [tau_{t-1}, beta_{t-1}, gamma_{t-1}, ..., gamma_{t-11}]'
        + [epsilon_{1,t}, epsilon_{2,t}, epsilon_{3,t}],

where M is a (1 x 13) matrix in the observation equation and Phi is a (13 x 13)
matrix in the state equation.  The parameters in the M and Phi matrices are
known.  The EM algorithm is used to estimate the variances of nu_t,
epsilon_{1,t}, epsilon_{2,t} and epsilon_{3,t}.
"""

import numpy as np
from numba import njit


@njit
def phi_mult_func(mat_in):
    """
    Premultiplies a matrix with the Phi matrix from the state equation.
    """
    
    # Dimensions of the input matrix
    dim1, dim2 = mat_in.shape
    
    # Output matrix
    mat_out = np.zeros((dim1, dim2))
    
    # First row
    mat_out[0, :] = mat_in[0, :] + mat_in[1, :]
    
    # Second row
    mat_out[1, :] = mat_in[1, :]
    
    # Third row
    for nnn in range(dim2):
        mat_out[2, nnn] = -np.sum(mat_in[2:dim1, nnn])
    
    # Remaining rows
    for nnn in range(dim1-3):
        mat_out[3+nnn, :] = mat_in[2+nnn, :]
    
    return mat_out


@njit
def kalman_func(s0, P0, zz, RR, QQ, WW):
    """
    The Kalman filter and smoother.
    
    This function is used to compute the expectation step of the EM algorithm.
    """
    
    # Dimensions of the data
    TT = zz[:, 0].size
    
    # Number of states
    rr = QQ[:, 0].size
    
    # Storage arrays for the Kalman filter
    s_pred = np.zeros((rr, TT))         # predictions of the state
    s_filt = np.zeros((rr, TT+1))       # filtered state
    P_pred = np.zeros((rr, rr, TT))     # prediction of variance of state
    P_filt = np.zeros((rr, rr, TT+1))   # filtered variance of state
    Jmats = np.zeros((rr, rr, TT))      # matrices for smoothing
    
    # Initial filtered values
    s_filt[:, 0] = s0[:, 0]
    P_filt[:, :, 0] = P0
    
    # Kalman filter
    llike = 0
    for jjj in range(TT):
        
        # Call filtered state from previous loop
        s_t1t1 = np.zeros((rr, 1))
        s_t1t1[:, 0] = s_filt[:, jjj]
        
        # Prediction of the state: s_{t|t-1} = E(s_t|z_1,...,z_{t-1})
        s_tt1 = np.zeros((rr, 1))
        s_tt1[:, :] = phi_mult_func(s_t1t1)
        s_pred[:, jjj] = s_tt1[:, 0]
        
        # Call filtered state variance from previous loop
        P_t1t1 = np.zeros((rr, rr))
        P_t1t1[:, :] = P_filt[:, :, jjj]
        
        # Prediction of state variance: P_{t|t-1} = Var(s_t|z_1,...,z_{t-1})
        PhiP = np.zeros((rr, rr))
        PhiP[:, :] = phi_mult_func(P_t1t1)
        
        PhiPPhi = np.zeros((rr, rr))
        PhiPPhi[:, :] = phi_mult_func(PhiP.T)
        
        P_tt1 = np.zeros((rr, rr))
        P_tt1[:, :] = PhiPPhi + QQ
        P_pred[:, :, jjj] = P_tt1[:, :]
        
        # Save matrix to be used for Kalman smoother
        Jtemp = np.zeros((rr, rr))
        Jtemp[:, :] = np.linalg.solve(P_tt1, PhiP)
        Jmats[:, :, jjj] = Jtemp.T
        
        # Additional steps that depend on if data is observed
        if WW[jjj, 0] == 1:
            # Variance of one-step ahead forecast
            H = np.zeros((1, 1))
            H[0, 0] = (P_tt1[0, 0] + P_tt1[0, 2] + P_tt1[2, 0] + P_tt1[2, 2]
                       + RR[0, 0])
            
            # Inverse of H
            Hinv = np.zeros((1, 1))
            Hinv[0, 0] = 1/H[0, 0]
            
            # Kalman gain
            K = np.zeros((rr, 1))
            K[:, 0] = (P_tt1[:, 0] + P_tt1[:, 2])*Hinv[0, 0]
            
            # One step ahead forecast error: v_t = z_t - E(y_t|z_1,...,z_{t-1})
            vv = np.zeros((1, 1))
            vv[0, 0] = zz[jjj, 0] - s_tt1[0, 0] - s_tt1[2, 0]
            
            # Updating step:
            # s_{t|t} = E(s_t|z_1,...,z_t) and P_{t|t} = Var(s_t|z_1,...,z_t)
            s_filt[:, jjj+1] = s_tt1[:, 0] + K[:, 0]*vv[0, 0]
            P_filt[:, :, jjj+1] = P_tt1 - K @ (P_tt1[0:1, :] + P_tt1[2:3, :])
            
            # Log-likelihood
            temp = np.zeros((1, 1))
            temp[0, 0] = vv[0, 0]*Hinv[0, 0]*vv[0, 0]
            llike = llike - (np.log(H[0, 0]) + temp[0, 0])/2
        else:
            # Kalman gain
            K = np.zeros((rr, 1))
            
            # Updating step:
            # s_{t|t} = E(s_t|z_1,...,z_t) and P_{t|t} = Var(s_t|z_1,...,z_t)
            s_filt[:, jjj+1] = s_tt1[:, 0]
            P_filt[:, :, jjj+1] = P_tt1
    
    # Storage arrays for the Kalman smoother
    s_smooth = np.zeros((rr, TT+1))         # smoothed state
    P_smooth = np.zeros((rr, rr, TT+1))     # smothed variance of state
    Pcov_smooth = np.zeros((rr, rr, TT))    # smoothed autocovariance of state
    
    # Last values
    s_smooth[:, TT:TT+1] = s_filt[:, TT:TT+1]
    P_smooth[:, :, TT] = P_filt[:, :, TT]
    
    P_tt = np.zeros((rr, rr))
    P_tt[:, :] = P_filt[:, :, TT-1]  
    
    temp1 = np.eye(rr)
    temp1[:, 0] = temp1[:, 0] - K[:, 0]
    temp1[:, 2] = temp1[:, 2] - K[:, 0]
    
    PhiP = np.zeros((rr, rr))
    PhiP[:, :] = phi_mult_func(P_tt)
    
    Pcov_smooth[:, :, TT-1] = temp1 @ PhiP
    
    # Some storage arrays to be used with numba @jit decorator
    J1_temp = np.zeros((rr, rr))
    J2_temp = np.zeros((rr, rr))
    
    # Kalman smoother
    for jjj in range(TT-1):
        # Define the index
        kk = TT - jjj - 1
        
        # Call some matrices
        J1_temp[:, :] = Jmats[:, :, kk]
        J2_temp[:, :] = Jmats[:, :, kk-1]
        P_tt[:, :] = P_filt[:, :, kk]
        
        # State smoother: E(s_t|z_1,...,z_T)
        temp = np.zeros((rr, 1))
        temp[:, :] = J1_temp @ (s_smooth[:, kk+1:kk+2] - s_pred[:, kk:kk+1])
        s_smooth[:, kk:kk+1] = s_filt[:, kk:kk+1] + temp
        
        # Variance smoother: Var(s_t|z_1,...,z_T)
        temp = np.zeros((rr, rr))
        temp[:, :] = J1_temp @ (P_smooth[:, :, kk+1] - P_pred[:, :, kk])
        P_smooth[:, :, kk] = P_tt + temp @ J1_temp.T
        
        # Autocovariance smoother: Cov(s_t, s_{t-1}|z_1,...,z_T)
        PhiP = np.zeros((rr, rr))
        PhiP[:, :] = phi_mult_func(P_tt)
        
        temp = np.zeros((rr, rr))
        temp[:, :] = P_tt + J1_temp @ (Pcov_smooth[:, :, kk] - PhiP)
        Pcov_smooth[:, :, kk-1] = temp @ J2_temp.T
    
    # Last loop of the Kalman smoother
    J1_temp[:, :] = Jmats[:, :, 0]
    P_tt[:, :] = P_filt[:, :, 0]
    
    temp = np.zeros((rr, 1))
    temp[:, :] = J1_temp @ (s_smooth[:, 1:2] - s_pred[:, 0:1])
    s_smooth[:, 0:1] = s_filt[:, 0:1] + temp
    
    temp = np.zeros((rr, rr))
    temp[:, :] = J1_temp @ (P_smooth[:, :, 1] - P_pred[:, :, 0]) @ J1_temp.T
    P_smooth[:, :, 0] = P_tt + temp
    
    return s_smooth, P_smooth, Pcov_smooth, llike


@njit
def maximize_func(ZZ, Wmats, RR, SS, PP, PPcov):
    """
    Function for maximization step in the EM algorithm.
    """
    
    # Dimensions of the data
    TT = ZZ[:, 0].size
    
    # Number of states
    rr = SS[:, 0].size
    
    # Output arrays
    Qupdate = np.zeros((rr, rr))
    Rupdate = np.zeros((1, 1))
    
    # Sums to compute updates of Q matrix
    ff0 = np.zeros((rr, TT))
    ff0[:, :] = SS[:, 0:TT]
    
    ff1 = np.zeros((rr, TT))
    ff1[:, :] = SS[:, 1:TT+1]
    
    sum1 = np.zeros((rr, rr))
    sum2 = np.zeros((rr, rr))
    sum3 = np.zeros((rr, rr))
    
    sum1[:, :] = ff0 @ ff0.T
    sum2[:, :] = ff1 @ ff0.T
    sum3[:, :] = ff1 @ ff1.T
    
    for jjj in range(TT):
        sum1 = sum1 + PP[:, :, jjj]
        sum2 = sum2 + PPcov[:, :, jjj]
        sum3 = sum3 + PP[:, :, jjj+1]
    
    # New value of Q matrix
    Phisum1 = np.zeros((rr, rr))
    Phisum1[:, :] = phi_mult_func(sum1)
    
    Phisum1Phi = np.zeros((rr, rr))
    Phisum1Phi[:, :] = phi_mult_func(Phisum1.T)
    
    Phisum2 = np.zeros((rr, rr))
    Phisum2[:, :] = phi_mult_func(sum2.T)
    
    temp = np.zeros((rr, rr))
    temp[:, :] = sum3 - Phisum2 - Phisum2.T + Phisum1Phi
    
    Qupdate[0, 0] = temp[0, 0]/TT
    Qupdate[1, 1] = temp[1, 1]/TT
    Qupdate[2, 2] = temp[2, 2]/TT
    
    # Loop to compute new value of the R matrix
    Rsum = np.zeros((1, 1))
    for jjj in range(TT):
        
        # Arrays used for computation
        tmp1 = np.zeros((1, 1))
        tmp2 = np.zeros((1, 1))
        
        # Add to Rsum depending on if variable is observed
        if Wmats[jjj, 0] == 1:
            tmp1[0, 0] = ZZ[jjj, 0] - SS[0, jjj+1] - SS[2, jjj+1]
            tmp2[0, 0] = (PP[0, 0, jjj+1] + PP[0, 2, jjj+1]
                          + PP[2, 0, jjj+1] + PP[2, 2, jjj+1])
            
            # Relevant sum
            Rsum[0, 0] = Rsum[0, 0] + tmp1[0, 0]*tmp1[0, 0] + tmp2[0, 0]
        else:
            # Relevant sum
            Rsum[0, 0] = Rsum[0, 0] + RR[0, 0]
    
    # Divide by the sample size
    Rupdate[0, 0] = Rsum[0, 0]/TT
    
    return Qupdate, Rupdate


def em_func(zz, Rin, Qin, WW, maxloop, threshold):
    """
    Function to execute the expectation maximization (EM) algorithm.
    """
    
    # Number of states
    rr = Qin[:, 0].size
    
    # Initial conditions of state for Kalman filter
    init_state = np.zeros((rr, 1))
    
    # Initial variance of state for Kalman filter
    A_init = np.zeros((rr, rr))
    A_init[0, 0] = 0.98
    A_init[0, 1] = 1
    A_init[1, 1] = 0.98
    A_init[2, 2:rr] = -0.98
    A_init[3:rr, 2:rr-1] = np.eye(10)
    
    Qvec = np.zeros((rr*rr, 1))
    Qvec[:, 0] = Qin.flatten()
    
    Mat = np.zeros((rr*rr, rr*rr))
    Mat[:, :] = np.eye(rr*rr) - np.kron(A_init, A_init)
    
    Matinv = np.zeros((rr*rr, rr*rr))
    Matinv[:, :] = np.linalg.inv(Mat)
    
    initP = np.zeros((rr*rr, 1))
    initP[:, :] = Matinv @ Qvec
    initP = np.reshape(initP, (rr, rr))
    
    # Initial expectation step via the Kalman filter and smoother
    state, P, Pcov, Loglike = kalman_func(init_state, initP, zz, Rin, Qin, WW)
    
    # Store the initial log-likelihood
    LL_save = np.zeros(maxloop+1)
    LL_save[0] = Loglike
    
    # Define the arrays for holding updated parameters
    Rnew = np.zeros((1, 1))
    Qnew = np.zeros((rr, rr))
    
    # Initial updates
    Rnew[0, 0] = Rin[0, 0]
    Qnew[0, 0] = Qin[0, 0]
    Qnew[1, 1] = Qin[1, 1]
    Qnew[2, 2] = Qin[2, 2]
    
    # EM algorithm loops
    for loop in range(maxloop):
        # Maximization step
        Qmax, Rmax = maximize_func(zz, WW, Rnew, state, P, Pcov)
        
        # New value of Q matrix
        Qnew[0, 0] = Qmax[0, 0]
        Qnew[1, 1] = Qmax[1, 1]
        Qnew[2, 2] = Qmax[2, 2]
        
        # New value of R matrix
        Rnew[0, 0] = Rmax[0, 0]
        
        # Initial values for expectation step
        init_state[:, 0] = state[:, 0]
        
        # Initial variance for Kalman filter
        Qvec = np.zeros((rr*rr, 1))
        Qvec[:, 0] = Qnew.flatten()
        
        initP = np.zeros((rr*rr, 1))
        initP[:, :] = Matinv @ Qvec
        initP = np.reshape(initP, (rr, rr))
        
        # Expectation step via the Kalman filter and smoother
        state, P, Pcov, Loglike = kalman_func(init_state, initP,
                                              zz, Rnew, Qnew, WW)
        
        # Store the log-likelihood
        LL_save[loop+1] = Loglike
        
        # Check for convergence
        dist = np.absolute(Loglike - LL_save[loop])
        if dist < threshold:
            break
        
    # Remove log-likelihood values not used
    LL_save = LL_save[0:loop+2]
    
    # Define output arrays
    Rout = np.zeros((1, 1))
    Qout = np.zeros((rr, rr))
    
    # Populate output arrays
    Rout[0, 0] = Rnew[0, 0]
    Qout[0, 0] = Qnew[0, 0]
    Qout[1, 1] = Qnew[1, 1]
    Qout[2, 2] = Qnew[2, 2]
    
    return state, P, Qout, Rout, LL_save, loop


def sa_func(series, loopmax, crit, str_name):
    """
    Function to seasonally adjust the data
    """
    
    # Number of observations
    TT = series[:, 0].size
    
    # Array to account for zero observations
    ww = np.zeros((TT, 1))
    mask = (series[:, 0] >= 1)
    ww[mask, :] = 1
    
    # Convert data to logs and set zeros to NaN
    series[mask, 0] = np.log(series[mask, 0])
    mask0 = (mask == 0)
    series[mask0, 0] = np.nan
    
    # Guess of the parameters
    sigma_v = np.nanvar(series)/1.5
    sigma_1 = np.nanvar(series)/50
    sigma_2 = np.nanvar(series)/50
    sigma_3 = np.nanvar(series)/50
    
    Qmat = np.zeros((13, 13))
    Qmat[0, 0] = sigma_1
    Qmat[1, 1] = sigma_2
    Qmat[2, 2] = sigma_3
    
    Rmat = np.zeros((1, 1))
    Rmat[0, 0] = sigma_v
    
    # The EM algorithm
    state_hat, Phat, Qhat, Rhat, Loglikes, ell = \
        em_func(series, Rmat, Qmat, ww, loopmax, crit)
    
    # Check if the EM algoithm finished before reaching max iterations
    if ell == loopmax-1:
        print('Seasonal adj for ' + str_name + ' is done.'
              + ' No EM algorithm solution.'
              + ' Maximum number of iterations reached.')
    else:
        print('Seasonal adj for ' + str_name + ' is done.'
              + ' EM algorithm solution found on loop ' + str(ell) + '.')
    
    # Seasonally adjust the data
    seasonal = np.zeros((TT, 1))
    seasonal[:, 0] = state_hat[2, 1:TT+1]
    
    series_sa = np.zeros((TT, 1))
    series_sa[mask, 0] = np.exp(series[mask, 0] - seasonal[mask, 0])
    series_sa[mask0, 0] = 0
    
    return series_sa

