# -*- coding: utf-8 -*-
"""
This program computes the WARN factor with an AR(1) structure and also an AR(2)
structure.  It then computes the correlation between these factors.  This
provides the result in footnote 11 of "Advance Layoff Notices and Aggregate
Job Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    seasonal_module.py : module with functions to seasonally adjust the data
    warn_module.py : module with functions to estimate the DFM
    WARNData_NSA_20211230.xlsx : Excel file with the state-level WARN data

This program prints the correlation in footnote 11.
"""


#%% Import Packages and Functions

# Get replication files directory
import os
path_dir = os.path.dirname(os.getcwd())

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Import Pandas and NumPy
import pandas as pd
import numpy as np

# Functions for seasonal adjustment
from modules.seasonal_module import sa_func

# Functions for estimating the DFM
from modules.dfm_em_module import make_guess, em_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the Program

"""
This section contains the parameters of the program.

The default values for the paper are as follows:
    str_date = '20211230'
    str_warn = '19960701'
    nloop = 20000
    thresh_sa = 1e-3
    thresh = 1e-5
"""

# String with the date of the data to call
str_date = '20211230'   # Format: 20200115 corresponds to January 15, 2020

# String to choose first date for WARN data
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# Parameters of the EM algorithm for seasonal adjustment
nloop = 20000       # maximum number of iterations in the EM algrorithm
thresh_sa = 1e-3    # threshold for convergence

# Parameters of the EM algorithm for the dynamic factor model
thresh = 1e-5       # threshold for convergence

"""
No further parameters need to be set beyond this line in the program.
"""


#%% Load the WARN Data

# Dataframe for number of workers affected by WARN notices by month and state
str_file_name = 'WARNData_NSA_' + str_date + '.xlsx'
path_warn_data = os.path.join(path_dir, 'data_input', str_file_name)
df_warn_all = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)

# Drop small states or states with too many observations with a value of 0
df_warn = df_warn_all.drop(columns=['AK', 'ID', 'KY', 'ND', 'NM',
                                    'OK', 'RI', 'UT', 'WV'])

# Number of states in the sample
nvars = len(df_warn.columns)


#%% Loop to Seasonally Adjust the Data

# Dataframe to store seasonally adjusted data
df_warn_sa = pd.DataFrame(index=df_warn.index, columns=df_warn.columns)

# Loop for seasonal adjustment
for loop in range(nvars):
    # State name
    state_name = df_warn.columns[loop]
    
    # Dates with observed data
    state_index = df_warn.loc[:, state_name].dropna().index
    
    # Number of observations excluding NaNs
    nsample = len(state_index)
    
    # Convert the data to numpy while dropping NaNs
    X_nsa = np.zeros((nsample, 1))
    X_nsa[:, 0] = df_warn.loc[:, state_name].dropna().to_numpy()
    
    # Seasonally adjust the data
    data_sa = sa_func(X_nsa, nloop, thresh_sa, state_name)
    
    # Store seasonally adjusted data in dataframe
    df_warn_sa.loc[state_index, state_name] = data_sa[:, 0]
    
    # Delete temprary variables
    del state_name, state_index, nsample, data_sa


#%% Adjust the Data and Construct the Selection Matrices for Missing Data

# Number of observations
nsample = len(df_warn_sa.loc[str_warn:, :].index)

# Convert the data to numpy
X = np.zeros((nsample, nvars))
X[:, :] = df_warn_sa.loc[str_warn:, :].to_numpy()

# Set zero values ot NaN and construct the selection matrices
W = np.zeros((nvars, nvars, nsample))
for jj in range(nsample):
    # Set zero values to NaN
    for ii in range(nvars):
        if X[jj, ii] < 1:
            X[jj, ii] = np.nan
    
    # Selection matrices
    W[:, :, jj] = np.diag(np.isnan(X[jj, :])==0)

# Delete temprary variables
del jj


#%% Initial Guess for the EM Algorithm with One Lag

# Take logs of the data
X = np.log(X)

# Guess of model parameters
nlags = 1
Fguess, Aguess, Qguess, Lguess, Rguess, Zstd, Avg, Var = make_guess(X, nlags)

# Guess of the initial state
F0guess = np.zeros((nlags, 1))

# Guess of covaraiance of initial state
Qvec = np.zeros((nlags*nlags, 1))
Qvec[0, 0] = Qguess[0, 0]

tmp = np.zeros((nlags*nlags, nlags*nlags))
tmp[:, :] = np.eye(nlags*nlags) - np.kron(Aguess, Aguess)

P0guess = np.zeros((nlags*nlags, 1))
P0guess[:, :] = np.linalg.solve(tmp, Qvec)
P0guess = np.reshape(P0guess, (nlags, nlags))

# Delete temprary variables
del Qvec, tmp


#%% EM Algorithm to Compute Model Parameters and WARN Factor with One Lag

Fhat1, Phat1, Ahat1, Qhat1, Lhat1, Rhat1, Loglikes1, ell1 = \
    em_func(F0guess, P0guess, Zstd, Lguess, Rguess, Aguess, Qguess,
            W, nloop, thresh)

# Check if the EM algoithm finished before reaching max iterations
print(' ')
if ell1 == nloop-1:
    print('No solution found with EM algorithm. '
          + 'Maximum number of iterations reached.')
else:
    print('Solution to EM algorithm found on loop ' + str(ell1))

# Compute WARN factor in term of units of employees affected
tmp = np.zeros((nvars, 1))
warn_hat1 = np.zeros((nsample, 1))
for jj in range(nsample):
    tmp[:, 0] = np.multiply(np.sqrt(Var[:, 0]), (Lhat1[:, 0]*Fhat1[0, jj+1]))
    warn_hat1[jj, 0] = np.sum(np.exp(Avg + tmp))

# Delete temprary variables
del ell1, tmp, jj, Fguess, Aguess, Qguess, Lguess, Rguess, Zstd, Avg, Var


#%% Initial Guess for the EM Algorithm with Two Lags

# Guess of model parameters
nlags = 2
Fguess, Aguess, Qguess, Lguess, Rguess, Zstd, Avg, Var = make_guess(X, nlags)

# Guess of the initial state
F0guess = np.zeros((nlags, 1))

# Guess of covaraiance of initial state
Qvec = np.zeros((nlags*nlags, 1))
Qvec[0, 0] = Qguess[0, 0]

tmp = np.zeros((nlags*nlags, nlags*nlags))
tmp[:, :] = np.eye(nlags*nlags) - np.kron(Aguess, Aguess)

P0guess = np.zeros((nlags*nlags, 1))
P0guess[:, :] = np.linalg.solve(tmp, Qvec)
P0guess = np.reshape(P0guess, (nlags, nlags))

# Delete temprary variables
del Qvec, tmp


#%% EM Algorithm to Compute Model Parameters and WARN Factor with Two Lags

Fhat2, Phat2, Ahat2, Qhat2, Lhat2, Rhat2, Loglikes2, ell2 = \
    em_func(F0guess, P0guess, Zstd, Lguess, Rguess, Aguess, Qguess,
            W, nloop, thresh)

# Check if the EM algoithm finished before reaching max iterations
print(' ')
if ell2 == nloop-1:
    print('No solution found with EM algorithm. '
          + 'Maximum number of iterations reached.')
else:
    print('Solution to EM algorithm found on loop ' + str(ell2))

# Compute WARN factor in term of units of employees affected
tmp = np.zeros((nvars, 1))
warn_hat2 = np.zeros((nsample, 1))
for jj in range(nsample):
    tmp[:, 0] = np.multiply(np.sqrt(Var[:, 0]), (Lhat2[:, 0]*Fhat2[0, jj+1]))
    warn_hat2[jj, 0] = np.sum(np.exp(Avg + tmp))

# Delete temprary variables
del ell2, tmp, jj, Fguess, Aguess, Qguess, Lguess, Rguess, Zstd, Avg, Var


#%% Correlation Between the Factors

# Compute the correlations
vals = np.zeros((2, nsample))
vals[0, :] = Fhat1[0, 1:nsample+1]
vals[1, :] = Fhat2[0, 1:nsample+1]
corr_lags = np.corrcoef(vals)

# Print the results
print(' ')
print('The correlation between a factor estimated with an AR(1)')
print('and a factor estimated with an AR(2) is')
print(corr_lags[1, 0])
print(' ')


#%% Disply the Run Time in Minutes

print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


