# -*- coding: utf-8 -*-
"""
This program produces the output from the dynamic factor model (DFM) for
"Advance Layoff Notices and Aggregate Job Loss" 
by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    seasonal_module.py : module with functions to seasonally adjust the data
    warn_module.py : module with functions to estimate the DFM
    WARNData_NSA_20211230.xlsx : Excel file with the state-level WARN data
    nber_recessions.xlsx : Excel file with NBER recession indicators

The main output of this program is
    df_results : pandas dataframe
                 This dataframe includes 4 variables
                 (1) WARN_factor : factor from the DFM in Section 4
                 (2) MSE : mean squared error of the WARN factor
                 (3) WARN_hat : number of workers affected by WARN notices as
                                implied by the WARN factor
                                see equation (4) of the paper
                 (4) WARN_sum : sum of states with no missing observations
                                beginning in January 2006

This program also saves
    WARN_factor.pdf : Figure 4 in the paper
    WARNFactors_20211230.xlsx : an Excel file that holds the df_results DataFrame
"""


#%% Import Packages and Functions

# Get replication files directory
import os
path_dir = os.path.dirname(os.getcwd())

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Import Pandas, NumPy and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# Functions for seasonal adjustment
from modules.seasonal_module import sa_func

# Functions for estimating the DFM
from modules.dfm_em_module import make_autocov, make_guess, em_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the Program

"""
This section contains the parameters of the program.

The default values for the paper are as follows:
    str_date = '20211230'
    str_warn = '19960701'
    str_sum = '20060101'
    nlags = 1
    nloop = 20000
    thresh_sa = 1e-3
    thresh = 1e-5
    nsim = 500
    np.random.seed(1)
"""

# String with the date of the data to call
str_date = '20211230'   # Format: 20200115 corresponds to January 15, 2020

# String to choose first date for WARN data
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# String to choose first date for summing the balanced panel
# This date must be later than the str_warn date
str_sum = '20060101'    # Format: 20060101 corresponds to January 1, 2006

# Number of lags in the state equation
nlags = 1

# Parameters of the EM algorithm for seasonal adjustment
nloop = 20000       # maximum number of iterations in the EM algrorithm
thresh_sa = 1e-3    # threshold for convergence

# Parameters of the EM algorithm for the dynamic factor model
thresh = 1e-5       # threshold for convergence

# Parameters for bootstrapping the mean squared errors
nsim = 500          # Number of simulations for producing mean squared error
np.random.seed(1)   # Seed the random number generator

"""
No further parameters need to be set beyond this line in the program.
"""


#%% Load the WARN Data

# Dataframe for number of workers affected by WARN notices by month and state
str_file_name = 'WARNData_NSA_' + str_date + '.xlsx'
path_warn_data = os.path.join(path_dir, 'data_input', str_file_name)
df_warn_all = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)

# Drop small states or states with too many observations with a value of 0
df_warn = df_warn_all.drop(columns=['AK', 'ID', 'KY', 'ND', 'NM',
                                    'OK', 'RI', 'UT', 'WV'])

# Number of states in the sample
nvars = len(df_warn.columns)


#%% Load the NBER Recession Indicators for the Figures

path_rec_data = os.path.join(path_dir, 'data_input', 'nber_recessions.xlsx')
df_recession = pd.read_excel(path_rec_data, 'Sheet1', index_col=0)
df_recession = df_recession.loc[str_warn:, :]


#%% Loop to Seasonally Adjust the Data

# Dataframe to store seasonally adjusted data
df_warn_sa = pd.DataFrame(index=df_warn.index, columns=df_warn.columns)

# Loop for seasonal adjustment
for loop in range(nvars):
    # State name
    state_name = df_warn.columns[loop]
    
    # Dates with observed data
    state_index = df_warn.loc[:, state_name].dropna().index
    
    # Number of observations excluding NaNs
    nsample = len(state_index)
    
    # Convert the data to NumPy while dropping NaNs
    X_nsa = np.zeros((nsample, 1))
    X_nsa[:, 0] = df_warn.loc[:, state_name].dropna().to_numpy()
    
    # Seasonally adjust the data
    data_sa = sa_func(X_nsa, nloop, thresh_sa, state_name)
    
    # Store seasonally adjusted data in dataframe
    df_warn_sa.loc[state_index, state_name] = data_sa[:, 0]
    
    # Delete temprary variables
    del state_name, state_index, nsample, data_sa


#%% Adjust the Data and Construct the Selection Matrices for Missing Data

# Number of observations
nsample = len(df_warn_sa.loc[str_warn:, :].index)

# Convert the data to NumPy
X = np.zeros((nsample, nvars))
X[:, :] = df_warn_sa.loc[str_warn:, :].to_numpy()

# Set zero values to NaN and construct the selection matrices
W = np.zeros((nvars, nvars, nsample))
for jj in range(nsample):
    # Set zero values to NaN
    for ii in range(nvars):
        if X[jj, ii] < 1:
            X[jj, ii] = np.nan
    
    # Selection matrices
    W[:, :, jj] = np.diag(np.isnan(X[jj, :])==0)

# Delete temprary variables
del jj


#%% Initial Guess for the EM Algorithm

# Take logs of the data
X = np.log(X)

# Guess of model parameters
Fguess, Aguess, Qguess, Lguess, Rguess, Zstd, Avg, Var = make_guess(X, nlags)

# Guess of the initial state
F0guess = np.zeros((nlags, 1))

# Guess of covaraiance of initial state
Qvec = np.zeros((nlags*nlags, 1))
Qvec[0, 0] = Qguess[0, 0]

tmp = np.zeros((nlags*nlags, nlags*nlags))
tmp[:, :] = np.eye(nlags*nlags) - np.kron(Aguess, Aguess)

P0guess = np.zeros((nlags*nlags, 1))
P0guess[:, :] = np.linalg.solve(tmp, Qvec)
P0guess = np.reshape(P0guess, (nlags, nlags))

# Delete temprary variables
del Qvec, tmp


#%% EM Algorithm to Compute Model Parameters and WARN Factor

Fhat, Phat, Ahat, Qhat, Lhat, Rhat, Loglikes, ell = \
    em_func(F0guess, P0guess, Zstd, Lguess, Rguess, Aguess, Qguess,
            W, nloop, thresh)

# Check if the EM algoithm finished before reaching max iterations
if ell == nloop-1:
    print('No solution found with EM algorithm. '
          + 'Maximum number of iterations reached.')
else:
    print('Solution to EM algorithm found on loop ' + str(ell))

# Compute WARN factor in term of units of employees affected
tmp = np.zeros((nvars, 1))
warn_hat = np.zeros((nsample, 1))
for jj in range(nsample):
    tmp[:, 0] = np.multiply(np.sqrt(Var[:, 0]), (Lhat[:, 0]*Fhat[0, jj+1]))
    warn_hat[jj, 0] = np.sum(np.exp(Avg + tmp))

# Delete temprary variables
del ell, tmp, jj


#%% Bootstrap to Compute Mean Squared Error

# Compute matrices for simulation
Qhalf = np.zeros((nlags, nlags))
Qhalf[:, :] = np.sqrt(Qhat)

Rhalf = np.zeros((nvars, nvars))
Rhalf[:, :] = np.sqrt(Rhat)

Sigma = make_autocov(Ahat, Qhat)
Shalf = np.zeros((nlags, nlags))
Shalf[:, :] = np.linalg.cholesky(Sigma)

# Bootstrap simulations
mse = np.zeros((nsample, nlags))
for sim in range(nsim):
    # Storage arrays
    Fsim = np.zeros((nlags, nsample+1))
    Xsim = np.zeros((nsample, nvars))
    
    # Inititalize with the actual data to match the pattern of NaNs 
    Xsim[:, :] = X[:, :]
    
    # Simulate the shocks
    Esim = np.zeros((nvars, nsample))
    Esim[:, :] = Rhalf @ np.random.standard_normal(size=(nvars, nsample))
    Usim = np.zeros((nlags, nsample))
    Usim[0, :] = Qhalf[0, 0]*np.random.standard_normal(size=nsample)
    
    # Initial state
    Fsim[:, 0:1] = Shalf @ np.random.standard_normal(size=(nlags, 1))
    
    # Loop to simulate
    Ztemp = np.zeros((nvars, 1))
    Xtemp = np.zeros((nvars, 1))
    for jj in range(nsample):
        # Simulate the factor
        Fsim[:, jj+1:jj+2] = Ahat @ Fsim[:, jj:jj+1] + Usim[:, jj:jj+1]
        
        # Simulated data with no missing observations
        Ztemp[:, 0] = Lhat[:, 0]*Fsim[0, jj+1] + Esim[:, jj]
        Xtemp[:, 0] = Avg[:, 0] + np.multiply(np.sqrt(Var[:, 0]), Ztemp[:, 0])
        
        # Match the pattern of missing data
        mask = (np.sum(W[:, :, jj], axis=1) == 1)
        Xsim[jj, mask] = Xtemp[mask, 0]
    
    # Averages of the data
    avg_boot = np.zeros((nvars, 1))
    avg_boot[:, 0] = np.nanmean(Xsim, axis=0)
    
    # Variances of the data
    var_boot = np.zeros((nvars, 1))
    var_boot[:, 0] = np.nanvar(Xsim, axis=0)
    
    # Standardize the data: de-mean, unit variance, set NaNs to zero
    Zstd_boot = np.zeros((nsample, nvars))
    for jj in range(nvars):
        mask = (np.isnan(Xsim[:, jj]) == 0)
        Zstd_boot[mask, jj] =\
            (Xsim[mask, jj] - avg_boot[jj, 0])/np.sqrt(var_boot[jj, 0])    
    
    # Run the EM algorithm
    F0_boot = np.zeros((nlags, 1))
    Fboot, Pboot, Aboot, Qboot, Lboot, Rboot, LogLboot, ell = \
        em_func(F0_boot, Sigma, Zstd_boot, Lhat, Rhat, Ahat, Qhat,
                W, nloop, thresh)
    
    # Check if the EM algoithm finished before reaching max iterations
    if ell == nloop-1:
        print('No solution found with EM algorithm. '
              + 'Maximum number of iterations reached.')
    else:
        print('Solution to EM algorithm found on loop ' + str(ell))
    
    # Stack the P matrices
    Pstack = np.zeros((nsample, nlags))
    for jj in range(nlags):
        Pstack[:, jj] = Pboot[jj, jj, 1:nsample+1]
    
    # Initial calculation for mean squared error
    diff = np.zeros((nsample, nlags))
    diff[:, :] = Fboot[0:nlags, 1:nsample+1].T - Fsim[0:nlags, 1:nsample+1].T
    mse = mse + np.power(diff, 2) - Pstack
    
    # Print the simulation loop
    print('Simulation loop ' + str(sim)
          + ' out of ' + str(nsim) + ' is complete.')

# Stack the P matrices
Pstack = np.zeros((nsample, nlags))
for jj in range(nlags):
    Pstack[:, jj] = Phat[jj, jj, 1:nsample+1]

# Compute the mean squared error
mse = mse/nsim + 2*Pstack

# Delete temporary variables
del F0_boot, Fboot, Pboot, Aboot, Qboot, Lboot, Rboot, LogLboot, ell, Pstack
del diff, mask, sim, jj


#%% Use the Mean Squares Errors to Compute Confidence Intervals

# Define the arrays to hold the confidence intervals
ci68 = np.zeros((nsample, 2))
ci95 = np.zeros((nsample, 2))

# Compute the standard error
se = np.zeros((nsample, 1))
se[:, 0] = np.sqrt(mse[:, 0])

# Populate the arrays
ci68[:, 0] = Fhat[0, 1:nsample+1] - 0.99446*se[:, 0]
ci68[:, 1] = Fhat[0, 1:nsample+1] + 0.99446*se[:, 0]

ci95[:, 0] = Fhat[0, 1:nsample+1] - 1.96*se[:, 0]
ci95[:, 1] = Fhat[0, 1:nsample+1] + 1.96*se[:, 0]


#%% Compute the Sum of the Balanced Panel

# WARN data beginning in date in str_sum
data = np.zeros((len(df_warn_all.loc[str_sum:].index), len(df_warn_all.columns)))
data[:, :] = df_warn_all.loc[str_sum:].to_numpy()

# Sum the data that are part of a balanced panel
warn_sum = np.zeros((data[:, 0].size, 1))
str_states = 'the states in this balanced panel are'
count_states = 0
for jj in range(len(df_warn_all.columns)):
    indicator = np.sum(np.isnan(data[:, jj]))
    if indicator == 0:
        warn_sum = warn_sum + data[:, jj:jj+1]
        str_states = str_states + ' ' + df_warn_all.columns[jj] + ','
        count_states = count_states + 1
str_states = str_states[0:-1]

# Adjust the dimensions of the sum
tmp = nsample - data[:, 0].size
warn_sum = np.vstack((np.full((tmp, 1), np.nan), warn_sum))

# Delete temprary variables
del data, indicator, jj, tmp


#%% Collect the Output and Save to Excel

# Dataframe for the sheet holding the data
results = np.hstack((Fhat[0:1, 1:nsample+1].T,
                     mse[:, 0:1], warn_hat, warn_sum))
res_cols = ['WARN Factor', 'MSE', 'WARN_hat', 'WARN_sum']
df_results = pd.DataFrame(results, index=df_warn_sa.loc[str_warn:, :].index,
                          columns=res_cols)

# Dataframe for the README sheet
col1 = ['This file contains data associated with "Advance Layoff Notices and '
        + 'Labor Market Forecasting" by Pawel M. Krolikowski and Kurt G. '
        + 'Lunsford.']
col1.append('This paper is Federal Reserve Bank of Cleveland Working Paper '
            + 'no. 20-03.')
col1.append(np.nan)
col1.append('"Sheet1" contains all data in this file. The data are labeled as '
            + 'follows:')
col2 = [np.nan, np.nan, np.nan, np.nan]
col3 = [np.nan, np.nan, np.nan, np.nan]

col1.append(np.nan)
col2.append('WARN Factor:')
col3.append('these data are the estimates of the factor from the dynamic '
            + 'factor model in Section 4 of the paper')

col1.append(np.nan)
col2.append('MSE:')
col3.append('these data are the estimated mean squared errors of the WARN '
            + 'factor')

col1.append(np.nan)
col2.append('WARN_hat:')
col3.append('these data are the output of equation (5) in the paper, which '
            + 'converts the WARN factor into the number of workers affected '
            + 'by WARN notices')

col1.append(np.nan)
col2.append('WARN_sum:')
col3.append('these data are a sum of ' 
            + str(count_states) 
            + ' states (not seasonally adjusted) that form a balanced panel '
            + 'beginning in January 2006')

col1.append(np.nan)
col2.append(np.nan)
col3.append(str_states)

df_readme = pd.DataFrame({'col1': col1, 'col2': col2, 'col3': col3})

# Save to Excel
str_file = 'WARNFactors_' + str_date + '.xlsx'
with pd.ExcelWriter(str_file, datetime_format='YYYY-MM') as writer:
    df_readme.to_excel(writer, sheet_name='README', header=False, index=False, startrow=1, startcol=1)
    df_results.to_excel(writer, sheet_name='Sheet1')

# Delete temprary variables
del str_file


#%% Plot Figure 6

# The sample for the plot
sample = df_results.index
nplot = len(sample)

# Figure and axes instances
fig, ax = plt.subplots(2, 1, figsize=(7, 7))

# Plot the WARN factor
ax[0].fill_between(sample, 10000*df_recession.loc[sample, 'USREC'],
                -10000*df_recession.loc[sample, 'USREC'],
                color='black', alpha=0.15, linewidth=0)
ax[0].plot(sample, np.zeros(nplot), color='black', linewidth=1)
ax[0].fill_between(sample, ci95[0:nplot, 0], ci95[0:nplot, 1],
                   color='black', alpha=0.40, linewidth=0)
ax[0].plot(sample, df_results.loc[sample, 'WARN Factor'],
           color='black', linewidth=1.5)
ax[0].set_xlim(sample[0], sample[-1])
ax[0].set_ylim(-3.4, 6.2)
ax[0].set_ylabel('Standard deviations', fontsize=10)
ax[0].set_title('WARN Factor and 95 Percent Confidence Interval', fontsize=12)

# Plot the WARN_hat
ax[1].fill_between(sample, 10000*df_recession.loc[sample, 'USREC'],
                -10000*df_recession.loc[sample, 'USREC'],
                color='black', alpha=0.15, linewidth=0)
ax[1].plot(sample, np.zeros(nplot), color='black', linewidth=1)
ax[1].plot(sample, df_results.loc[sample, 'WARN_hat']/1000,
        color='black', linewidth=1.5)
ax[1].set_xlim(sample[0], sample[-1])
ax[1].set_ylim(0, 260)
ax[1].set_ylabel('Number of workers (Thousands, SA)', fontsize=10)
ttt = 'The Number of WARN Layoffs Implied by the WARN Factor ($\widehat{WARN}_t$)'
ax[1].set_title(ttt, fontsize=12)

fig.tight_layout()

# Save the figure
plt.savefig('WARN_factor.pdf', bbox_inches='tight')


#%% Disply the Run Time in Minutes

print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


