# -*- coding: utf-8 -*-
"""
This program produces Figure 5 in "Advance Layoff Notices and Aggregate Job
Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    WARNFactors_20211230.xlsx : the output of warn_dfm_program.py in
                                WARN_RepFiles\section_4
    mls_data.xlsx : Excel file with data from the Mass Layoff Statistics (MLS)

This program saves
    WARN_hat_MLS.pdf : Figure 5 in the paper
"""


#%% Import Packages and Functions

# Pandas, NumPy and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# Enable latex
plt.rc('text', usetex=True)


#%% Load the Data

# Get replication files directory
import os
path_dir = os.path.dirname(os.getcwd())

# Load the WARN factor and WARN_hat
path_warn_data = os.path.join(path_dir, 'section_4', 'WARNFactors_20211230.xlsx')
df_warn = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)

# Load the MLS data
path_mls_data = os.path.join(path_dir, 'data_input', 'mls_data.xlsx')
df_mls = pd.read_excel(path_mls_data, 'Sheet1', index_col=0)

# Read in recession dates and adjust the sample
path_rec_data = os.path.join(path_dir, 'data_input', 'nber_recessions.xlsx')
df_recession = pd.read_excel(path_rec_data, 'Sheet1', index_col=0)
df_recession = df_recession.loc[df_mls.index[0]:, :]


#%% Merge All Data into a New DataFrame

# Date index for the dataframe
ind = pd.date_range(start=df_mls.index[0], end=df_warn.index[-1], freq='MS')

# Column labels for the dataframe
col_labels = ['factor', 'warn_hat', 'mls',
              'jolts_small', 'jolts_big', 'recession']

# The new dataframe
df_figures = pd.DataFrame(index=ind, columns=col_labels)
df_figures.loc[:, 'factor'] = df_warn.loc[:, 'WARN Factor']
df_figures.loc[:, 'warn_hat'] = df_warn.loc[:, 'WARN_hat']
df_figures.loc[:, 'mls'] = df_mls.loc[:, 'mls_sa']
df_figures.loc[:, 'recession'] = df_recession.loc[:, 'USREC']


#%% Plot Figure 5

# The sample for the plot
sample = df_figures.loc[df_warn.index[0]:df_mls.index[-1], :].index
nplot = len(sample)

# Figure and axes instances
fig, ax = plt.subplots()

# Plot WARN_hat along with MLS
ax.fill_between(sample, 10000*df_figures.loc[sample, 'recession'],
                -10000*df_figures.loc[sample, 'recession'],
                color='black', alpha=0.15, linewidth=0)
ax.plot(sample, np.zeros(nplot), color='black', linewidth=1)
line1, = ax.plot(sample, df_figures.loc[sample, 'warn_hat']/1000,
                 color='black', linewidth=1.5)
ax.set_xlim(sample[0], sample[-1])
ax.set_ylim(0, 74)
ax.set_ylabel('Number of workers (Thousands, SA)', fontsize=10)

ax2 = ax.twinx()
line2, = ax2.plot(sample, df_figures.loc[sample, 'mls']/1000,
                  color='black', linewidth=1.5, alpha=0.6)
ax2.set_xlim(sample[0], sample[-1])
ax2.set_ylim(0, 370)
ax2.set_ylabel('Number of initial claimants (Thousands, SA)', fontsize=10)

ax.legend([line1, line2], ['$\widehat{WARN}_t$ (left)', 'MLS (right)'],
          loc='upper left')

#fig.tight_layout()

# Save the figure
plt.savefig('WARN_hat_MLS.pdf', bbox_inches='tight')

