# -*- coding: utf-8 -*-
"""
This program produces Table 3 and Figure 7 in "Advance Layoff Notices and
Aggregate Job Loss" by Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    var_data.xlsx : Excel file with data for each variable in the VAR

The main output of this program is
    table_3 : an array holding the values for Table 3 in the paper
    irf : an array holding the impulse response functions (IRFs) shown in
          Figure 9 in the paper
    ci_high95 : the upper ends of the confidence intervals shown in Figure 7
    ci_low95 : the lower ends of the cnfience intevrals shown in Figure 7

This program saves
    IRFs.pdf : Figure 7 in the paper
"""

#%% Import Packages and Functions

# Get replication files directory
import os
path_dir = os.path.dirname(os.getcwd())

# Pandas, NumPy and Matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# Enable latex
plt.rc('text', usetex=True)

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the Program

# String to choose last date used in the VAR
str_last = '20191231'   # Format: 20191231 corresponds to December 31, 2019

# Number of variables in the VAR
nvars = 6

# Number of lags in the VAR
nlags = 2

# Maximum horizon to forecast
nirf = 37

# Number of bootstrap replications
nboot = 5000

# Seed of random number generator
np.random.seed(1)


#%% Load and Adjust the Data

# Load the VAR data
path_var_data = os.path.join(path_dir, 'data_input', 'var_data.xlsx')
df_var = pd.read_excel(path_var_data, 'Sheet1', index_col=0)

# Log of f, s, unemployment rate and UI claims
df_var.iloc[:, 0:4] = np.log(df_var.iloc[:, 0:4])

# Difference of log unemployment rate
df_var.loc[:, 'urate'] = df_var.loc[:, 'urate'].diff()

# Difference of log unemployment rate
df_var.loc[:, 'uspriv'] = df_var.loc[:, 'uspriv'].diff()

# Difference of WARN hat
df_var.loc[:, 'warn_hat'] = df_var.loc[:, 'warn_hat'].diff()


#%% In-Sample Regression

# Number of data observations
nobs = len(df_var.loc[:str_last, :].index)

# Convert dataframe data to a NumPy array
var_data = np.zeros((nobs-1, nvars))
var_data[:, 0] = 100*df_var.iloc[1:nobs, 0].to_numpy()
var_data[:, 1] = 100*df_var.iloc[1:nobs, 1].to_numpy()
var_data[:, 2] = 100*df_var.iloc[1:nobs, 2].to_numpy()
var_data[:, 3] = 100*df_var.iloc[1:nobs, 3].to_numpy()
var_data[:, 4] = df_var.iloc[1:nobs, 4].to_numpy()
var_data[:, 5] = df_var.iloc[1:nobs, 5].to_numpy()/1000

# Effective sample size of the VAR
nsample = nobs - nlags - 1

# Left-hand side array
lhs = np.zeros((nsample, nvars))
lhs[:, :] = var_data[nlags:nobs-1, :]

# Right-hand side array
rhs = np.ones((nsample, nvars*nlags+1))
for jj in range(nlags):
    rhs[:, nvars*jj:nvars*(jj+1)] = var_data[nlags-jj-1:nobs-jj-2, :]

# Least squares regression
mat1 = rhs.T @ rhs
mat2 = rhs.T @ lhs
gamma = np.linalg.solve(mat1, mat2)

# Regression residuals
errors = lhs - rhs @ gamma

# Compute standard errors equation by equation
se = np.zeros((nvars*nlags+1, nvars))
for jj in range(nvars):
    # Product of right-hand side matrix and residuals
    zz = np.zeros((nsample, nvars))
    zz = np.multiply((errors[:, jj:jj+1] @ np.ones((1, nvars*nlags+1))), rhs)
    
    # Covariance matrix
    V = nsample*np.linalg.inv(mat1) @ (zz.T @ zz) @ np.linalg.inv(mat1)
    
    #Standard errors
    se[:, jj] = np.sqrt(np.diag(V)/nsample)

# t statistics
tstats = np.zeros((nvars*nlags+1, nvars))
tstats[:, :] = np.divide(gamma, se)


#%% An Array for Table 3

# Populate the table
table_3 = np.zeros((2*nlags, nvars))
for jj in range(nlags):
    table_3[jj*2, :] = gamma[(jj+1)*nvars-1, :]
    table_3[jj*2+1, :] = se[(jj+1)*nvars-1, :]

# Round to 2 digits for presentation
table_3 = np.round(table_3, 2)


#%% Compute Impulse Response Functions

# History used to compute IRFs
history = np.zeros((1, nvars*nlags))

mat = np.linalg.cholesky((errors.T @ errors)/nsample)

# Initial impulse response is standard deviation of WARN equation residuals
irf = np.zeros((nirf, nvars))
irf[0, nvars-1] = mat[nvars-1, nvars-1]

# Loop to compute additional impulse responses
for jj in range(nirf-1):
    history[0, nvars:nvars*nlags] = history[0, 0:nvars*(nlags-1)]
    history[0, 0:nvars] = irf[jj:jj+1, :]
    irf[jj+1:jj+2, :] = history @ gamma[0:nvars*nlags, :]

# Adjust the IRFs for better interpretation
irf[:, 2] = np.cumsum(irf[:, 2])


#%% Bootstrap the IRFs

# Storage array
irf_boot = np.zeros((nirf, nvars, nboot))

# Bootstrap loop
for boot in range(nboot):
    # The index to resample the residuals
    index = np.random.randint(0, nsample, nsample)
    
    # Resample the residuals
    e_boot = np.zeros((nsample, nvars))
    e_boot[:, :] = errors[index, :]
    
    # Storage arrays
    lhs_boot = np.zeros((nsample, nvars))
    rhs_boot = np.ones((nsample, nvars*nlags+1))
    
    # Recursively produce the endogenous variables
    rhs_boot[0, :] = rhs[0, :]      # Initial conditions
    for jj in range(nsample-1):
        # Left-hand side array
        lhs_boot[jj, :] = rhs_boot[jj, :] @ gamma + e_boot[jj, :]
        
        # Right-hand side array
        rhs_boot[jj+1, 0:nvars] = lhs_boot[jj, :]
        rhs_boot[jj+1, nvars:nvars*nlags] = rhs_boot[jj, 0:nvars*(nlags-1)]
    
    # Last value of left-hand side array
    lhs_boot[nsample-1, :] = rhs_boot[nsample-1, :] @ gamma + e_boot[nsample-1, :]
    
    # Least squares regression
    mat1_boot = rhs_boot.T @ rhs_boot
    mat2_boot = rhs_boot.T @ lhs_boot
    gamma_boot = np.linalg.solve(mat1_boot, mat2_boot)
    
    # Regression residuals
    errors_boot = np.zeros((nsample, nvars))
    errors_boot[:, :] = lhs_boot - rhs_boot @ gamma_boot
    
    # History used to compute IRFs
    history = np.zeros((1, nvars*nlags))
    
    # Initial impulse response is standard deviation of WARN equation residuals
    irf_temp = np.zeros((nirf, nvars))
    irf_temp[0, nvars-1] = np.sqrt((errors_boot[:, nvars-1:nvars].T
                                    @ errors_boot[:, nvars-1:nvars])/nsample)
    
    # Loop to compute additional impulse responses
    for jj in range(nirf-1):
        history[0, nvars:nvars*nlags] = history[0, 0:nvars*(nlags-1)]
        history[0, 0:nvars] = irf_temp[jj:jj+1, :]
        irf_temp[jj+1:jj+2, :] = history @ gamma_boot[0:nvars*nlags, :]
    
    # Adjust the IRFs for better interpretation
    irf_temp[:, 2] = np.cumsum(irf_temp[:, 2])
    
    # Store the bootstrapped IRFs
    irf_boot[:, :, boot] = irf_temp[:, :]
    
    print('This is bootstrap loop number ' + str(boot) + '.')


#%% Compute the Confidence Intervals

# Sort the bootstrapped impulse responses
irf_sort = np.zeros((nirf, nvars, nboot))
irf_sort[:, :, :] = np.sort(irf_boot, axis=2)

# Set the level of the confidence inetrval
level_low95 = int(0.025*nboot)
level_high95 = int(0.975*nboot)

# Percentile confidence intervals
ci_low95 = np.zeros((nirf, nvars))
ci_high95 = np.zeros((nirf, nvars))

for jj in range(nvars):
    ci_low95[:, jj] = irf_sort[:, jj, level_low95]
    ci_high95[:, jj] = irf_sort[:, jj, level_high95]


#%% Plot the Figure 7

# Sample to plot on
sample = np.arange(0, nirf)

# Figure and axes instances
fig, ax = plt.subplots(2, 3, figsize=(9, 7))

ax[0, 0].plot(sample, irf[:, 0], linewidth=2, color='black')
ax[0, 0].fill_between(sample, ci_high95[:, 0], ci_low95[:, 0],
                      color='black', alpha=0.30, linewidth=0)
ax[0, 0].plot(sample, np.zeros(nirf), color='black', linewidth=0.5)
ax[0, 0].set_xlim(sample[0], sample[-1])
ax[0, 0].set_xlabel('Months', fontsize=12)
ax[0, 0].set_ylabel('Percent', fontsize=12)
ax[0, 0].tick_params(axis='x', labelsize=10)
ax[0, 0].tick_params(axis='y', labelsize=10)
ax[0, 0].set_title('Job finding rate', fontsize=12)

ax[0, 1].plot(sample, irf[:, 1], linewidth=2, color='black')
ax[0, 1].fill_between(sample, ci_high95[:, 1], ci_low95[:, 1],
                      color='black', alpha=0.30, linewidth=0)
ax[0, 1].plot(sample, np.zeros(nirf), color='black', linewidth=0.5)
ax[0, 1].set_xlim(sample[0], sample[-1])
ax[0, 1].set_xlabel('Months', fontsize=12)
ax[0, 1].set_ylabel('Percent', fontsize=12)
ax[0, 1].set_yticks([0, 0.5, 1, 1.5])
ax[0, 1].tick_params(axis='x', labelsize=10)
ax[0, 1].tick_params(axis='y', labelsize=10)
ax[0, 1].set_title('Job separation rate', fontsize=12)

ax[0, 2].plot(sample, irf[:, 2], linewidth=2, color='black')
ax[0, 2].fill_between(sample, ci_high95[:, 2], ci_low95[:, 2],
                      color='black', alpha=0.30, linewidth=0)
ax[0, 2].plot(sample, np.zeros(nirf), color='black', linewidth=0.5)
ax[0, 2].set_xlim(sample[0], sample[-1])
ax[0, 2].set_xlabel('Months', fontsize=12)
ax[0, 2].set_ylabel('Percent', fontsize=12)
ax[0, 2].set_yticks([0, 0.5, 1, 1.5])
ax[0, 2].tick_params(axis='x', labelsize=10)
ax[0, 2].tick_params(axis='y', labelsize=10)
ax[0, 2].set_title('Unemployment rate', fontsize=12)

ax[1, 0].plot(sample, irf[:, 3], linewidth=2, color='black')
ax[1, 0].fill_between(sample, ci_high95[:, 3], ci_low95[:, 3],
                      color='black', alpha=0.30, linewidth=0)
ax[1, 0].plot(sample, np.zeros(nirf), color='black', linewidth=0.5)
ax[1, 0].set_xlim(sample[0], sample[-1])
ax[1, 0].set_xlabel('Months', fontsize=12)
ax[1, 0].set_ylabel('Percent', fontsize=12)
ax[1, 0].set_yticks([0, 0.5, 1, 1.5])
ax[1, 0].tick_params(axis='x', labelsize=10)
ax[1, 0].tick_params(axis='y', labelsize=10)
ax[1, 0].set_title('Initial claims', fontsize=12)

ax[1, 1].plot(sample, irf[:, 4], linewidth=2, color='black')
ax[1, 1].fill_between(sample, ci_high95[:, 4], ci_low95[:, 4],
                      color='black', alpha=0.30, linewidth=0)
ax[1, 1].plot(sample, np.zeros(nirf), color='black', linewidth=0.5)
ax[1, 1].set_xlim(sample[0], sample[-1])
ax[1, 1].set_xlabel('Months', fontsize=12)
ax[1, 1].set_ylabel('Number of employees (Thousands)', fontsize=12)
ax[1, 1].set_yticks([-25, -20, -15, -10, -5, 0, 5])
ax[1, 1].tick_params(axis='x', labelsize=10)
ax[1, 1].tick_params(axis='y', labelsize=10)
ax[1, 1].set_title('Private employment changes', fontsize=12)

line1, = ax[1, 2].plot(sample, irf[:, 5], linewidth=2, color='black')
line2 = ax[1, 2].fill_between(sample, ci_high95[:, 5], ci_low95[:, 5],
                              color='black', alpha=0.30, linewidth=0)
ax[1, 2].plot(sample, np.zeros(nirf), color='black', linewidth=0.5)
ax[1, 2].set_xlim(sample[0], sample[-1])
ax[1, 2].set_xlabel('Months', fontsize=12)
ax[1, 2].set_ylabel('Number of workers (Thousands)', fontsize=12)
ax[1, 2].tick_params(axis='x', labelsize=10)
ax[1, 2].tick_params(axis='y', labelsize=10)
ax[1, 2].set_title('$\widehat{WARN}_t$ changes', fontsize=12)

fig.tight_layout()

# Save the figure
plt.savefig('IRFs.pdf', bbox_inches='tight')


#%% Disply the Run Time

# This line display the run time in minutes
print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


