# -*- coding: utf-8 -*-
"""
This program computes the WARN factor using the sample July 1996 through 
December 2019. It then computes the correlation between this factor and the 
factor estimated from July 1996 through November 2021. This provides the
result in footnote 12 of "Advance Layoff Notices and Aggregate Job Loss" by
Pawel M. Krolikowski and Kurt G. Lunsford.

This program calls the following files:
    seasonal_module.py : module with functions to seasonally adjust the data
    warn_module.py : module with functions to estimate the DFM
    WARNData_NSA_20211230.xlsx : Excel file with the state-level WARN data

This program print the correlation in footnote 12.
"""


#%% Import Packages and Functions

# Get replication files directory
import os
path_dir = os.path.dirname(os.getcwd())

# Insert the replication files directory to the system paths if necessary
import sys

check = 0
for jj in range(len(sys.path)):
    check = check + (sys.path[jj] == path_dir)

if check == 0:
    # Use 1 to make this the second directory in sys.path
    sys.path.insert(1, path_dir)

# Import Pandas and NumPy
import pandas as pd
import numpy as np

# Functions for seasonal adjustment
from modules.seasonal_module import sa_func

# Functions for estimating the DFM
from modules.dfm_em_module import make_guess, em_func

# Start time used to measure the run time of this program
import time
start_time = time.time()


#%% Parameters of the Program

"""
This section contains the parameters of the program.

The default values for the paper are as follows:
    str_date = '20211230'
    str_warn = '19960701'
    str_end = '20191231'
    nlags = 1
    nloop = 20000
    thresh_sa = 1e-3
    thresh = 1e-5
"""

# String with the date of the data to call
str_date = '20211230'   # Format: 20200115 corresponds to January 15, 2020

# String to choose first date for WARN data
str_warn = '19960701'   # Format: 19960701 corresponds to July 1, 1996

# String to remove pandemic observations
str_end = '20191231'

# Number of lags in the state equation
nlags = 1

# Parameters of the EM algorithm for seasonal adjustment
nloop = 20000       # maximum number of iterations in the EM algrorithm
thresh_sa = 1e-3    # threshold for convergence

# Parameters of the EM algorithm for the dynamic factor model
thresh = 1e-5       # threshold for convergence

"""
No further parameters need to be set beyond this line in the program.
"""


#%% Load the WARN Data

# Dataframe for number of workers affected by WARN notices by month and state
str_file_name = 'WARNData_NSA_' + str_date + '.xlsx'
path_warn_data = os.path.join(path_dir, 'data_input', str_file_name)
df_warn_all = pd.read_excel(path_warn_data, 'Sheet1', index_col=0)

# Drop small states or states with too many observations with a value of 0
df_warn = df_warn_all.drop(columns=['AK', 'ID', 'KY', 'ND', 'NM',
                                    'OK', 'RI', 'UT', 'WV'])

# Drop pandemic observations
df_warn = df_warn.loc[:str_end, :]

# Number of states in the sample
nvars = len(df_warn.columns)


#%% Loop to Seasonally Adjust the Data

# Dataframe to store seasonally adjusted data
df_warn_sa = pd.DataFrame(index=df_warn.index, columns=df_warn.columns)

# Loop for seasonal adjustment
for loop in range(nvars):
    # State name
    state_name = df_warn.columns[loop]
    
    # Dates with observed data
    state_index = df_warn.loc[:, state_name].dropna().index
    
    # Number of observations excluding NaNs
    nsample = len(state_index)
    
    # Convert the data to NumPy while dropping NaNs
    X_nsa = np.zeros((nsample, 1))
    X_nsa[:, 0] = df_warn.loc[:, state_name].dropna().to_numpy()
    
    # Seasonally adjust the data
    data_sa = sa_func(X_nsa, nloop, thresh_sa, state_name)
    
    # Store seasonally adjusted data in dataframe
    df_warn_sa.loc[state_index, state_name] = data_sa[:, 0]
    
    # Delete temprary variables
    del state_name, state_index, nsample, data_sa


#%% Adjust the Data and Construct the Selection Matrices for Missing Data

# Number of observations
nsample = len(df_warn_sa.loc[str_warn:, :].index)

# Convert the data to NumPy
X = np.zeros((nsample, nvars))
X[:, :] = df_warn_sa.loc[str_warn:, :].to_numpy()

# Set zero values to NaN and construct the selection matrices
W = np.zeros((nvars, nvars, nsample))
for jj in range(nsample):
    # Set zero values to NaN
    for ii in range(nvars):
        if X[jj, ii] < 1:
            X[jj, ii] = np.nan
    
    # Selection matrices
    W[:, :, jj] = np.diag(np.isnan(X[jj, :])==0)

# Delete temprary variables
del jj


#%% Initial Guess for the EM Algorithm

# Take logs of the data
X = np.log(X)

# Guess of model parameters
Fguess, Aguess, Qguess, Lguess, Rguess, Zstd, Avg, Var = make_guess(X, nlags)

# Guess of the initial state
F0guess = np.zeros((nlags, 1))

# Guess of covaraiance of initial state
Qvec = np.zeros((nlags*nlags, 1))
Qvec[0, 0] = Qguess[0, 0]

tmp = np.zeros((nlags*nlags, nlags*nlags))
tmp[:, :] = np.eye(nlags*nlags) - np.kron(Aguess, Aguess)

P0guess = np.zeros((nlags*nlags, 1))
P0guess[:, :] = np.linalg.solve(tmp, Qvec)
P0guess = np.reshape(P0guess, (nlags, nlags))

# Delete temprary variables
del Qvec, tmp


#%% EM Algorithm to Compute Model Parameters and WARN Factor

Fhat, Phat, Ahat, Qhat, Lhat, Rhat, Loglikes, ell = \
    em_func(F0guess, P0guess, Zstd, Lguess, Rguess, Aguess, Qguess,
            W, nloop, thresh)

# Check if the EM algoithm finished before reaching max iterations
if ell == nloop-1:
    print('No solution found with EM algorithm. '
          + 'Maximum number of iterations reached.')
else:
    print('Solution to EM algorithm found on loop ' + str(ell))

# Compute WARN factor in term of units of employees affected
tmp = np.zeros((nvars, 1))
warn_hat = np.zeros((nsample, 1))
for jj in range(nsample):
    tmp[:, 0] = np.multiply(np.sqrt(Var[:, 0]), (Lhat[:, 0]*Fhat[0, jj+1]))
    warn_hat[jj, 0] = np.sum(np.exp(Avg + tmp))

# Delete temprary variables
del ell, tmp, jj


#%% Load the WARN Factor from the Full Sample

# Load the data
df_factors = pd.read_excel('WARNFactors_20211230.xlsx', 'Sheet1', index_col=0)

# The WARN factor
Fhat_long = np.zeros((1, nsample))
Fhat_long[0, :] = df_factors.loc[str_warn:str_end, 'WARN Factor'].to_numpy()


#%% Correlation Between the Factors

# Compute the correlations
vals = np.zeros((2, nsample))
vals[0, :] = Fhat[0, 1:nsample+1]
vals[1, :] = Fhat_long[0, 0:nsample]
corr_lags = np.corrcoef(vals)

# Print the results
print(' ')
print('The correlation between a factor estimated on Jul 1996 through Dec 2019')
print('and a factor estimated on Jul 1996 through Nov 2021 is')
print(corr_lags[1, 0])
print(' ')


#%% Disply the Run Time in Minutes

print('This program finished in '
      + str(np.round((time.time()-start_time)/60, decimals=2)) 
      + ' minutes.')


